#!/usr/bin/perl

# Script to run cross-validated HMM models on the data.
# Sergey Kirshner, 01/09/2004

use POSIX qw(floor);

$num_args=$#ARGV+1;

if( $num_args==0 ) {
  # No arguments passed -- displaying help
  print "mvnhmm-data-stat.perl <data dimensionality> <number of sequences> <length of each sequence> <number of sequences to leave out (with cross-validation)>\n";
  print "Data dimensionality: >0\n";
  print "Number of sequences: >0\n";
  print "Length of each sequence: >0\n";
  print "Number of sequences to leave out with cross-validation: >=0.  Use 0 for no cross-validation\n";

  print "\n";
  print "In order for the script to run, two directories must be present in the execution directory: 'data' and 'outputs'.  The results of the runs will be placed into 'outputs' directory.  'data' directory must contain the dataset in the file 'data' (link to a data set).\n";
  exit( 0 );
} elsif( $num_args<4 ) {
  # Not enough arguments passed
  print "Need four arguments passed\n";
  exit( -1 );
} else {
  # Collecting necessary parameters from the command line

  # Data dimensionality
  $output_dim=$ARGV[0];
  if( $output_dim<0 ) {
    print "Invalid dimensionality of the data set: $output_dim\n";
    exit( -1 );
  }
  print "Dimensionality of the data: $output_dim\n";

  # Number of sequences
  $num_seqs=$ARGV[1];
  if( $num_seqs<1 ) {
    print "Invalid number of sequences: $num_seqs\n";
    exit( -1 );
  }
  print "Number of sequences: $num_seqs\n";

  # Length of each sequence
  $seq_length=$ARGV[2];
  if( $seq_length<1 ) {
    print "Invalid length of each sequence: $seq_length\n";
    exit( -1 );
  }
  print "Length of each sequence: $seq_length\n";

  # Number of sequences to leave out for cross-validation
  $xval_out=$ARGV[3];
  if( $xval_out<0 ) {
    print "Invalid number of sequences to leave out: $xval_out\n";
    exit( -1 );
  }
  if( $xval_out==0 ) {
    print "Cross-validation type: none\n";
    $xval_type="none\n";
  } elsif( $num_seqs/$xval_out<2 ) {
    print "Number of sequences to leave-out is too large -- should be at most a half of all examples\n";
    exit( -1 );
  } else {
    print "Cross-validation type: leave-$xval_out-out\n";
    $xval_type="leave_n_out";
  }
}

# Removing previously created parameter file
`rm -f params_current`;

# Analysis action can calculate a number (6) of statistics
@analysis_option=("mean", "correlation", "persistence", "dry", "wet", "information");
# 0: proportion of rainy days per station
# 1: correlation of occurences between pairs of stations
# 2: rain/no rain persistence per station
# 3: dry spell length distribution per station
# 4: wet spell length distribution per station
# 5: mutual information

# Need to loop over all statistics
$index_limit=6;

for( $mode_index=0; $mode_index<$index_limit; $mode_index++ ) {
  # Looping over possible modes

  # Building up the parameter file

  # Creating a file
  `touch params_current`;

  # Action analysis
  `echo 'action analysis' >> params_current`;

  # Evaluation type
  `echo 'analysis ${analysis_option[$mode_index]}' >> params_current`;

  # Number of data components
  `echo 'num_discrete_data_components $output_dim' >> params_current`;
  `echo 'num_real_data_components 0' >> param_current`;

  # Analysis uses the data
  `echo 'data data/data' >> params_current`;

  # Number of data sequences

  # Analysis action requires the total number of sequences (as simulated) to be listed

  # Total number of sequences is the number of cross-validated sets times the number of sequences
  # in each set (number left-out)
  if( $xval_out==0 ) {
    # No cross-validation
    $num_total_seqs=$num_seqs;
  } else {
    # Cross-validation
    $num_total_seqs=floor($num_seqs/$xval_out)*$xval_out;
  }

  `echo 'num_data_sequences $num_total_seqs' >> params_current`;

  # Length of each data sequence
  `echo 'data_sequence_length $seq_length' >> params_current`;

  # Cross-validation type
  `echo 'xval_type $xval_type' >> params_current`;
	
  if( $xval_type eq 'leave_n_out' ) {
    # Number of examples to leave-out for cross-validation
    `echo 'examples_out $xval_out' >> params_current`;
  }

  # Actions simulation and analysis require a number of simulations

  # Number of simulations
  `echo 'num_simulations 1' >> params_current`;

  # Output
  `echo 'output outputs/stat_${analysis_option[$mode_index]}' >> params_current`;

  # Running the experiment
  print "Calculating statistics ${analysis_option[$mode_index]}\n";

  `/bin/nice -19 mvnhmm params_current`;

  # Deleting the parameter file
  `rm -f params_current`;
}

