/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* distmupdate.cpp */

/* Includes routines for M-step update for distributions */

/* Includes */
#include "struct.h"

/* Powers of 2 from 0 to 30 */
static long power2[] = {
  1,         2,         4,          8,    
  16,        32,        64,         128, 
  256,       512,       1024,       2048, 
  4096,      8192,      16384,      32768,
  65536,     131072,    262144,     524288, 
  1048576,   2097152,   4194304,    8388608,   
  16777216,  33554432,  67108864,   134217728, 
  268435456, 536870912, 1073741824
};
extern long SINGULAR_FAIL;
extern long is_done;
extern long num_failed;
extern double maxent_epsilon;

void Distribution::UpdateEmissionMixture( Data* data, double** prob_s ) {

  double*** mix_comp_prob;

  /* Index variable(s) */
  long i,j,n,t;

  /* Temporary variable(s) */
  double sum;

  /* Allocating P(mixing component|X,S) */
  mix_comp_prob=new double**[num_states];
  for( i=0; i<num_states; i++ )
    {
      mix_comp_prob[i]=new double*[data->num_seqs];
      for( n=0; n<data->num_seqs; n++ )
	mix_comp_prob[i][n]=new double[data->sequence[n]->seq_length];
    }
  
  /* Need to calculate P(mixing component|X,S) */
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      {
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  {
	    if( t==0 )
	      mix_comp_prob[i][n][t]=state_prob[i]*
		subdist[i]->prob( data->sequence[n]->entry[t], NULL );
	    else
	      mix_comp_prob[i][n][t]=state_prob[i]*
		subdist[i]->prob( data->sequence[n]->entry[t],
				  data->sequence[n]->entry[t-1] );
	    sum+=mix_comp_prob[i][n][t];
	  }
	for( i=0; i<num_states; i++ )
	  mix_comp_prob[i][n][t]/=sum;
      }
  
  /* Updating P(mixing component|S) */
  sum=0.0;
  for( i=0; i<num_states; i++ )
    {
      state_prob[i]=0.0;
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  state_prob[i]+=prob_s[n][t]*mix_comp_prob[i][n][t];
      sum+=state_prob[i];
    }
  for( i=0; i<num_states; i++ )
    state_prob[i]/=sum;
  
  /* Passing weights of the data to the mixture components */
  for( i=0; i<num_states; i++ )
    {
      sum=0.0;
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  {	    
	    mix_comp_prob[i][n][t]*=prob_s[n][t];
	    sum+=mix_comp_prob[i][n][t];
	  }
      
      /* Updating mixture components */
      subdist[i]->UpdateEmissionParameters( data, mix_comp_prob[i], sum );
    }
  
  /* Deallocating P(mixing component|X,S) */
  for( i=0; i<num_states; i++ )
    {
      for( n=0; n<data->num_seqs; n++ )
	delete mix_comp_prob[i][n];
      delete mix_comp_prob[i];
    }
  delete mix_comp_prob;
  
  return;
}

void Distribution::UpdateEmissionBernoulli( Data* data, double** prob_s ) {
  /* !!! Univariate Bernoulli !!! */
  
  /* Index variable(s) */
  long b,n,t;

  /* Temporary variables */
  double sum;

  for( b=0; b<num_states; b++ )
    state_prob[b]=0.0;
  
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      if( !is_missing( data->sequence[n]->entry[t]->ddata[dim_index[0]] ) )
	state_prob[data->sequence[n]->entry[t]->ddata[dim_index[0]]]+=prob_s[n][t];
  
  /* Normalizing the probabilities */
  sum=0.0;
  for( b=0; b<num_states; b++ )
    sum+=state_prob[b];
  
  for( b=0; b<num_states; b++ )
    state_prob[b]/=sum;
  
  return;
}

void Distribution::UpdateEmissionConditionalBernoulli( Data* data, double** prob_s ) {
  /* !!! Unvariate conditional Bernoulli !!! */
  
  /* Index variable(s) */
  long b,n,t,i,j;

  /* Temporary variable(s) */
  double sum;

  for( b=0; b<num_states; b++ )
    {
      /* First sequence entry probabilities */
      state_prob[b]=0.0;
      for( n=0; n<data->num_seqs; n++ )
	{
	  if( data->sequence[n]->entry[0]->ddata[dim_index[0]]==b )
	    state_prob[b]+=prob_s[n][0];
	  
	  for( t=1; t<data->sequence[n]->seq_length; t++ )
	    {
	      if( data->sequence[n]->entry[t]->ddata[dim_index[0]]==b )
		cond_state_prob[data->sequence[n]->entry[t-1]->ddata[dim_index[0]]][b]+=prob_s[n][t];
	      
	      if( data->sequence[n]->entry[t]->ddata[dim_index[0]]==b )
		state_prob[b]+=prob_s[n][t];
	    }
	}
    }
  
  /* Normalizing */
  sum=0.0;
  for( b=0; b<num_states; b++ )
    sum+=state_prob[b];
  for( b=0; b<num_states; b++ )
    state_prob[b]/=sum;
  
  for( i=0; i<num_states; i++ )
    {
      sum=0.0;
      for( j=0; j<num_states; j++ )
	sum+=cond_state_prob[i][j];
      for( j=0; j<num_states; j++ )
	cond_state_prob[i][j]/=sum;
    }
  
  return;
}

void Distribution::UpdateEmissionChowLiu( Data* data, double** prob_s, double norm_const ) {

  double**** pair_prob;         // Projection on the pair of the variables   
  double** MI;                  // Mutual information
  double best_MI;               // Best value of mutual information in the current pass
  long* best_edge;              // The list of current best edges

  Envelope** envelope;
  long num_envelopes;
  long** node_index;
  double** cond_prob;
  double** temp_ind_prob;       // Future individual probabilities

  /* Updates */
  double** mult;                // Total update from instantiated nodes
  double** sent_backward;       // Update sent in a backward pass (to the parent)
  double** received_backward;   // Total update received from uninstantiated children
  double** received_forward;    // Update sent in a forward pass (from the parent)

  long* visited;              // Indicator of visited variables for bivariate probabilities

  /* Index variable(s) */
  long i,j,n,t,i1,i2,env,m1,b,e,ei,b1,b2,i_bi,m_bi,b_bi,e_bi;

  /* Temporary variable(s) */
  double sum;
  DataPoint* datum;
  double* temp_cond_prob;
  double** temp_received_backward;
  double** temp_sent_backward;

  /* Allocating the array of individual probabilities */
  temp_ind_prob=new double*[dim];
  for( i=0; i<dim; i++ )
    temp_ind_prob[i]=new double[num_states];

  /* Allocating the array of pairwise probabilities */
  pair_prob=new double***[dim];
  for( i=0; i<dim; i++ )
    {
      pair_prob[i]=new double**[i];
      for( j=0; j<i; j++ )
	{
	  pair_prob[i][j]=new double*[num_states];
	  for( i1=0; i1<num_states; i1++ )
	    pair_prob[i][j][i1]=new double[num_states];
	}
    }
  
  /* Allocating mutual information lower triangular matrix */
  MI=new double*[dim];
  for( i=0; i<dim; i++ )
    MI[i]=new double[dim];
  
  /* Array of best edges */
  best_edge=new long[dim-1];
  
  /* Array of envelopes */
  envelope=new Envelope*[dim];
  node_index=new long*[dim];
  for( i=0; i<dim; i++ )
    node_index[i]=new long[2];

  /* Initializing */
  
  /* Individual probability projections */
  for( i=0; i<dim; i++ )
    for( i1=0; i1<num_states; i1++ )
      temp_ind_prob[i][i1]=0.0;
  
  /* Pairwise probability projections */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      for( i1=0; i1<num_states; i1++ )
	for( i2=0; i2<num_states; i2++ )
	  pair_prob[i][j][i1][i2]=0.0;
  
  /* Mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      MI[i][j]=0.0;

  /* Initializing structures for count collection */
  mult=new double*[dim];
  for( i=0; i<dim; i++ )
    mult[i]=new double[num_states];
  sent_backward=new double*[dim];
  for( i=0; i<dim; i++ )
    sent_backward[i]=new double[num_states];
  received_backward=new double*[dim];
  for( i=0; i<dim; i++ )
    received_backward[i]=new double[num_states];
  received_forward=new double*[dim];
  for( i=0; i<dim; i++ )
    received_forward[i]=new double[num_states];
  cond_prob=new double*[dim];
  for( i=0; i<dim; i++ )
    cond_prob[i]=new double[num_states];
  temp_cond_prob=new double[num_states];
  temp_sent_backward=new double*[dim];
  for( i=0; i<dim; i++ )
    temp_sent_backward[i]=new double[num_states];
  temp_received_backward=new double*[dim];
  for( i=0; i<dim; i++ )
    temp_received_backward[i]=new double[num_states];
  visited=new long[dim];

  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      { /* Computing univariate and bivariate probabilities for variable in the current observation */
	datum=data->sequence[n]->entry[t];	

	/* Initializing node_index to -1 */
	for( i=0; i<dim; i++ )
	  node_index[i][0]=-1;

	/* Computing the envelopes */
	num_envelopes=compute_envelopes( datum, node_index, envelope );

	/* Initializing cond_prob in case the node is isolated */
	for( i=0; i<dim; i++ )
	  for( b=0; b<num_states; b++ )
	    cond_prob[i][b]=ind_prob[i][b];

	/* Computing posterior probabilities for missing values */ 
	for( env=0; env<num_envelopes; env++ )
	  /* For each envelope */
	  if( is_missing(datum->ddata[dim_index[envelope[env]->node[0]]]) )
	    /* Envelope contains missing values */
	    if( envelope[env]->num_nodes==1 )
	      { /* Envelope contains exactly one missing variable. */
		m1=envelope[env]->node[0];
		for( b=0; b<num_states; b++ )
		  cond_prob[m1][b]=ind_prob[m1][b];
		sum=0.0;
		for( b=0; b<num_states; b++ )
		  {/* Summing over the missing variable */
		    for( e=0; e<envelope[env]->num_edges; e++ )
		      {
			ei=envelope[env]->edge[e];
			i1=edge[ei][0];
			i2=edge[ei][1];
			
			if( m1==i1 )
			  cond_prob[m1][b]*=edge_prob[ei][b][datum->ddata[dim_index[i2]]]/
			    (ind_prob[i1][b]*ind_prob[i2][datum->ddata[dim_index[i2]]]);
			else
			  cond_prob[m1][b]*=edge_prob[ei][datum->ddata[dim_index[i1]]][b]/
			    (ind_prob[i1][datum->ddata[dim_index[i1]]]*ind_prob[i2][b]);		    
		      }
		    sum+=cond_prob[m1][b];
		  } /* Summing over the missing variable */
		
		for( b=0; b<num_states; b++ )
		  cond_prob[m1][b]/=sum;
	      } /* Envelope contains exactly one missing variable. */
	    else
	      { /* Envelope contains more than one missing variable. */

		/* Initializing multiplicative contributions of leaves connected to each
		   missing variable */
		for( i=0; i<envelope[env]->num_nodes; i++ )
		  for( b=0; b<num_states; b++ )
		    {
		      mult[envelope[env]->node[i]][b]=1.0;
		      /* No need to initialize sent messages. */
		      received_backward[envelope[env]->node[i]][b]=1.0;
		      /* No need to initialize received message in the forward pass 
			 since it is received from one source. */
		    }

		/* Propagating contributions in a backward pass */
		for( e=envelope[env]->num_edges-1; e>=0; e-- )
		  { /* Edges in a backwards pass */
		    ei=envelope[env]->edge[e];
		    i1=edge[ei][0];
		    i2=edge[ei][1];

		    if( !is_missing(datum->ddata[dim_index[i1]] ) )
		      /* Second node of the edge is missing. */
		      for( b=0; b<num_states; b++ )
			mult[i2][b]*=edge_prob[ei][datum->ddata[dim_index[i1]]][b]/ind_prob[i2][b];
		    //			  (ind_prob[i1][datum->ddata[dim_index[i1]]]*ind_prob[i2][b]);
		    else
		      if( !is_missing( datum->ddata[dim_index[i2]] ) )
			/* First node of the edge is missing. */
			for( b=0; b<num_states; b++ )
			  mult[i1][b]*=edge_prob[ei][b][datum->ddata[dim_index[i2]]]/ind_prob[i1][b];
		    //			    (ind_prob[i1][b]*ind_prob[i2][datum->ddata[dim_index[i2]]]);
		      else
			/* Both nodes are missing. */
			if( node_index[i1][1]<node_index[i2][1] )
			  { /* i1 is the parent of i2 */
			    /* Updating contributions to i1 */
			    for( b1=0; b1<num_states; b1++ )
			      { /* Computing the contribution to i1 with value value1 from child i2 */
				sent_backward[i2][b1]=0.0;
				for( b2=0; b2<num_states; b2++ )
				  sent_backward[i2][b1]+=
				    mult[i2][b2]*received_backward[i2][b2]*edge_prob[ei][b1][b2]/ind_prob[i1][b1];
				/* Updating the parent */
				received_backward[i1][b1]*=sent_backward[i2][b1];
			      } /* Computing the contribution to i1 with value b1 from child i2 */
			  } /* i1 is the parent of i2 */
			else
			  { /* i2 is the parent of i1 */
			    /* Updating contributions to i2 */
			    for( b2=0; b2<num_states; b2++ )
			      { /* Computing the contribution to i2 with value b2 from child i1 */
				sent_backward[i1][b2]=0.0;
				for( b1=0; b1<num_states; b1++ )
				  sent_backward[i1][b2]+=
				    mult[i1][b1]*received_backward[i1][b1]*edge_prob[ei][b1][b2]/ind_prob[i2][b2];
				/* Updating the parent */
				received_backward[i2][b2]*=sent_backward[i1][b2];
			      } /* Computing the contribution to i2 with value value2 from child i1 */
			  } /* i2 is the parent of i1 */
		  } /* Edges in a backwards pass */
		
		/* Updating the root of the envelope */
		m1=envelope[env]->node[0];
		sum=0.0;
		for( b=0; b<num_states; b++ )
		  {
		    received_forward[m1][b]=ind_prob[m1][b]*mult[m1][b]*received_backward[m1][b];
		    cond_prob[m1][b]=received_forward[m1][b];
		    sum+=cond_prob[m1][b];
		  }

		/* Normalizing */
		for( b=0; b<num_states; b++ )
		  cond_prob[m1][b]/=sum;

		/* Forwards pass */
		for( e=0; e<envelope[env]->num_edges; e++ )
		  { /* Edges in a forward pass */
		    ei=envelope[env]->edge[e];
		    i1=edge[ei][0];
		    i2=edge[ei][1];

		    /* Only care for the edges connecting two missing variables */
		    if( is_missing(datum->ddata[dim_index[i1]]) && is_missing(datum->ddata[dim_index[i2]]) )
		      /* Both variables are missing */
		      if( node_index[i1][1]<node_index[i2][1] )
			{ /* i1 is the parent of i2 */
			  /* Computing update from i1 to i2 */
			  for( b2=0; b2<num_states; b2++ )
			    {
			      received_forward[i2][b2]=0.0;
			      for( b1=0; b1<num_states; b1++ )
				received_forward[i2][b2]+=
				  received_forward[i1][b1]*(edge_prob[ei][b1][b2]/(ind_prob[i2][b2]*ind_prob[i1][b1]))
				  /sent_backward[i2][b1];
			    }

			  /* Computing the conditional probability for i2 */
			  sum=0.0;
			  for( b=0; b<num_states; b++ )
			    {
			      received_forward[i2][b]*=ind_prob[i2][b]*mult[i2][b]*received_backward[i2][b];
			      cond_prob[i2][b]=received_forward[i2][b];
			      sum+=cond_prob[i2][b];
			    }

			  /* Normalizing */
			  for( b=0; b<num_states; b++ )
			    cond_prob[i2][b]/=sum;
			} /* i1 is the parent of i2 */
		      else
			{ /* i2 is the parent of i1 */
			  /* Computing update from i2 to i1 */
			  for( b1=0; b1<num_states; b1++ )
			    {
			      received_forward[i1][b1]=0.0;
			      for( b2=0; b2<num_states; b2++ )
				received_forward[i1][b1]+=
				  received_forward[i2][b2]*(edge_prob[ei][b1][b2]/(ind_prob[i1][b1]*ind_prob[i2][b2]))
				  /sent_backward[i1][b2];
			    }

			  /* Computing the conditional probability for i1 */
			  sum=0.0;
			  for( b=0; b<num_states; b++ )
			    {
			      received_forward[i1][b]*=ind_prob[i1][b]*mult[i1][b]*received_backward[i1][b];
			      cond_prob[i1][b]=received_forward[i1][b];
			      sum+=cond_prob[i1][b];
			    }

			  /* Normalizing */
			  for( b=0; b<num_states; b++ )
			    cond_prob[i1][b]/=sum;
			} /* i2 is the parent of i1 */
		  } /* Edges in a forward pass */

		/* Computing bivariate probabilities by setting a value of each node and completing the inference */

		/* Last edge to consider in an envelope */
		e_bi=envelope[env]->num_edges;
		for( i_bi=envelope[env]->num_nodes-1; i_bi>=1; i_bi-- )
		  { /* Looping over nodes of the envelope */		    
		    m_bi=envelope[env]->node[i_bi];

		    /* Finding the edge with m_bi as a child */
		    do
		      {
			e_bi--;
			ei=envelope[env]->edge[e_bi];
			i1=edge[ei][0];
			i2=edge[ei][1];
		      }			
		    while( !is_missing(datum->ddata[dim_index[i1]]) ||
			   !is_missing(datum->ddata[dim_index[i2]]) );

		    for( b_bi=0; b_bi<num_states; b_bi++ )
		      { /* Looping on the value of the current missing node */
			/* Assigning the value for the last node */
			/* Resetting visited indicator */
			for( i=0; i<i_bi; i++ )
			  visited[envelope[env]->node[i]]=0;

			ei=envelope[env]->edge[e_bi];
			i1=edge[ei][0];
			i2=edge[ei][1];			

			/* Updating the parent of the current node */
			if( i2==m_bi )
			  { /* i1 is the parent of the current node */
			    /* Computing the contributions from the current node */
			    for( b1=0; b1<num_states; b1++ )
			      {
				temp_sent_backward[i2][b1]=edge_prob[ei][b1][b_bi]/ind_prob[i1][b1];

				/* Recomputing parent's received_backward value */
				temp_received_backward[i1][b1]=
				  received_backward[i1][b1]*temp_sent_backward[i2][b1]/sent_backward[i2][b1];
			      }
			    visited[i1]=1;
			  } /* i1 is the parent of the current node */
			else
			  { /* i2 is the parent of the current node */
			    /* Computing the contributions from the current node */
			    for( b2=0; b2<num_states; b2++ )
			      {
				temp_sent_backward[i1][b2]=edge_prob[ei][b_bi][b2]/ind_prob[i2][b2];

				/* Recomputing parent's received_backward value */
				temp_received_backward[i2][b2]=
				  received_backward[i2][b2]*temp_sent_backward[i1][b2]/sent_backward[i1][b2];
			      }

			    visited[i2]=1;
			  } /* i2 is the parent of the current node */

			/* Propagating contributions in a backward pass */
			for( e=e_bi-1; e>=0; e-- )
			  { /* Edges in a backwards pass */
			    ei=envelope[env]->edge[e];
			    i1=edge[ei][0];
			    i2=edge[ei][1];
			    
			    /* Only care for edges connecting two uninstantiated variables */
			    if( is_missing(datum->ddata[dim_index[i1]]) && is_missing(datum->ddata[dim_index[i2]]) )
			      /* Both nodes are missing. */
			      if( node_index[i1][1]<node_index[i2][1] )
				/* i1 is the parent of i2 */
				  
				/* Updating contributions to i1 */
				if( visited[i2] )
				  { /* i2 is on the path from changed node */
				    /* Updating contributions to i1 */
				    for( b1=0; b1<num_states; b1++ )
				      { /* Updating received_backward for i1 */
					temp_sent_backward[i2][b1]=0.0;
					for( b2=0; b2<num_states; b2++ )
					  temp_sent_backward[i2][b1]+=
					    mult[i2][b2]*temp_received_backward[i2][b2]*edge_prob[ei][b1][b2]/ind_prob[i1][b1];
					/* Updating the parent */					  
					temp_received_backward[i1][b1]=
					  received_backward[i1][b1]*temp_sent_backward[i2][b1]/sent_backward[i2][b1];
				      } /* Updating received_backward for i1 */
				    visited[i1]=1;
				  } /* i2 is on the path from changed node */
				else
				  { /* i2 is not on the path from changed node */
				    /* Setting temp_sent_backward and temp_received_backward */
				    for( b1=0; b1<num_states; b1++ )
				      temp_sent_backward[i2][b1]=sent_backward[i2][b1];
				    for( b2=0; b2<num_states; b2++ )
				      temp_received_backward[i2][b2]=received_backward[i2][b2];
				  } /* i2 is not on the path from changed node */
			      else
				/* i2 is the parent of i1 */
				  
				/* Updating contributions to i2 */
				if( visited[i1] )
				  { /* i1 is on the path from changed node */
				    /* Updating contributions to i2 */
				    for( b2=0; b2<num_states; b2++ )
				      { /* Updating received_backward for i2 */
					temp_sent_backward[i1][b2]=0.0;
					for( b1=0; b1<num_states; b1++ )
					  temp_sent_backward[i1][b2]+=
					    mult[i1][b1]*temp_received_backward[i1][b1]*edge_prob[ei][b1][b2]/ind_prob[i2][b2];
					/* Updating the parent */					  
					temp_received_backward[i2][b2]=
					  received_backward[i2][b2]*temp_sent_backward[i1][b2]/sent_backward[i1][b2];
				      } /* Updating received_backward for i2 */
				    visited[i2]=1;
				  } /* i1 is on the path from changed node */
				else
				  { /* i1 is not on the path from changed node */
				    /* Setting temp_sent_backward and temp_received_backward */
				    for( b2=0; b2<num_states; b2++ )
				      temp_sent_backward[i1][b2]=sent_backward[i1][b2];
				    for( b1=0; b1<num_states; b1++ )
				      temp_received_backward[i1][b1]=received_backward[i1][b1];
				  } /* i1 is not on the path from changed node */
			    
			  } /* Edges in a backwards pass */
			
			/* Updating the root of the envelope */
			m1=envelope[env]->node[0];
			sum=0.0;
			for( b=0; b<num_states; b++ )
			  {
			    received_forward[m1][b]=ind_prob[m1][b]*mult[m1][b]*temp_received_backward[m1][b];
			    temp_cond_prob[b]=received_forward[m1][b];
			    sum+=temp_cond_prob[b];
			  }

			/* Normalizing */
			for( b=0; b<num_states; b++ )
			  temp_cond_prob[b]/=sum;
			
			/* Updating the bivariate count table */
			for( b=0; b<num_states; b++ )
			  if( m1>m_bi )
			    pair_prob[m1][m_bi][b][b_bi]+=cond_prob[m_bi][b_bi]*temp_cond_prob[b]*prob_s[n][t];
			  else
			    pair_prob[m_bi][m1][b_bi][b]+=cond_prob[m_bi][b_bi]*temp_cond_prob[b]*prob_s[n][t];

			/* Forwards pass */
			for( e=0; e<e_bi; e++ )
			  { /* Edges in a forward pass */
			    ei=envelope[env]->edge[e];
			    i1=edge[ei][0];
			    i2=edge[ei][1];
			    
			    /* Only care for the edges connecting two missing variables */
			    if( is_missing(datum->ddata[dim_index[i1]]) && is_missing(datum->ddata[dim_index[i2]]) )
			      /* Both variables are missing */
			      if( node_index[i1][1]<node_index[i2][1] )
				{ /* i1 is the parent of i2 */
				  /* Computing update from i1 to i2 */
				  for( b2=0; b2<num_states; b2++ )
				    {
				      received_forward[i2][b2]=0.0;
				      for( b1=0; b1<num_states; b1++ )
					received_forward[i2][b2]+=
					  received_forward[i1][b1]*(edge_prob[ei][b1][b2]/(ind_prob[i2][b2]*ind_prob[i1][b1]))
					  /temp_sent_backward[i2][b1];
				    }
				  
				  /* Computing the conditional probability for i2 */
				  sum=0.0;
				  for( b=0; b<num_states; b++ )
				    {
				      received_forward[i2][b]*=ind_prob[i2][b]*mult[i2][b]*temp_received_backward[i2][b];
				      temp_cond_prob[b]=received_forward[i2][b];
				      sum+=temp_cond_prob[b];
				    }
				      
				  /* Normalizing */
				  for( b=0; b<num_states; b++ )
				    temp_cond_prob[b]/=sum;
				      
				  /* Updating the bivariate count table */
				  for( b=0; b<num_states; b++ )
				    if( i2>m_bi )
				      pair_prob[i2][m_bi][b][b_bi]+=cond_prob[m_bi][b_bi]*temp_cond_prob[b]*prob_s[n][t];
				    else
				      pair_prob[m_bi][i2][b_bi][b]+=cond_prob[m_bi][b_bi]*temp_cond_prob[b]*prob_s[n][t];
				} /* i1 is the parent of i2 */
			      else
				{ /* i2 is the parent of i1 */
				  /* Computing update from i2 to i1 */
				  for( b1=0; b1<num_states; b1++ )
				    {
				      received_forward[i1][b1]=0.0;
				      for( b2=0; b2<num_states; b2++ )
					received_forward[i1][b1]+=
					  received_forward[i2][b2]*(edge_prob[ei][b1][b2]/(ind_prob[i1][b1]*ind_prob[i2][b2]))
					  /temp_sent_backward[i1][b2];
				    }
				      
				  /* Computing the conditional probability for i1 */
				  sum=0.0;
				  for( b=0; b<num_states; b++ )
				    {
				      received_forward[i1][b]*=ind_prob[i1][b]*mult[i1][b]*temp_received_backward[i1][b];
				      temp_cond_prob[b]=received_forward[i1][b];
				      sum+=temp_cond_prob[b];
				    }
				      
				  /* Normalizing */
				  for( b=0; b<num_states; b++ )
				    temp_cond_prob[b]/=sum;
				  
				  /* Updating the bivariate count table */
				  for( b=0; b<num_states; b++ )
				    if( i1>m_bi )
				      pair_prob[i1][m_bi][b][b_bi]+=cond_prob[m_bi][b_bi]*temp_cond_prob[b]*prob_s[n][t];
				    else
				      pair_prob[m_bi][i1][b_bi][b]+=cond_prob[m_bi][b_bi]*temp_cond_prob[b]*prob_s[n][t];
				} /* i2 is the parent of i1 */
			  } /* Edges in a forward pass */
		      } /* Looping on the value of the current missing node */
		  } /* Looping over nodes of the envelope */
	      } /* Envelope contains more than one missing variable. */
	
	/* Deallocating the envelopes */
	for( env=0; env<num_envelopes; env++ )
	  delete envelope[env];

	/* Collecting the counts */
	for( i=0; i<dim; i++ )
	  if( is_missing( data->sequence[n]->entry[t]->ddata[dim_index[i]] ) )
	    for( b=0; b<num_states; b++ )
	      temp_ind_prob[i][b]+=prob_s[n][t]*cond_prob[i][b];
	  else
	    temp_ind_prob[i][data->sequence[n]->entry[t]->ddata[dim_index[i]]]+=prob_s[n][t];

	for( i=1; i<dim; i++ )
	  for( j=0; j<i; j++ )
	    if( is_missing( data->sequence[n]->entry[t]->ddata[dim_index[i]] ) )
	      { /* First node of the pair is missing */
		if( !is_missing( data->sequence[n]->entry[t]->ddata[dim_index[j]] ) )
		  /* Second node of the pair is not missing */
		  for( b=0; b<num_states; b++ )
		    pair_prob[i][j][b][data->sequence[n]->entry[t]->ddata[dim_index[j]]]+=
		      prob_s[n][t]*cond_prob[i][b];
		else
		  /* Both nodes are missing */
		  if( (node_index[i][0]!=node_index[j][0]) || node_index[i][0]==-1 )
		    /* Nodes come from different envelopes or one is not from an envelope at all */
		    for( b1=0; b1<num_states; b1++ )
		      for( b2=0; b2<num_states; b2++ )
			pair_prob[i][j][b1][b2]+=prob_s[n][t]*cond_prob[i][b1]*cond_prob[j][b2];
	      } /* First node of the pair is missing */
	    else
	      { /* First node of the pair is not missing */
		if( is_missing( data->sequence[n]->entry[t]->ddata[dim_index[j]] ) )
		  /* Second node of the pair is missing */
		  for( b=0; b<num_states; b++ )
		    pair_prob[i][j][data->sequence[n]->entry[t]->ddata[dim_index[i]]][b]+=
		      prob_s[n][t]*cond_prob[j][b];
		else
		  /* Both nodes are present */
		  pair_prob[i][j][data->sequence[n]->entry[t]->ddata[dim_index[i]]]
		    [data->sequence[n]->entry[t]->ddata[dim_index[j]]]+=prob_s[n][t];		  
	      } /* First node of the pair is not missing */
      } /* Computing univariate and bivariate probabilities for variable in the current observation */

  /* Calculating probability distribution projections from the counts */
  /* The normalizing factor is the normalizing constant! */
  for( i=0; i<dim; i++ )
    {
      sum=0.0;
      for( i1=0; i1<num_states; i1++ )
	sum+=temp_ind_prob[i][i1];
      
      for( i1=0; i1<num_states; i1++ )
	ind_prob[i][i1]=temp_ind_prob[i][i1]/sum;
    }  

  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      {
	sum=0.0;
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    sum+=pair_prob[i][j][i1][i2];
	
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    pair_prob[i][j][i1][i2]/=sum;
      }

  /* Deallocating the array of envelopes */
  delete envelope;
  for( i=0; i<dim; i++ )
    delete node_index[i];
  delete node_index;

  /* Deallocating count collection structures */
  for( i=0; i<dim; i++ )
    delete mult[i];
  delete mult;
  for( i=0; i<dim; i++ )
    delete sent_backward[i];
  delete sent_backward;
  for( i=0; i<dim; i++ )
    delete received_backward[i];
  delete received_backward;
  for( i=0; i<dim; i++ )
    delete received_forward[i];
  delete received_forward;
  for( i=0; i<dim; i++ )
    delete cond_prob[i];
  delete cond_prob;
  delete temp_cond_prob;
  for( i=0; i<dim; i++ )
    delete temp_sent_backward[i];
  delete temp_sent_backward;
  for( i=0; i<dim; i++ )
    delete temp_received_backward[i];
  delete temp_received_backward;
  delete visited;

  /* Calculating mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      for( i1=0; i1<num_states; i1++ )
	{
	  MI[i][j]-=xlogx(ind_prob[i][i1]);
	  MI[i][j]-=xlogx(ind_prob[j][i1]);
	  for( i2=0; i2<num_states; i2++ )
	    MI[i][j]+=xlogx(pair_prob[i][j][i1][i2]);
	}

  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      MI[j][i]=MI[i][j];
  
  /* Initializing the list of connected nodes */
  node_used[dim-1]=1;
  for( i=0; i<dim-1; i++ )
    node_used[i]=0;
  
  /* Initializing the array of best edges */
  /* Current best edge is always from the only "attached" node */
  for( i=0; i<dim-1; i++ )
    best_edge[i]=dim-1;
  
  /* Inserting dim-1 edges */
  for( num_edges=0; num_edges<dim-1; num_edges++ )
    {
      best_MI=NEG_INF;
      i1=-1;
      for( i=0; i<dim-1; i++ )
	if( !node_used[i] )
	  if( MI[i][best_edge[i]]>best_MI )
	    { /* Found a better edge */
	      i1=i;
	      best_MI=MI[i][best_edge[i]];
	    } /* Found a better edge */
      
      /* Adding the edge */
      node_used[i1]=1;
      if( i1>best_edge[i1] )
	{
	  edge[num_edges][0]=i1;
	  edge[num_edges][1]=best_edge[i1];
	  for( i=0; i<num_states; i++ )
	    for( j=0; j<num_states; j++ )
	      edge_prob[num_edges][i][j]=pair_prob[i1][best_edge[i1]][i][j];
	}
      else
	{
	  edge[num_edges][0]=best_edge[i1];
	  edge[num_edges][1]=i1;
	  for( i=0; i<num_states; i++ )
	    for( j=0; j<num_states; j++ )
	      edge_prob[num_edges][i][j]=pair_prob[best_edge[i1]][i1][i][j];
	}
      
      /* Storing the information about the edge */
      edge_MI[num_edges]=best_MI;
      /* Rank is not yet available */
      
      /* Adjusting the list of best edges to not-yet-connected nodes */
      for( i=0; i<dim-1; i++ )
	if( !node_used[i] )
	  if( MI[i][i1]>MI[i][best_edge[i]] )
	    best_edge[i]=i1;
    }
  
  /* Calculating the ranks of the edges */
  for( i=0; i<dim-1; i++ )
    edge_rank[i]=1;
  
  for( i=0; i<dim-1; i++ )
    for( j=0; j<i; j++ )
      if( edge_MI[i]<edge_MI[j] )
	edge_rank[i]++;
      else
	edge_rank[j]++;
  
  /* !!! Pruning edges if needed !!! */
  /* Ignoring CONST_MAX_NUM_EDGES */
  /* Ignoring cl_threshold */
  
  /* Deallocating temporary individual probabilities */
  for( i=0; i<dim; i++ )
    delete temp_ind_prob[i];
  delete temp_ind_prob;

  /* Deallocaing pairwise probabilities */
  for( i=0; i<dim; i++ )
    {
      for( j=0; j<i; j++ )
	{
	  for( i1=0; i1<num_states; i1++ )
	    delete pair_prob[i][j][i1];
	  delete pair_prob[i][j];
	}
      delete pair_prob[i];
    }
  delete pair_prob;
  
  /* Deallocating mutual information */
  for( i=0; i<dim; i++ )
    delete MI[i];
  delete MI;
  
  /* Deallocating the array of best edges */
  delete best_edge;
  
  return;
}

void Distribution::UpdateEmissionConditionalChowLiu( Data* data, double** prob_s ) {
  /* Indices:
     0,..,dim-1 -- previous entry observations
     dim,..,2*dim-1 -- current entry observations
  */

  /* Allocating the array of pairwise probabilities */
  /* pair_prob[i][j][i1][i2]:
     i is the node for the current observation
     j is the node for the previous observation
     i1 -- value for node i
     i2 -- value for node j */
  
  double**** pair_prob;         // Projection on the pair of the variables
  double** MI;                  // Mutual information
  long* best_edge;              // The list of current best edges
  double best_MI;               // Best value of mutual information in the current pass

  /* Index variable(s) */
  long i,j,i1,i2,n,t;

  /* Temporary variable(s) */
  double sum;

  pair_prob=new double***[2*dim];
  for( i=dim; i<2*dim; i++ )
    {
      pair_prob[i]=new double**[i];
      for( j=0; j<i; j++ )
	{
	  pair_prob[i][j]=new double*[num_states];
	  for( i1=0; i1<num_states; i1++ )
	    pair_prob[i][j][i1]=new double[num_states];
	}
    }
  
  /* Allocating mutual information lower triangular matrix */
  MI=new double*[2*dim];
  for( i=0; i<2*dim; i++ )
    MI[i]=new double[2*dim];
  
  /* Array of best edges */
  best_edge=new long[2*dim];
  
  /* Initializing */
  
  /* Individual probability projections */
  for( i=0; i<2*dim; i++ )
    for( i1=0; i1<num_states; i1++ )
      ind_prob[i][i1]=0.0;
  
  /* Pairwise probability projections */
  for( i=dim; i<2*dim; i++ )
    for( j=0; j<i; j++ )
      for( i1=0; i1<num_states; i1++ )
	for( i2=0; i2<num_states; i2++ )
	  pair_prob[i][j][i1][i2]=0.0;
  
  /* Mutual information */
  for( i=dim; i<2*dim; i++ )
    for( j=0; j<i; j++ )
      MI[i][j]=0.0;
  
  /* Looping through the data to collect the counts */
  for( n=0; n<data->num_seqs; n++ )
    {
      for( t=0; t<data->sequence[n]->seq_length; t++ )
	for( i=0; i<dim; i++ )
	  { /* !!! Checking for missing data !!! */
	    if( !is_missing( data->sequence[n]->entry[t]->ddata[dim_index[i]] ) )
	      {
		ind_prob[i+dim][data->sequence[n]->entry[t]->ddata[dim_index[i]]]+=prob_s[n][t];
		for( j=0; j<i; j++ )
		  /* !!! !!! */
		  if( !is_missing( data->sequence[n]->entry[t]->ddata[dim_index[j]] ) )
		    pair_prob[i+dim][j+dim][data->sequence[n]->entry[t]->ddata[dim_index[i]]]
		      [data->sequence[n]->entry[t]->ddata[dim_index[j]]]+=prob_s[n][t];
	      }
	  } /* !!! Checking for missing data !!! */
      
      for( t=1; t<data->sequence[n]->seq_length; t++ )
	for( i=0; i<dim; i++ )
	  { /* !!! Checking for missing data !!! */
	    if( !is_missing( data->sequence[n]->entry[t-1]->ddata[dim_index[i]] ) )
	      {
		ind_prob[i][data->sequence[n]->entry[t-1]->ddata[dim_index[i]]]+=prob_s[n][t];
		for( j=0; j<dim; j++ )
		  /* !!! !!! */
		  if( !is_missing( data->sequence[n]->entry[t]->ddata[dim_index[j]] ) )
		    pair_prob[j+dim][i][data->sequence[n]->entry[t]->ddata[dim_index[j]]]
		      [data->sequence[n]->entry[t-1]->ddata[dim_index[i]]]+=prob_s[n][t];
	      }
	  } /* !!! Checking for missing data !!! */
    }
  
  /* Calculating probability distribution projections from the counts */
  /* The normalizing factor is the normalizing constant! */
  for( i=0; i<2*dim; i++ )
    {
      sum=0.0;
      for( i1=0; i1<num_states; i1++ )
	sum+=ind_prob[i][i1];
      
      for( i1=0; i1<num_states; i1++ )
	ind_prob[i][i1]/=sum;
    }
  
  for( i=dim; i<2*dim; i++ )
    for( j=0; j<i; j++ )
      {
	sum=0.0;
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    sum+=pair_prob[i][j][i1][i2];
	
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    pair_prob[i][j][i1][i2]/=sum;
      }
  
  /* Calculating mutual information */
  for( i=dim; i<2*dim; i++ )
    for( j=0; j<i; j++ )
      for( i1=0; i1<num_states; i1++ )
	{
	  MI[i][j]-=xlogx(ind_prob[i][i1]);
	  MI[i][j]-=xlogx(ind_prob[j][i1]);
	  for( i2=0; i2<num_states; i2++ )
	    MI[i][j]+=xlogx(pair_prob[i][j][i1][i2]);
	}
  
  for( i=dim; i<2*dim; i++ )
    for( j=0; j<i; j++ )
      MI[j][i]=MI[i][j];
  
  /* Initializing the list of connected nodes */
  for( i=0; i<dim; i++ )
    node_used[i]=1;
  for( i=dim; i<2*dim; i++ )
    node_used[i]=0;
  
  /* Initializing the array of best edges */
  /* Initially, the best edges are computed from previous entry measurements to the current */
  for( i=dim; i<2*dim; i++ )
    { /* To node i */
      best_edge[i]=-1;
      best_MI=NEG_INF;
      for( j=0; j<dim; j++ )
	if( MI[i][j]>best_MI )
	  {
	    best_MI=MI[i][j];
	    best_edge[i]=j;
	  }	
    } /* To node i */
  
  /* Inserting dim edges */
  for( num_edges=0; num_edges<dim; num_edges++ )
    {
      best_MI=NEG_INF;
      i1=-1;
      for( i=dim; i<2*dim; i++ )
	if( !node_used[i] )
	  if( MI[i][best_edge[i]]>best_MI )
	    { /* Found a better edge */
	      i1=i;
	      best_MI=MI[i][best_edge[i]];
	    } /* Found a better edge */
      
      /* Adding the edge */
      node_used[i1]=1;
      if( i1>best_edge[i1] )
	{
	  edge[num_edges][0]=i1;
	  edge[num_edges][1]=best_edge[i1];
	  for( i=0; i<num_states; i++ )
	    for( j=0; j<num_states; j++ )
	      edge_prob[num_edges][i][j]=pair_prob[i1][best_edge[i1]][i][j];
	}
      else
	{
	  edge[num_edges][0]=best_edge[i1];
	  edge[num_edges][1]=i1;
	  for( i=0; i<num_states; i++ )
	    for( j=0; j<num_states; j++ )
	      edge_prob[num_edges][i][j]=pair_prob[best_edge[i1]][i1][i][j];
	}
      
      /* Storing the information about the edge */
      edge_MI[num_edges]=best_MI;
      /* Rank is not yet available */
      
      /* Adjusting the list of best edges to not-yet-connected nodes */
      for( i=dim; i<2*dim; i++ )
	if( !node_used[i] )
	  if( MI[i][i1]>MI[i][best_edge[i]] )
	    best_edge[i]=i1;
    }
  
  /* Calculating the ranks of the edges */
  for( i=0; i<dim; i++ )
    edge_rank[i]=1;
  
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      if( edge_MI[i]<edge_MI[j] )
	edge_rank[i]++;
      else
	edge_rank[j]++;
  
  /* Post pruning */
  /* CONST_MAX_NUM_EDGES */
  /* cl_threshold */
  
  /* Deallocaing pairwise probabilities */
  for( i=dim; i<2*dim; i++ )
    {
      for( j=0; j<i; j++ )
	{
	  for( i1=0; i1<num_states; i1++ )
	    delete pair_prob[i][j][i1];
	  delete pair_prob[i][j];
	}
      delete pair_prob[i];
    }
  delete pair_prob;
  
  /* Deallocating mutual information */
  for( i=0; i<2*dim; i++ )
    delete MI[i];
  delete MI;
  
  /* Deallocating the array of best edges */
  delete best_edge;
  
  return;
}

void Distribution::UpdateEmissionFullBivariateMaxEnt( Data* data, double** prob_s, double norm_const ) {
  /* !!! No missing values !!! */

  double **count;
  long num_combs;
  double* uprob;
  long num_unchanged;

  double nu,f;

  /* Index variable(s) */
  long i,j,n,t,i1,i2,b;

  /* Temporary variable(s) */
  double sum;

  /* Initializing the constraint counts */
  count=new double*[dim];
  for( i=0; i<dim; i++ )
    {
      count[i]=new double[i+1];
      for( j=0; j<i+1; j++ )
	count[i][j]=0.0;
    }
  
  /* Calculating counts */
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      for( i=0; i<dim; i++ )
	if( data->sequence[n]->entry[t]->ddata[dim_index[i]]==1 )
	  for( j=0; j<i+1; j++ )
	    if( data->sequence[n]->entry[t]->ddata[dim_index[j]]==1 )
	      count[i][j]+=prob_s[n][t];
  
  /* Normalizing counts */
  for( i=0; i<dim; i++ )
    for( j=0; j<i+1; j++ )
      {
	count[i][j]/=norm_const;
	if( count[i][j]<CONSTRAINT_MIN_VALUE )
	  count[i][j]=CONSTRAINT_MIN_VALUE;
	else if( count[i][j]>CONSTRAINT_MAX_VALUE )
	  count[i][j]=CONSTRAINT_MAX_VALUE;
      }
  
  
  /* Initializing the parameters */
  for( i=0; i<dim; i++ )
    for( j=0; j<i+1; j++ )
      sigma[i][j]=0.0;
  det=(double)dim*CONST_LN2;
  
  /* Determining the total number of combinations */
  num_combs=power2[dim];
  
  /* Allocating the array of unnormalized probabilities */
  uprob=new double[num_combs];
  
  /* Initializing the array */
  for( b=0; b<num_combs; b++ )
    uprob[b]=1.0;
  
  num_unchanged=0;
  i=0; j=0;
  
  while( num_unchanged<dim*(dim+1)/2 )
    {
      /* Calculating the probability mass for entries satisfying the constraint */
      sum=0.0;
      if( i==j )
	{ /* Univariate constraint */
	  for( i1=0; i1<power2[dim-i-1]; i1++ )
	    for( b=i1*power2[i+1]+power2[i]; b<(i1+1)*power2[i+1]; b++ )
	      sum+=uprob[b];
	  sum*=exp(-det);
	} /* Univariate constraint */
      else
	{ /* Bivariate constraint */
	  for( i1=0; i1<power2[dim-i-1]; i1++ )
	    for( i2=0; i2<power2[i-j-1]; i2++ )
	      for( b=i1*power2[i+1]+power2[i]+i2*power2[j+1]+power2[j];
		   b<i1*power2[i+1]+power2[i]+(i2+1)*power2[j+1]; b++ )
		sum+=uprob[b];
	  sum*=exp(-det);
	} /* Bivariate constraint */
      
      /* Gradient */
      nu=count[i][j]-sum;
      
      if( fabs(nu)>MAXENT_EPSILON )
	{ /* Constraint not satisfied -- updating the parameter */
	  if( sum>2.0*CONSTRAINT_MAX_VALUE || sum<0.5*CONSTRAINT_MIN_VALUE )
	    /* Too sensitive -- will skip for now */
	    num_unchanged++;
	  else
	    {
	      f=log(count[i][j])-log(1.0-count[i][j])+log(1.0-sum)-log(sum);
	      sigma[i][j]+=f;
	      f=exp(f);
	      
	      /* Updating the probabilities */
	      if( i==j )
		{ /* Univariate constraint */
		  for( i1=0; i1<power2[dim-i-1]; i1++ )
		    for( b=i1*power2[i+1]+power2[i]; b<(i1+1)*power2[i+1]; b++ )
		      uprob[b]*=f;
		} /* Univariate constraint */
	      else
		{ /* Bivariate constraint */
		  for( i1=0; i1<power2[dim-i-1]; i1++ )
		    for( i2=0; i2<power2[i-j-1]; i2++ )
		      for( b=i1*power2[i+1]+power2[i]+i2*power2[j+1]+power2[j];
			   b<i1*power2[i+1]+power2[i]+(i2+1)*power2[j+1]; b++ )
			uprob[b]*=f;
		} /* Bivariate constraint */
	      
	      /* Updating the normalization constant */
	      det-=log(1.0-count[i][j])-log(1.0-sum);
	      
	      num_unchanged=0;
	    }
	} /* Constraint not satisfied -- updating the parameter */
      else
	/* Constraint satisfied */
	num_unchanged++;
      
      /* Determining the next constraint to update */
      if( j==i )
	{
	  j=0;
	  if( i==dim-1 )
	    i=0;
	  else
	    i++;
	}
      else
	j++;
    }
  
  /* Deallocating the array of unnormalized probabilities */
  delete uprob;
  
  /* Deallocating the constraint counts */
  for( i=0; i<dim; i++ )
    delete count[i];
  delete count;
  
  return;
}

void Distribution::UpdateEmissionPUCMaxEnt( Data* data, double** prob_s, double norm_const ) {
  /* !!! Assuming no missing values !!! */
  
  double **score, **alpha;
  long num_points;
  long** flat_data;    // Flattened data
  double* w;           // Flat array of weights
  double **count;
  long *old_num_features, **old_feature_index;
  double baseline_ll;
  double new_ll;
  double** old_sigma;
  long from_index, to_index;
  double* uprob;
  long **descendant, **candidate;
  double* factor_ll;
  double best_score;

  /* Index variable(s) */
  long i,j,n,t,i1;

  /* Temporary variables */
  double temp1;

  /* Creating a flat data set */
  num_points=data->num_points();
  w=new double[num_points];
  flat_data=new long*[num_points];
  i1=0;
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      {
	flat_data[i1]=new long[dim];
	for( i=0; i<dim; i++ )
	  flat_data[i1][i]=data->sequence[n]->entry[t]->ddata[dim_index[i]];
	w[i1]=prob_s[n][t];
	i1++;
      }
  
  
  /* Computing mass of univariate and bivariate occurences */
  uprob=new double[dim];
  count=new double*[dim];
  for( i=0; i<dim; i++ )
    count[i]=new double[dim];
  
  for( i=0; i<dim; i++ )
    {
      uprob[i]=0.0;
      for( n=0; n<num_points; n++ )
	if( flat_data[n][i] )
	  uprob[i]+=w[n];
      
      /* Making sure the sum is not too small or too large */
      /* Guaranteeing the probability under the constraint to be in (1e-12, 1-1e-12) */
      if( uprob[i]>norm_const*CONSTRAINT_MAX_VALUE )
	uprob[i]=norm_const*CONSTRAINT_MAX_VALUE;
      else if( uprob[i]<norm_const*CONSTRAINT_MIN_VALUE )
	uprob[i]=norm_const*CONSTRAINT_MIN_VALUE;
      
      for( j=0; j<i; j++ )
	{
	  count[i][j]=0.0;
	  for( n=0; n<num_points; n++ )
	    if( flat_data[n][i]==1 && flat_data[n][j]==1 )
	      count[i][j]+=w[n];
	  
	  /* Making sure the sum is not too small or too large */
	  /* Guaranteeing the probability under the constraint to be in (1e-12, 1-1e-12) */
	  if( count[i][j]>norm_const*CONSTRAINT_MAX_VALUE )
	    count[i][j]=norm_const*CONSTRAINT_MAX_VALUE;
	  else if( count[i][j]<norm_const*CONSTRAINT_MIN_VALUE )
	    count[i][j]=norm_const*CONSTRAINT_MIN_VALUE;
	  
	  count[j][i]=count[i][j];
	}
      count[i][i]=uprob[i];
    }
  
  /* Computing parameters for the current structure */
  baseline_ll=0.0;
  /* Allocating and initializing the array for the old structure */
  old_num_features=new long[dim];
  old_feature_index=new long*[dim];
  old_sigma=new double*[dim];
  for( i=0; i<dim; i++ )
    {
      temp1=
	learn_univariate_conditional_maxent( num_points, flat_data, w, norm_const, i,
					     count[i], num_features[i], sigma[i], feature_index[i] );
      
      if( !finite(temp1) )
	{
	  /* Trying to set all parameters to zero */
	  for( j=0; j<num_features[i]; j++ )
	    sigma[i][j]=0.0;
	  
	  temp1=
	    learn_univariate_conditional_maxent( num_points, flat_data, w, norm_const, i,
						 count[i], num_features[i], sigma[i], feature_index[i] );
	  baseline_ll+=
	    learn_univariate_conditional_maxent( num_points, flat_data, w, norm_const, i,
						 count[i], num_features[i], sigma[i], feature_index[i] );
	  
	  if( !finite(temp1) )
	    /* Cannot find a solution */
	    temp1=NEG_INF;
	}
      
      baseline_ll+=temp1;
      
      old_num_features[i]=num_features[i];
      old_feature_index[i]=new long[dim];
      old_sigma[i]=new double[dim];
      for( j=0; j<num_features[i]; j++ )
	{
	  old_feature_index[i][j]=feature_index[i][j];
	  old_sigma[i][j]=sigma[i][j];
	}
    }
  
  /* Factor log-likelihood */
  factor_ll=new double[dim];
  
  /* Initializing to having only the univariate features */
  for( i=0; i<dim; i++ )
    {
      num_features[i]=1;
      feature_index[i][0]=i;
    }
  
  for( i=0; i<dim; i++ )
    {
      sigma[i][0]=log(uprob[i])-log(norm_const-uprob[i]);
      factor_ll[i]=uprob[i]*sigma[i][0]-norm_const*log(1.0+exp(sigma[i][0]));
    }
  
  /* Allocating and initializing the matrix indicating the descendants in the network */
  // descendant[i][j]=1 means that j is a descendant of i.  i is a descendant of i be default.
  descendant=new long*[dim];
  for( i=0; i<dim; i++ )
    {
      descendant[i]=new long[dim];
      for( j=0; j<dim; j++ )
	descendant[i][j]=0;
    }
  
  /* Allocating and initializing the matrix of candidate features for each factor */
  candidate=new long*[dim];
  for( i=0; i<dim; i++ )
    {
      candidate[i]=new long[dim];
      for( j=0; j<dim; j++ )
	candidate[i][j]=1;
      candidate[i][i]=0;
    }
  
  /* Initializing the array of scores */
  score=new double*[dim];
  for( i=0; i<dim; i++ )
    score[i]=new double[dim];
  
  /* Initializing the array of starting values for parameters as determined by the scoring procedure */
  alpha=new double*[dim];
  for( i=0; i<dim; i++ )
    alpha[i]=new double[dim];
  
#ifdef BLAH
  /* Computing the score for each candidate */
  for( i=0; i<dim; i++ )
    InformationGain( data, prob_s, norm_const, candidate[i], score[i], alpha[i], i );
#endif
  /* Efficient computation available when only one conditioning variable */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      {
	score[i][j]=xlogx(count[i][j])+xlogx(uprob[i]-count[i][j])+xlogx(uprob[j]-count[i][j])
	  +xlogx(norm_const+count[i][j]-uprob[i]-uprob[j])+xlogx(norm_const)
	  -xlogx(uprob[i])-xlogx(uprob[j])-xlogx(norm_const-uprob[i])-xlogx(norm_const-uprob[j]);
	score[i][j]/=norm_const;
	score[j][i]=score[i][j];
	/* alpha's are irrelevant for pairs */
	alpha[i][j]=0.0;
	alpha[j][i]=0.0;
      }
  
  temp1=1.0;
  best_score=0.0;
  /* Determining the best score */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      if( candidate[i][j] )
	if( score[i][j]>best_score+COMP_EPSILON )
	  {
	    best_score=score[i][j];
	    from_index=j;
	    to_index=i;
	  }
	else
	  {
	    if( fabs( score[i][j]-best_score )<=COMP_EPSILON )
	      { /* Two scores are identical -- deciding which edge to pick */
		temp1+=1.0;
		if( drand48()<1.0/temp1 )
		  { /* Pick the newer edge */
		    from_index=j;
		    to_index=i;
		  } /* Pick the newer edge */
	      } /* Two scores are identical -- deciding which edge to pick */
	  }
  
  while( best_score>maxent_epsilon )
    { /* Adding an edge */
      
      /* Updating the indices in the factor */
      feature_index[to_index][num_features[to_index]]=from_index;
      
      /* Initializing the value corresponding to the added feature */
      sigma[to_index][num_features[to_index]]=alpha[to_index][from_index];
      
      num_features[to_index]++;
      
      /* Updating the parameters */
      if( num_features[to_index]==2 )
	{ /* Special case of two variables with bivariate features */
	  sigma[to_index][0]=log(uprob[to_index]-count[to_index][from_index])
	    -log(norm_const+count[to_index][from_index]-uprob[to_index]-uprob[from_index]);
	  sigma[to_index][1]=log(count[to_index][from_index])
	    +log(norm_const+count[to_index][from_index]-uprob[to_index]-uprob[from_index])
	    -log(uprob[to_index]-count[to_index][from_index])
	    -log(uprob[from_index]-count[to_index][from_index]);
	  
	  factor_ll[to_index]=uprob[to_index]*sigma[to_index][0]+
	    count[to_index][from_index]*sigma[to_index][1]-
	    uprob[from_index]*log(1.0+exp(sigma[to_index][0]+sigma[to_index][1]))-
	    (norm_const-uprob[from_index])*log(1.0+exp(sigma[to_index][0]));
	} /* Special case of two variables with bivariate features */
      else
	factor_ll[to_index]=
	  learn_univariate_conditional_maxent( num_points, flat_data, w, norm_const, to_index, count[to_index], 
					       num_features[to_index], sigma[to_index], feature_index[to_index] );
      
      candidate[to_index][from_index]=0;
      
      /* Updating the list of descendants */
      descendant[from_index][to_index]=1;
      for( i=0; i<dim; i++ )
	if( descendant[to_index][i] )
	  descendant[from_index][i]=1;
      
      /* Updating descendant table for all nodes having to_index as a descendant */
      for( i=0; i<dim; i++ )
	if( descendant[i][from_index] )
	  for( j=0; j<dim; j++ )
	    if( descendant[i][j]==1 || descendant[from_index][j]==1 )
	      descendant[i][j]=1;
      
      /* Updating the list of candidates */
      for( i=0; i<dim; i++ )
	for( j=0; j<dim; j++ )
	  if( candidate[i][j] )
	    if( descendant[i][j] )
	      candidate[i][j]=0;
      
      /* Recomputing the scores for the factor to_index */
      InformationGain( num_points, flat_data, w, norm_const, count[to_index], candidate[to_index], 
		       score[to_index], alpha[to_index], to_index );
      
      temp1=1.0;
      best_score=0.0;
      /* Determining the best score */
      for( i=0; i<dim; i++ )
	for( j=0; j<dim; j++ )
	  if( candidate[i][j] )
	    if( score[i][j]>best_score+COMP_EPSILON )
	      {
		best_score=score[i][j];
		from_index=j;
		to_index=i;
	      }
	    else
	      if( fabs( score[i][j]-best_score )<=COMP_EPSILON )
		{ /* Two scores are identical -- deciding which edge to pick */
		  temp1+=1.0;
		  if( drand48()<1.0/temp1 )
		    { /* Pick the newer edge */
		      from_index=j;
		      to_index=i;
		    } /* Pick the newer edge */
		} /* Two scores are identical -- deciding which edge to pick */
    } /* Adding an edge */
  
  /* Computing the new log-likelihood */
  new_ll=0.0;
  for( i=0; i<dim; i++ )
    new_ll+=factor_ll[i];
  
  if( new_ll>baseline_ll )
    {
      /* Deleting arrays with previous structure */
      delete old_num_features;
      for( i=0; i<dim; i++ )
	{
	  delete old_feature_index[i];
	  delete old_sigma[i];
	}
      delete old_feature_index;
      delete old_sigma;
    }
  else
    {
      /* Deleting arrays with current structure */
      delete num_features;
      for( i=0; i<dim; i++ )
	{
	  delete feature_index[i];
	  delete sigma[i];
	}
      delete feature_index;
      delete sigma;
      
      /* Reusing the old structure */
      num_features=old_num_features;
      feature_index=old_feature_index;
      sigma=old_sigma;
    }
  
  /* Deallocating the arrays */
  for( i=0; i<dim; i++ )
    delete alpha[i];
  delete alpha;
  
  for( i=0; i<dim; i++ )
    delete score[i];
  delete score;
  
  for( i=0; i<dim; i++ )
    delete candidate[i];
  delete candidate;
  
  for( i=0; i<dim; i++ )
    delete descendant[i];
  delete descendant;
  
  delete uprob;
  for( i=0; i<dim; i++ )
    delete count[i];
  delete count;
  
  delete factor_ll;
  
  /* Deallocating flat data set */
  delete w;
  for( n=0; n<num_points; n++ )
    delete flat_data[n];
  delete flat_data;
    
  return;
}

void Distribution::UpdateEmissionConditionalPUCMaxEnt( Data* data, double** prob_s, double norm_const ) {
  /* !!! Assuming no missing values !!! */
  
  double **score, **alpha;
  long num_points;
  long** flat_data;    // Flattened data
  double* w;           // Flat array of weights
  double **count;
  long *old_num_features, **old_feature_index;
  double baseline_ll;
  double new_ll;
  double** old_sigma;
  long from_index, to_index;
  double* uprob;
  long **descendant, **candidate;
  double* factor_ll;
  double best_score;

  /* Index variable(s) */
  long i,j,n,t,i1;

  /* Temporary variables */
  double temp1;

  /* Creating a flat data set 
     
  Only time points 2..T are used.
  
  components 0..dim-1 -- current time point
  components dim..2*dim-1 -- previous time point
  */
  
  /* Computing first entry probability of one */
  for( i=0; i<dim; i++ )
    {
      state_prob[i]=0.0;
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  if( data->sequence[n]->entry[t]->ddata[dim_index[i]]==1 )
	    state_prob[i]+=prob_s[n][t];
      
      state_prob[i]/=norm_const;
    }
  
  num_points=data->num_points()-data->num_seqs;
  w=new double[num_points];
  flat_data=new long*[num_points];
  i1=0;
  for( n=0; n<data->num_seqs; n++ )
    for( t=1; t<data->sequence[n]->seq_length; t++ )
      {
	flat_data[i1]=new long[2*dim];
	for( i=0; i<dim; i++ )
	  {
	    flat_data[i1][i]=data->sequence[n]->entry[t]->ddata[dim_index[i]];
	    flat_data[i1][i+dim]=data->sequence[n]->entry[t-1]->ddata[dim_index[i]];
	  }
	w[i1]=prob_s[n][t];
	i1++;
      }
  
  /* Computing mass of univariate and bivariate occurences */
  uprob=new double[2*dim];
  count=new double*[dim];
  for( i=0; i<dim; i++ )
    count[i]=new double[2*dim];
  
  for( i=0; i<dim; i++ )
    {
      uprob[i]=0.0;
      for( n=0; n<num_points; n++ )
	if( flat_data[n][i]==1 )
	  uprob[i]+=w[n];
      
      uprob[i+dim]=0.0;
      for( n=0; n<num_points; n++ )
	if( flat_data[n][i+dim]==1 )
	  uprob[i+dim]+=w[n];
      
      /* Making sure the sum is not too small or too large */
      /* Guaranteeing the probability under the constraint to be in (1e-12, 1-1e-12) */
      if( uprob[i]>norm_const*CONSTRAINT_MAX_VALUE )
	uprob[i]=norm_const*CONSTRAINT_MAX_VALUE;
      else if( uprob[i]<norm_const*CONSTRAINT_MIN_VALUE )
	uprob[i]=norm_const*CONSTRAINT_MIN_VALUE;
      if( uprob[i+dim]>norm_const*CONSTRAINT_MAX_VALUE )
	uprob[i+dim]=norm_const*CONSTRAINT_MAX_VALUE;
      else if( uprob[i+dim]<norm_const*CONSTRAINT_MIN_VALUE )
	uprob[i+dim]=norm_const*CONSTRAINT_MIN_VALUE;
      
      for( j=i+1; j<2*dim; j++ )
	{
	  count[i][j]=0.0;
	  for( n=0; n<num_points; n++ )
	    if( flat_data[n][i]==1 && flat_data[n][j]==1 )
	      count[i][j]+=w[n];
	  
	  /* Making sure the sum is not too small or too large */
	  /* Guaranteeing the probability under the constraint to be in (1e-12, 1-1e-12) */
	  if( count[i][j]>norm_const*CONSTRAINT_MAX_VALUE )
	    count[i][j]=norm_const*CONSTRAINT_MAX_VALUE;
	  else if( count[i][j]<norm_const*CONSTRAINT_MIN_VALUE )
	    count[i][j]=norm_const*CONSTRAINT_MIN_VALUE;
	  
	  if( j<dim )
	    count[j][i]=count[i][j];
	}
      count[i][i]=uprob[i];
    }
  
  /* Computing parameters for the current structure */
  baseline_ll=0.0;
  /* Allocating and initializing the array for the old structure */
  old_num_features=new long[dim];
  old_feature_index=new long*[dim];
  old_sigma=new double*[dim];
  for( i=0; i<dim; i++ )
    {
      temp1=learn_univariate_conditional_maxent( num_points, flat_data, w, norm_const, i,
						 count[i], num_features[i], sigma[i], feature_index[i] );
      
      if( !finite(temp1) )
	{
	  /* Trying to set all parameters to zero */
	  for( j=0; j<num_features[i]; j++ )
	    sigma[i][j]=0.0;
	  
	  temp1=
	    learn_univariate_conditional_maxent( num_points, flat_data, w, norm_const, i,
						 count[i], num_features[i], sigma[i], feature_index[i] );
	  baseline_ll+=
	    learn_univariate_conditional_maxent( num_points, flat_data, w, norm_const, i,
						 count[i], num_features[i], sigma[i], feature_index[i] );
	  
	  if( !finite(temp1) )
	    /* Cannot find a solution */
	    temp1=NEG_INF;
	}
      
      baseline_ll+=temp1;
      
      old_num_features[i]=num_features[i];
      old_feature_index[i]=new long[dim];
      old_sigma[i]=new double[dim];
      for( j=0; j<num_features[i]; j++ )
	{
	  old_feature_index[i][j]=feature_index[i][j];
	  old_sigma[i][j]=sigma[i][j];
	}
    }
  
  /* Factor log-likelihood */
  factor_ll=new double[dim];
  
  /* Initializing to having only the univariate features */
  for( i=0; i<dim; i++ )
    {
      num_features[i]=1;
      feature_index[i][0]=i;
    }
  
  for( i=0; i<dim; i++ )
    {
      sigma[i][0]=log(uprob[i])-log(norm_const-uprob[i]);
      factor_ll[i]=uprob[i]*sigma[i][0]-norm_const*log(1.0+exp(sigma[i][0]));
    }
  
  /* Allocating and initializing the matrix indicating the descendants in the network */
  /* descendant[i][j]=1 means that j is a descendant of i.  i is a descendant of i be default. */
  descendant=new long*[dim];
  for( i=0; i<dim; i++ )
    {
      descendant[i]=new long[dim];
      for( j=0; j<dim; j++ )
	descendant[i][j]=0;
    }
  
  /* Allocating and initializing the matrix of candidate features for each factor */
  candidate=new long*[dim];
  for( i=0; i<dim; i++ )
    {
      candidate[i]=new long[2*dim];
      for( j=0; j<2*dim; j++ )
	candidate[i][j]=1;
      candidate[i][i]=0;
    }
  
  /* Initializing the array of scores */
  score=new double*[dim];
  for( i=0; i<dim; i++ )
    score[i]=new double[2*dim];
  
  /* Initializing the array of starting values for parameters as determined by the scoring procedure */
  alpha=new double*[dim];
  for( i=0; i<dim; i++ )
    alpha[i]=new double[2*dim];
  
#ifdef BLAH
  /* Computing the score for each candidate */
  for( i=0; i<dim; i++ )
    InformationGain( data, prob_s, norm_const, candidate[i], score[i], alpha[i], i );
#endif
  /* Efficient computation available when only one conditioning variable */
  for( i=0; i<dim; i++ )
    for( j=i+1; j<2*dim; j++ )
      {
	score[i][j]=xlogx(count[i][j])+xlogx(uprob[i]-count[i][j])+xlogx(uprob[j]-count[i][j])
	  +xlogx(norm_const+count[i][j]-uprob[i]-uprob[j])+xlogx(norm_const)
	  -xlogx(uprob[i])-xlogx(uprob[j])-xlogx(norm_const-uprob[i])-xlogx(norm_const-uprob[j]);
	score[i][j]/=norm_const;
	if( j<dim )
	  score[j][i]=score[i][j];
	/* alpha's are irrelevant for pairs */
	alpha[i][j]=0.0;
	if( j<dim )
	  alpha[j][i]=0.0;
      }
  
  temp1=1.0;
  best_score=0.0;
  /* Determining the best score */
  for( i=0; i<dim; i++ )
    for( j=0; j<2*dim; j++ )
      if( candidate[i][j] )
	if( score[i][j]>best_score+COMP_EPSILON )
	  {
	    best_score=score[i][j];
	    from_index=j;
	    to_index=i;
	  }
	else
	  {
	    if( fabs( score[i][j]-best_score )<=COMP_EPSILON )
	      { /* Two scores are identical -- deciding which edge to pick */
		temp1+=1.0;
		if( drand48()<1.0/temp1 )
		  { /* Pick the newer edge */
		    from_index=j;
		    to_index=i;
		  } /* Pick the newer edge */
	      } /* Two scores are identical -- deciding which edge to pick */
	  }
  
  while( best_score>MIN_ME_LL_CHANGE )
    { /* Adding an edge */
      
      /* Updating the indices in the factor */
      feature_index[to_index][num_features[to_index]]=from_index;
      
      /* Initializing the value corresponding to the added feature */
      sigma[to_index][num_features[to_index]]=alpha[to_index][from_index];
      
      num_features[to_index]++;
      
      /* Updating the parameters */
      if( num_features[to_index]==2 )
	{ /* Special case of two variables with bivariate features */
	  sigma[to_index][0]=log(uprob[to_index]-count[to_index][from_index])
	    -log(norm_const+count[to_index][from_index]-uprob[to_index]-uprob[from_index]);
	  sigma[to_index][1]=log(count[to_index][from_index])
	    +log(norm_const+count[to_index][from_index]-uprob[to_index]-uprob[from_index])
	    -log(uprob[to_index]-count[to_index][from_index])
	    -log(uprob[from_index]-count[to_index][from_index]);
	  
	  factor_ll[to_index]=uprob[to_index]*sigma[to_index][0]+
	    count[to_index][from_index]*sigma[to_index][1]-
	    uprob[from_index]*log(1.0+exp(sigma[to_index][0]+sigma[to_index][1]))-
	    (norm_const-uprob[from_index])*log(1.0+exp(sigma[to_index][0]));
	} /* Special case of two variables with bivariate features */
      else
	factor_ll[to_index]=
	  learn_univariate_conditional_maxent( num_points, flat_data, w, norm_const, to_index, count[to_index], 
					       num_features[to_index], sigma[to_index], feature_index[to_index] );
      
      candidate[to_index][from_index]=0;
      
      if( from_index<dim )
	{
	  /* Updating the list of descendants */
	  descendant[from_index][to_index]=1;
	  for( i=0; i<dim; i++ )
	    if( descendant[to_index][i] )
	      descendant[from_index][i]=1;
	  
	  /* Updating descendant table for all nodes having to_index as a descendant */
	  for( i=0; i<dim; i++ )
	    if( descendant[i][from_index] )
	      for( j=0; j<dim; j++ )
		if( descendant[i][j]==1 || descendant[from_index][j]==1 )
		  descendant[i][j]=1;
	  
	  /* Updating the list of candidates */
	  for( i=0; i<dim; i++ )
	    for( j=0; j<dim; j++ )
	      if( candidate[i][j] )
		if( descendant[i][j] )
		  candidate[i][j]=0;
	}
      
      /* Recomputing the scores for the factor to_index */
      InformationGain( num_points, flat_data, w, norm_const, count[to_index], candidate[to_index], 
		       score[to_index], alpha[to_index], to_index );
      
      temp1=1.0;
      best_score=0.0;
      /* Determining the best score */
      for( i=0; i<dim; i++ )
	for( j=0; j<2*dim; j++ )
	  if( candidate[i][j] )
	    if( score[i][j]>best_score+COMP_EPSILON )
	      {
		best_score=score[i][j];
		from_index=j;
		to_index=i;
	      }
	    else
	      if( fabs( score[i][j]-best_score )<=COMP_EPSILON )
		{ /* Two scores are identical -- deciding which edge to pick */
		  temp1+=1.0;
		  if( drand48()<1.0/temp1 )
		    { /* Pick the newer edge */
		      from_index=j;
		      to_index=i;
		    } /* Pick the newer edge */
		} /* Two scores are identical -- deciding which edge to pick */
    } /* Adding an edge */
  
  /* Computing the new log-likelihood */
  new_ll=0.0;
  for( i=0; i<dim; i++ )
    new_ll+=factor_ll[i];
  
  if( new_ll>baseline_ll )
    {
      /* Deleting arrays with previous structure */
      delete old_num_features;
      for( i=0; i<dim; i++ )
	{
	  delete old_feature_index[i];
	  delete old_sigma[i];
	}
      delete old_feature_index;
      delete old_sigma;
    }
  else
    {
      /* Deleting arrays with current structure */
      delete num_features;
      for( i=0; i<dim; i++ )
	{
	  delete feature_index[i];
	  delete sigma[i];
	}
      delete feature_index;
      delete sigma;
      
      /* Reusing the old structure */
      num_features=old_num_features;
      feature_index=old_feature_index;
      sigma=old_sigma;
    }
  
  /* Deallocating the arrays */
  for( i=0; i<dim; i++ )
    delete alpha[i];
  delete alpha;
  
  for( i=0; i<dim; i++ )
    delete score[i];
  delete score;
  
  for( i=0; i<dim; i++ )
    delete candidate[i];
  delete candidate;
  
  for( i=0; i<dim; i++ )
    delete descendant[i];
  delete descendant;
  
  delete uprob;
  for( i=0; i<dim; i++ )
    delete count[i];
  delete count;
  
  delete factor_ll;
  
  /* Deallocating flat data set */
  delete w;
  for( n=0; n<num_points; n++ )
    delete flat_data[n];
  delete flat_data;
  
  return;
}

void Distribution::UpdateEmissionDeltaExponential( Data* data, double** prob_s, double norm_const ) {

  /* Index variable(s) */
  long i,n,t;

  /* Temporary variable(s) */
  double sum;
  double *temp, *temp_prob_mix, *temp_exp_value;

  /* Flags */
  long leave_unchanged=0;

  /* Updating parameters for delta-exponential distribution */
  temp_prob_mix=new double[num_states];
  temp_exp_value=new double[num_states-1];
  temp=new double[num_states-1];
  
  /* Calculating mixing parameters and the exponential parameter */
  for( i=0; i<num_states; i++ )
    temp_prob_mix[i]=0.0;
  for( i=0; i<num_states-1; i++ )
    temp_exp_value[i]=0.0;
  
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      if( !is_missing( data->sequence[n]->entry[t]->rdata[dim_index[0]] ) )
	{
	  if( data->sequence[n]->entry[t]->rdata[dim_index[0]]==0.0 )
	    temp_prob_mix[0]+=prob_s[n][t];
	  else
	    {
	      sum=0.0;
	      for( i=0; i<num_states-1; i++ )
		{
		  temp[i]=mix_prob[i+1]*
		    exp_dist(data->sequence[n]->entry[t]->rdata[dim_index[0]],exp_param[i]);
		  sum+=temp[i];
		}
	      
	      for( i=0; i<num_states-1; i++ )
		{
		  if( sum==0.0 )
		    leave_unchanged=1;
		  temp[i]*=(prob_s[n][t]/sum);
		  temp_prob_mix[i+1]+=temp[i];
		  temp_exp_value[i]+=temp[i]*data->sequence[n]->entry[t]->rdata[dim_index[0]];
		}
	    }
	}
  
  /* Normalizing mixing probabilities */
  sum=0.0;
  for( i=0; i<num_states; i++ )
    if( temp_prob_mix[i]==0.0 )
      leave_unchanged=1;
    else
      sum+=temp_prob_mix[i];

  if( !leave_unchanged )
    {
      for( i=0; i<num_states; i++ )
	mix_prob[i]=temp_prob_mix[i]/sum;
      
	/* Parameters for the exponential components */
      for( i=0; i<num_states-1; i++ )
	exp_param[i]=temp_prob_mix[i+1]/temp_exp_value[i];
    }
  
  delete temp_prob_mix;
  delete temp_exp_value;
  delete temp;

  return;
}

void Distribution::UpdateEmissionDeltaGamma( Data* data, double** prob_s, double norm_const ) {

  double f,fprime;

  /* Index variable(s) */
  long i,n,t;

  /* Temporary variable(s) */
  double sum;
  double *temp, *temp_prob_mix, *temp_exp_value, *temp_exp_log;

  /* Flags */
  long leave_unchanged=0;

  /* Updating parameters for delta-gamma distribution */
  temp_prob_mix=new double[num_states];
  temp_exp_value=new double[num_states-1];
  temp_exp_log=new double[num_states-1];
  temp=new double[num_states-1];
  
  /* Calculating mixing parameters and the exponential parameter */
  for( i=0; i<num_states; i++ )
    temp_prob_mix[i]=0.0;
  for( i=0; i<num_states-1; i++ )
    {
      temp_exp_value[i]=0.0;
      temp_exp_log[i]=0.0;
    }
  
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      if( !is_missing( data->sequence[n]->entry[t]->rdata[dim_index[0]] ) )
	{
	  if( data->sequence[n]->entry[t]->rdata[dim_index[0]]==0.0 )
	      temp_prob_mix[0]+=prob_s[n][t];
	  else
	    {
	      sum=0.0;
	      for( i=0; i<num_states-1; i++ )
		{
		  temp[i]=mix_prob[i+1]*
		    gamma_dist(data->sequence[n]->entry[t]->rdata[dim_index[0]],
			       gamma_param1[i],gamma_param2[i]);
		  sum+=temp[i];
		}
	      for( i=0; i<num_states-1; i++ )
		{
		  if( sum==0.0 )
		    leave_unchanged=1;
		  temp[i]*=(prob_s[n][t]/sum);
		  temp_prob_mix[i+1]+=temp[i];
		  temp_exp_value[i]+=temp[i]*data->sequence[n]->entry[t]->rdata[dim_index[0]];
		  temp_exp_log[i]+=temp[i]*log(data->sequence[n]->entry[t]->rdata[dim_index[0]]);
		}
	    }
	}
  
  if( !leave_unchanged )
    {
      /* Parameters for the gamma components */
      for( i=0; i<num_states-1; i++ )
	{
	  temp[i]=gamma_param1[i];
	  
	  /* Starting from the previous value */
	  f=digamma(gamma_param1[i])-log(gamma_param1[i])-temp_exp_log[i]/temp_prob_mix[i+1]
	    -log(temp_prob_mix[i+1])+log(temp_exp_value[i]);
	  fprime=trigamma(gamma_param1[i])-1.0/gamma_param1[i];
	  
	  while( fabs(f)>NR_EPSILON && fabs(fprime)>NR_EPSILON )
	    {
	      gamma_param1[i]-=f/fprime;
	      f=digamma(gamma_param1[i])-log(gamma_param1[i])-temp_exp_log[i]/temp_prob_mix[i+1]
		-log(temp_prob_mix[i+1])+log(temp_exp_value[i]);
	      fprime=trigamma(gamma_param1[i])-1.0/gamma_param1[i];
	    }
	    
	  if( fabs(fprime)<NR_EPSILON || gamma_param1[i]<=0.0 )
	    gamma_param1[i]=temp[i];
	  else
	    gamma_param2[i]=gamma_param1[i]*temp_prob_mix[i+1]/temp_exp_value[i];
	}
    }
  
  /* Normalizing mixing probabilities */
  sum=0.0;
  for( i=0; i<num_states; i++ )
    if( temp_prob_mix[i]==0.0 )
      leave_unchanged=1;
    else
      sum+=temp_prob_mix[i];
  if( !leave_unchanged )
    for( i=0; i<num_states; i++ )
      mix_prob[i]=temp_prob_mix[i]/sum;
  
  delete temp_prob_mix;
  delete temp_exp_value;
  delete temp_exp_log;
  delete temp;

  return;
}

void Distribution::UpdateEmissionGaussian( Data* data, double** prob_s, double norm_const ) {

  /* Index variable(s) */
  long i,j,n,t;

  /* Zeroing out the mean */
  for( i=0; i<dim; i++ )
    mu[i]=0.0;
  
  /* Updating the mean */
  for( i=0; i<dim; i++ )
    { /* i-th component of the mean */
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  mu[i]+=(prob_s[n][t]*data->sequence[n]->entry[t]->rdata[dim_index[i]]);
      
      mu[i]/=norm_const;
    } /* i-th component of the mean */

  /* Zeroing out the covariance matrix */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      { 
	/* Storing sigma's values for the case of singular matrix */
	inv_sigma[i][j]=sigma[i][j];
	sigma[i][j]=0.0;
      }

  /* Updating the covariance matrix */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      {
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    sigma[i][j]+=(prob_s[n][t]*
			  (mu[i]-data->sequence[n]->entry[t]->rdata[dim_index[i]])*
			  (mu[j]-data->sequence[n]->entry[t]->rdata[dim_index[j]]));
	sigma[i][j]/=norm_const;
      }
  
  /* Calculating the inverse of the covariance matrix and its determinant */
  det=find_det( sigma, dim );
  if( SINGULAR_FAIL )
    { /* New matrix is singular -- replacing with the old one */
      /* !!! Need better solution !!! */
      for( i=0; i<dim; i++ )
	for( j=0; j<dim; j++ )
	  sigma[i][j]=inv_sigma[i][j];
      det=find_det( sigma, dim );
      
      if( !is_done )
	{ /* Stopping EM */
	  is_done=1;
	  num_failed++;
	} /* Stopping EM */
      
    } /* New matrix is singular -- replacing with the old one */
  
  find_inv( sigma, dim, inv_sigma );
  
  return;
}

void Distribution::UpdateEmissionARGaussian( Data* data, double** prob_s, double norm_const ) {

  double* prev_datum_ave;       // Weighted previous datum
  double **W_num, **W_denom;    // Matrices used to calculate W

  /* Index variable(s) */
  long i,j,n,t,i1,i2;

  /* Temporary variable(s) */
  double sum,temp1,temp2;

  /* First state updates */
  
  /* First state mean */
  for( i=0; i<dim; i++ )
    { /* i-th component of the first state mean */
      first_mu[i]=0.0;
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  first_mu[i]+=(prob_s[n][t]*data->sequence[n]->entry[t]->rdata[dim_index[i]]);
      first_mu[i]/=norm_const;
    } /* i-th component of the first state mean */
  
  /* First state covariance matrix */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      { /* (i,j)-th entry of the covariance matrix */
	/* Storing old values for the case the new matrix is singular */
	inv_first_sigma[i][j]=first_sigma[i][j];
	
	/* Updating */
	first_sigma[i][j]=0.0;
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    first_sigma[i][j]+=(prob_s[n][t]*
				(first_mu[i]-data->sequence[n]->entry[t]->rdata[dim_index[i]])*
				(first_mu[i]-data->sequence[n]->entry[t]->rdata[dim_index[j]]));	  
	first_sigma[i][j]/=norm_const;
      } /* (i,j)-th entry of the covariance matrix */
  
  /* Calculating the inverse of the first state covariance matrix and its determinant */
  first_det=find_det( first_sigma, dim );
  if( SINGULAR_FAIL )
    { /* New matrix is singular -- replacing with the old one */
      /* !!! Need better solution !!! */
      for( i=0; i<dim; i++ )
	for( j=0; j<dim; j++ )
	  first_sigma[i][j]=inv_first_sigma[i][j];
      first_det=find_det( first_sigma, dim );
      
      if( !is_done )
	{ /* Stopping EM */
	  is_done=1;
	  num_failed++;
	} /* Stopping EM */
      
    } /* New matrix is singular -- replacing with the old one */
  
  find_inv( first_sigma, dim, inv_first_sigma );
  
  if( SINGULAR_FAIL )
    return;
  
  /* Auto-regressive parameter updates */
  
  /* Normalization constant */
  sum=0.0;
  for( n=0; n<data->num_seqs; n++ )
    for( t=1; t<data->sequence[n]->seq_length; t++ )
      sum+=prob_s[n][t];
  
  /* Linear transformation matrix */
  
  /* Calculating constant related to the linear transformation matrix update */
  prev_datum_ave=new double[dim];
  
  W_num=new double*[dim];
  for( i=0; i<dim; i++ )
    W_num[i]=new double[dim];
  
  W_denom=new double*[dim];
  for( i=0; i<dim; i++ )
    W_denom[i]=new double[dim];
  
  for( i=0; i<dim; i++ )
    {
      prev_datum_ave[i]=0.0;
      for( n=0; n<data->num_seqs; n++ )
	for( t=1; t<data->sequence[n]->seq_length; t++ )
	  prev_datum_ave[i]+=prob_s[n][t]*data->sequence[n]->entry[t-1]->rdata[dim_index[i]];
      prev_datum_ave[i]/=sum;
    }
  
  /* Updating linear transformation matrix */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      { /* (i,j)-th entry */
	W_num[i][j]=0.0;
	for( n=0; n<data->num_seqs; n++ )
	    for( t=1; t<data->sequence[n]->seq_length; t++ )
	      W_num[i][j]+=(prob_s[n][t]*data->sequence[n]->entry[t-1]->rdata[dim_index[i]]*
			    (data->sequence[n]->entry[t-1]->rdata[dim_index[j]]-prev_datum_ave[j]));
      } /* (i,j)-th entry */
  
  find_inv( W_num, dim, W_denom );
  
  if( SINGULAR_FAIL )
    { /* New matrix is singular */
      if( !is_done )
	{ /* Stopping EM */
	  is_done=1;
	  num_failed++;
	} /* Stopping EM */
    } /* New matrix is singular */
  else
    {
      for( i=0; i<dim; i++ )
	for( j=0; j<dim; j++ )
	  { /* (i,j)-th entry */
	    W_num[i][j]=0.0;
	    for( n=0; n<data->num_seqs; n++ )
	      for( t=1; t<data->sequence[n]->seq_length; t++ )
		W_num[i][j]+=(prob_s[n][t]*data->sequence[n]->entry[t]->rdata[dim_index[i]]*
			      (data->sequence[n]->entry[t-1]->rdata[dim_index[j]]-prev_datum_ave[j]));
	  } /* (i,j)-th entry */
      
      /* Updated W is the product of W_num and W_denom */
      for( i=0; i<dim; i++ )
	for( j=0; j<dim; j++ )
	  {
	    W[i][j]=0.0;
	    for( i1=0; i1<dim; i1++ )
	      W[i][j]+=W_num[i][i1]*W_denom[i1][j];
	  }
    }
  
  
  delete prev_datum_ave;
  
  for( i=0; i<dim; i++ )
    delete W_num[i];
  delete W_num;
  
  for( i=0; i<dim; i++ )
    delete W_denom[i];
  delete W_denom;
  
  if( SINGULAR_FAIL )
    return;
  
  /* Updating translation and covariance */
  
  /* Normalization constant is the same as used with linear transformation update */
  
  /* Translation */
  for( i=0; i<dim; i++ )
    { /* i-th component */
      mu[i]=0.0;
      for( n=0; n<data->num_seqs; n++ )
	for( t=1; t<data->sequence[n]->seq_length; t++ )
	  {
	    temp1=0.0;
	    for( j=0; j<dim; j++ )
	      temp1+=W[i][j]*data->sequence[n]->entry[t-1]->rdata[dim_index[j]];
	    mu[i]+=(prob_s[n][t]*(data->sequence[n]->entry[t]->rdata[dim_index[i]]-temp1));
	  }
      
      mu[i]/=sum;
    } /* i-th component */
  
  /* Noise covariance */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      { /* (i,j)-th entry */
	/* Storing previous values */
	inv_sigma[i][j]=sigma[i][j];
	
	/* Updating */
	sigma[i][j]=0.0;
	for( n=0; n<data->num_seqs; n++ )
	  for( t=1; t<data->sequence[n]->seq_length; t++ )
	    {	      
	      temp1=mu[i]-data->sequence[n]->entry[t]->rdata[dim_index[i]];
	      for( i1=0; i1<dim; i1++ )
		temp1+=W[i][i1]*data->sequence[n]->entry[t-1]->rdata[dim_index[i1]];
	      
	      temp2=mu[j]-data->sequence[n]->entry[t]->rdata[dim_index[j]];
	      for( i2=0; i2<dim; i2++ )
		temp2+=W[j][i2]*data->sequence[n]->entry[t-1]->rdata[dim_index[i2]];
	      
	      sigma[i][j]+=(prob_s[n][t]*temp1*temp2);
	    }
	
	sigma[i][j]/=sum;
      } /* (i,j)-th entry */
  
  /* Calculating the inverse of the covariance matrix and its determinant */
  det=find_det( sigma, dim );
  if( SINGULAR_FAIL )
    { /* New matrix is singular -- replacing with the old one */
      /* !!! Need better solution !!! */
      for( i=0; i<dim; i++ )
	for( j=0; j<dim; j++ )
	  sigma[i][j]=inv_sigma[i][j];
      det=find_det( sigma, dim );
      
      if( !is_done )
	{ /* Stopping EM */
	  is_done=1;
	  num_failed++;
	} /* Stopping EM */
      
    } /* New matrix is singular -- replacing with the old one */
  
  find_inv( sigma, dim, inv_sigma );

  return;
}

void Distribution::UpdateEmissionGaussianChowLiu( Data* data, double** prob_s, double norm_const ) {

  double** MI;                  // Mutual information
  long* best_edge;              // The list of current best edges

  /* Index variable(s) */
  long i,j,n,t,i1;

  /* Temporary variable(s) */
  double temp1,best_MI;

  /* Zeroing out the mean */
  for( i=0; i<dim; i++ )
    mu[i]=0.0;
  
  /* Updating the mean */
  for( i=0; i<dim; i++ )
    { /* i-th component of the mean */
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  mu[i]+=(prob_s[n][t]*data->sequence[n]->entry[t]->rdata[dim_index[i]]);
      
      mu[i]/=norm_const;
    } /* i-th component of the mean */
  
  /* First, computing full covariance matrix */
  
  /* Zeroing out the covariance matrix */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      sigma[i][j]=0.0;
  
  /* Updating the covariance matrix */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      {
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    sigma[i][j]+=(prob_s[n][t]*
			  (mu[i]-data->sequence[n]->entry[t]->rdata[dim_index[i]])*
			  (mu[j]-data->sequence[n]->entry[t]->rdata[dim_index[j]]));
	sigma[i][j]/=norm_const;
      }
  
  
  /* Now, applying Chow-Liu tree algorithm to the covariance matrix */
  
  /* Transforming covariance matrix into a correlation matrix */
  for( i=0; i<dim; i++ )
    for( j=i+1; j<dim; j++ )
      {
	sigma[i][j]/=sqrt(sigma[i][i]*sigma[j][j]);
	sigma[j][i]=sigma[i][j];
      }
  
  /* Building the inverse covariance matrix */
  /* Zeroing out first */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      inv_sigma[i][j]=0.0;
  
  /* Starting with the identity matrix */
  for( i=0; i<dim; i++ )
    inv_sigma[i][i]=1.0;
  
  /* Allocating mutual information lower triangular matrix */
  MI=new double*[dim];
  for( i=0; i<dim; i++ )
    MI[i]=new double[dim];
  
  /* Array of best edges */
  best_edge=new long[dim-1];
  
  /* Initializing */
  
  /* Mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      {
	MI[i][j]=-0.5*log(1.0-sigma[i][j]*sigma[i][j]);
	MI[j][i]=MI[i][j];
      }
  
  /* Initializing the list of connected nodes */
  node_used[dim-1]=1;
  for( i=0; i<dim-1; i++ )
    node_used[i]=0;
  
  /* Initializing the array of best edges */
  /* Current best edge is always from the only "attached" node */
  for( i=0; i<dim-1; i++ )
    best_edge[i]=dim-1;
  
  /* Inserting dim-1 edges */
  for( num_edges=0; num_edges<dim-1; num_edges++ )
    {
      best_MI=NEG_INF;
      i1=-1;
      for( i=0; i<dim-1; i++ )
	if( !node_used[i] )
	  if( MI[i][best_edge[i]]>best_MI )
	    { /* Found a better edge */
	      i1=i;
	      best_MI=MI[i][best_edge[i]];
	    } /* Found a better edge */
      
      /* Adding the edge */
      node_used[i1]=1;
      if( i1>best_edge[i1] )
	{
	  edge[num_edges][0]=i1;
	  edge[num_edges][1]=best_edge[i1];
	}
      else
	{
	  edge[num_edges][0]=best_edge[i1];
	  edge[num_edges][1]=i1;
	}
      
      /* Storing the information about the edge */
      edge_MI[num_edges]=best_MI;
      
      /* Updating the inverse covariance matrix */
      temp1=sigma[edge[num_edges][0]][edge[num_edges][1]];
      inv_sigma[edge[num_edges][0]][edge[num_edges][0]]+=(temp1*temp1)/(1.0-temp1*temp1);
      inv_sigma[edge[num_edges][1]][edge[num_edges][1]]+=(temp1*temp1)/(1.0-temp1*temp1);
      inv_sigma[edge[num_edges][0]][edge[num_edges][1]]=-temp1/(1.0-temp1*temp1);
      inv_sigma[edge[num_edges][1]][edge[num_edges][0]]=inv_sigma[edge[num_edges][0]][edge[num_edges][1]];
      
      /* Adjusting the list of best edges to not-yet-connected nodes */
      for( i=0; i<dim-1; i++ )
	if( !node_used[i] )
	  if( MI[i][i1]>MI[i][best_edge[i]] )
	    best_edge[i]=i1;
    }
  
  /* Calculating the ranks of the edges */
  for( i=0; i<dim-1; i++ )
    edge_rank[i]=1;
  
  for( i=0; i<dim-1; i++ )
    for( j=0; j<i; j++ )
      if( edge_MI[i]<edge_MI[j] )
	edge_rank[i]++;
      else
	edge_rank[j]++;
  
  /* Deallocating mutual information */
  for( i=0; i<dim; i++ )
    delete MI[i];
  delete MI;
  
  /* Deallocating the array of best edges */
  delete best_edge;
  
  /* Adjusting the inverse covariance matrix -- dividing by appropriate deviation values */
  for( i=0; i<dim; i++ )
    inv_sigma[i][i]/=sigma[i][i];
  for( i=0; i<num_edges; i++ )
    {
      inv_sigma[edge[i][0]][edge[i][1]]/=sqrt(sigma[edge[i][0]][edge[i][0]]*sigma[edge[i][1]][edge[i][1]]);
      inv_sigma[edge[i][1]][edge[i][0]]=inv_sigma[edge[i][0]][edge[i][1]];
    }
  
  /* Calculating the new covariance matrix */
  /* !!! May be replaced by a more efficient procedure !!! */
  find_inv( inv_sigma, dim, sigma );
  /* Calculating the inverse of the covariance matrix and its determinant */
  det=find_det( sigma, dim );
  if( SINGULAR_FAIL )
    { /* New matrix is singular */
      if( !is_done )
	{ /* Stopping EM */
	  is_done=1;
	  num_failed++;
	} /* Stopping EM */
    } /* New matrix is singular */
  
  return;
}

double xlogx( double x ) {
  /* Function calculating x*ln(x) */

  /* x cannot be negative or NaN, and these values should trigger an error. */
  if( x<0.0 || isnan(x) )
    fprintf( stderr, "Warning: Unexpected value passed to x*log(x): x=%0.12le\n", x );
  
  if( x>=0.0 )
    {
      if( x<1e-012 )
	/* !!! lim x->0 x*log(x)=0 !!! */
	return( 0.0 );
      else
	return( x*log(x) );
    }  
}

