/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* em.cpp */

#include "em.h"

void HMM::EM( Data* data, Data* input_var, double precision ) {
  /* Index variable(s) */
  long i,n,t,j;

  /* Accumulator variable(s) */
  double sum;

  /* Variables */
  double old_ll_normalized; // Previous log-likelihood

  /* Previous emission variables */
  Distribution** prev_emission;

  long iteration;           // Current iteration number

  /* !!! !!! */
  if( !(type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM || type==MODEL_TYPE_MIX || type==MODEL_TYPE_NMIX) )
    {
      fprintf( stderr, "This model type is not implemented!\n" );
      exit( -1 );
    }

  prev_emission=new Distribution*[num_states];

  is_done=0;

  /* Since likelihood can be estimated after performing an E-step, doing just that. */

  /* Allocating necessary structures */
  AllocateEMStructures( data, input_var );
  
  /* E-step */
  /* Need to perform forward-backward algorithm */
  EStep( data, input_var );

  iteration=0;
  if( em_verbose )
    printf( "%ld:\tll=%0.12le\tper-dimension-ll=%0.12le\n", iteration, ll, ll_normalized );

  if( !finite( ll_normalized ) ) {
    if( em_verbose )
      {
	fprintf( stderr, "Log-likelihood is infinite\n" );
	num_failed++;
	is_done=1;
	//	WriteToFile( stderr );
      }   
  }

  while( !is_done )
    { /* Main loop */
      iteration++;
 
      /* M-step */

      /* Storing the previous value of the log-likelihood */
      old_ll_normalized=ll_normalized;

      /* Calculating necessary summaries */
      CalculateSummaries( data );
      
      /* Updating emission probabilities */
      for( i=0; i<num_states; i++ )
	{
	  prev_emission[i]=emission[i]->copy();
	  emission[i]->UpdateEmissionParameters( data, Anti[i], norm_const[i] );
	}

      /* Updating transition layer parameters */
      if( type==MODEL_TYPE_NHMM )
	/* Non-homogeneous HMM */
	input->UpdateTransitionParameters( input_var, Anti, Bntij, log_un_prob, iteration );
      else if( type==MODEL_TYPE_NMIX )
	/* Non-homogeneous mixture */
	input->UpdateTransitionParameters( input_var, Anti, NULL, log_un_prob, iteration );
      else if( type==MODEL_TYPE_HMM )
	{ /* Ordinary HMM */
	  if( !robust_first_state )
	    {
	      /* P(S_1) */
	      sum=0.0;
	      for( i=0; i<num_states; i++ )
		{
		  first[i]=0.0;
		  for( n=0; n<data->num_seqs; n++ )
		    first[i]+=Anti[i][n][0];
		  sum+=first[i];
		}
	      
	      /* Normalizing */
	      for( i=0; i<num_states; i++ )
		first[i]/=sum;
	    }
	  else
	    {
	      /* P(S_1) */
	      sum=0.0;
	      for( i=0; i<num_states; i++ )
		{
		  first[i]=0.0;
		  for( n=0; n<data->num_seqs; n++ )
		    for( t=0; t<data->sequence[n]->seq_length; t++ )
		      first[i]+=Anti[i][n][t];
		  sum+=first[i];
		}
	      
	      /* Normalizing */
	      for( i=0; i<num_states; i++ )
		first[i]/=sum;
	    }

	  /* Transition matrix */
	  for( j=0; j<num_states; j++ )
	    { /* Summing over S_{t-1} */
	      sum=0.0;
	      for( i=0; i<num_states; i++ )
		{ /* Summing over S_t */
		  trans[j][i]=0.0;
		  
		  for( n=0; n<data->num_seqs; n++ )
		    for( t=1; t<data->sequence[n]->seq_length; t++ )
		      trans[j][i]+=Bntij[i][j][n][t];

		  sum+=trans[j][i];
		} /* Summing over S_t */
    
	      /* Normalizing */
	      for( i=0; i<num_states; i++ )
		trans[j][i]/=sum;
	    } /* Summing over S_{t-1} */
	} /* Ordinary HMM */
      else if( type==MODEL_TYPE_MIX )
	{ /* Mixture of states */
	  for( i=0; i<num_states; i++ )
	    {
	      first[i]=0.0;
	      for( n=0; n<data->num_seqs; n++ )
		for( t=0; t<data->sequence[n]->seq_length; t++ )
		  first[i]+=Anti[i][n][t];
	    }
	  
	  sum=0.0;
	  for( i=0; i<num_states; i++ )
	    sum+=first[i];
	  
	  for( i=0; i<num_states; i++ )
	    first[i]/=sum;
	} /* Mixture of states */

      /* E-step */
      /* Need to perform forward-backward algorithm */
      EStep( data, input_var );

      if( em_verbose )
	printf( "%ld:\tll=%0.12le\tper-dimension ll=%0.12le\n", iteration, ll, ll_normalized );
	
      if( !finite( ll_normalized ) )
	{ /* Log-likelihood is infinite -- run failed */
	  num_failed++;
	  is_done=1;
	} /* Log-likelihood is infinite -- run failed */
      else if( ll_normalized-old_ll_normalized<precision )
	{ /* Checking whether the end condition is reached */
	  if( ll_normalized>=old_ll_normalized )
	    is_done=1;
	  else
	    {
	      if( em_verbose )
		printf( "Log-likelihood went down\n" );

	      /* Analyzing why log-likelihood went down */
	      /* Most likely, one of the emissions changed to a worse set of parameters */
	      for( i=0; i<num_states; i++ )
		if( prev_emission[i]->weighted_ll( data, Anti[i] )>
		    emission[i]->weighted_ll( data, Anti[i] ) )
		  { /* Emission i got worse */
		    delete emission[i];
		    emission[i]=prev_emission[i]->copy();
		  } /* Emissiong i got worse */

	      /* Redoing the E-step */
	      EStep( data, input_var );

	      if( em_verbose )
		printf( "Corrected ll=%0.12le\tper-dimension ll=%0.12le\n", ll, ll_normalized );

	      if( !finite( ll_normalized ) )
		{ /* Log-likelihood is infinite -- run failed */
		  num_failed++;
		  is_done=1;
		} /* Log-likelihood is infinite -- run failed */
	      else if( ll_normalized-old_ll_normalized<precision )
		{ /* Checking whether the end condition is reached */
		  if( ll_normalized>=old_ll_normalized )
		    is_done=1;
		  else
		    {
		      if( em_verbose )
			printf( "Corrected log-likelihood still went down\n" );
		      is_done=1;
		    }
		} /* Checking whether the end condition is reached */
	    }
	} /* Checking whether the end condition is reached */
	  
      for( i=0; i<num_states; i++ )
	delete prev_emission[i];
    } /* Main loop */
  
  /* Deallocating arrays */
  DeallocateEMStructures( data, input_var );

  delete prev_emission;

  return;
}

void HMM::EMGibbs( Data* data, Data* input_data, double precision ) {
  /* Index variable(s) */
  long i,n,t,j;

  /* Temporary variable(s) */
  long* temp_state;
  long* temp_entry;

  /* Accumulator variable(s) */
  double sum;

  /* Variables */
  double old_ll_normalized; // Previous log-likelihood

  /* Previous emission variables */
  Distribution** prev_emission;

  /* Variables holding accumulated simulation values */
  long*** sim_state;
  long*** sim_entry;
  long num_samples;         // Number of samples from the distribution

  long iteration;           // Current iteration number

  num_samples=GIBBS_SAMPLES_PER_STATE*num_states;
  sim_state=new long**[data->num_seqs];
  sim_entry=new long**[data->num_seqs];
  for( n=0; n<data->num_seqs; n++ )
    {
      sim_state[n]=new long*[num_samples];
      sim_entry[n]=new long*[num_samples];
    }

  if( em_verbose )
    printf( "EM-MCMC\n" );

  /* !!! !!! */
  if( !(type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM || type==MODEL_TYPE_MIX || type==MODEL_TYPE_NMIX) )
    {
      fprintf( stderr, "This model type is not implemented!\n" );
      exit( -1 );
    }

  prev_emission=new Distribution*[num_states];

  is_done=0;

  /* Since likelihood can be estimated after performing an E-step, doing just that. */

  /* Allocating necessary structures */
  AllocateEMStructures( data, input_data );
  
  /* E-step */
  /* Need to perform forward-backward algorithm */
  EStep( data, input_data );

  iteration=0;
  /* !!! Assuming initializations can handle missing data !!! */
  if( em_verbose )
    printf( "%ld:\tll=%0.12le\tper-dimension-ll=%0.12le\n", iteration, ll, ll_normalized );

  if( !finite( ll_normalized ) ) {
    if( em_verbose )
      {
	fprintf( stderr, "Log-likelihood is infinite\n" );
	num_failed++;
	is_done=1;
      }   
  }
  
  is_done=1;

  while( !is_done )
    { /* Main loop */
      iteration++;
 
      /* M-step */

      /* Storing the previous value of the log-likelihood */
      old_ll_normalized=ll_normalized;

      /* Initializing missing entries */
      for( n=0; n<data->num_seqs; n++ )
	InitializeMissingEntriesSequence( data->sequence[n] );

      /* Burn-in */
      for( j=0; j<GIBBS_BURNIN; j++ )
	{
	  
	}

      /* Calculating necessary summaries */
      CalculateSummaries( data );
      
      /* Updating emission probabilities */
      for( i=0; i<num_states; i++ )
	{
	  prev_emission[i]=emission[i]->copy();
	  emission[i]->UpdateEmissionParameters( data, Anti[i], norm_const[i] );
	}

      /* Updating transition layer parameters */
      if( type==MODEL_TYPE_NHMM )
	/* Non-homogeneous HMM */
	input->UpdateTransitionParameters( input_data, Anti, Bntij, log_un_prob, iteration );
      else if( type==MODEL_TYPE_NMIX )
	/* Non-homogeneous mixture */
	input->UpdateTransitionParameters( input_data, Anti, NULL, log_un_prob, iteration );
      else if( type==MODEL_TYPE_HMM )
	{ /* Ordinary HMM */
	  if( !robust_first_state )
	    {
	      /* P(S_1) */
	      sum=0.0;
	      for( i=0; i<num_states; i++ )
		{
		  first[i]=0.0;
		  for( n=0; n<data->num_seqs; n++ )
		    first[i]+=Anti[i][n][0];
		  sum+=first[i];
		}
	      
	      /* Normalizing */
	      for( i=0; i<num_states; i++ )
		first[i]/=sum;
	    }
	  else
	    {
	      /* P(S_1) */
	      sum=0.0;
	      for( i=0; i<num_states; i++ )
		{
		  first[i]=0.0;
		  for( n=0; n<data->num_seqs; n++ )
		    for( t=0; t<data->sequence[n]->seq_length; t++ )
		      first[i]+=Anti[i][n][t];
		  sum+=first[i];
		}
	      
	      /* Normalizing */
	      for( i=0; i<num_states; i++ )
		first[i]/=sum;
	    }

	  /* Transition matrix */
	  for( j=0; j<num_states; j++ )
	    { /* Summing over S_{t-1} */
	      sum=0.0;
	      for( i=0; i<num_states; i++ )
		{ /* Summing over S_t */
		  trans[j][i]=0.0;
		  
		  for( n=0; n<data->num_seqs; n++ )
		    for( t=1; t<data->sequence[n]->seq_length; t++ )
		      trans[j][i]+=Bntij[i][j][n][t];

		  sum+=trans[j][i];
		} /* Summing over S_t */
    
	      /* Normalizing */
	      for( i=0; i<num_states; i++ )
		trans[j][i]/=sum;
	    } /* Summing over S_{t-1} */
	} /* Ordinary HMM */
      else if( type==MODEL_TYPE_MIX )
	{ /* Mixture of states */
	  for( i=0; i<num_states; i++ )
	    {
	      first[i]=0.0;
	      for( n=0; n<data->num_seqs; n++ )
		for( t=0; t<data->sequence[n]->seq_length; t++ )
		  first[i]+=Anti[i][n][t];
	    }
	  
	  sum=0.0;
	  for( i=0; i<num_states; i++ )
	    sum+=first[i];
	  
	  for( i=0; i<num_states; i++ )
	    first[i]/=sum;
	} /* Mixture of states */

      /* E-step */
      /* Need to perform forward-backward algorithm */
      EStep( data, input_data );

      if( em_verbose )
	printf( "%ld:\tLower bound on ll=%0.12le\tper-dimension ll=%0.12le\n", iteration, ll, ll_normalized );
	
      if( !finite( ll_normalized ) )
	{ /* Log-likelihood is infinite -- run failed */
	  num_failed++;
	  is_done=1;
	} /* Log-likelihood is infinite -- run failed */
      else if( ll_normalized-old_ll_normalized<precision )
	{ /* Checking whether the end condition is reached */
	  if( ll_normalized>=old_ll_normalized )
	    is_done=1;
	  else
	    {
	      if( em_verbose )
		printf( "Log-likelihood went down\n" );

	      /* Analyzing why log-likelihood went down */
	      /* Most likely, one of the emissions changed to a worse set of parameters */
	      for( i=0; i<num_states; i++ )
		if( prev_emission[i]->weighted_ll( data, Anti[i] )>
		    emission[i]->weighted_ll( data, Anti[i] ) )
		  { /* Emission i got worse */
		    delete emission[i];
		    emission[i]=prev_emission[i]->copy();
		  } /* Emissiong i got worse */

	      /* Redoing the E-step */
	      EStep( data, input_data );

	      if( em_verbose )
		printf( "Corrected lower bound on ll=%0.12le\tper-dimension ll=%0.12le\n", ll, ll_normalized );

	      if( !finite( ll_normalized ) )
		{ /* Log-likelihood is infinite -- run failed */
		  num_failed++;
		  is_done=1;
		} /* Log-likelihood is infinite -- run failed */
	      else if( ll_normalized-old_ll_normalized<precision )
		{ /* Checking whether the end condition is reached */
		  if( ll_normalized>=old_ll_normalized )
		    is_done=1;
		  else
		    {
		      if( em_verbose )
			printf( "Corrected lower-bound on log-likelihood still went down\n" );
		      is_done=1;
		    }
		} /* Checking whether the end condition is reached */
	    }
	} /* Checking whether the end condition is reached */
	  
      for( i=0; i<num_states; i++ )
	delete prev_emission[i];
    } /* Main loop */
  
  /* Deallocating arrays */
  DeallocateEMStructures( data, input_data );

  delete prev_emission;

  /* Deallocating the simulation arrays */
  for( n=0; n<data->num_seqs; n++ )
    {
      delete sim_state[n];
      delete sim_entry[n];
    }
  delete sim_state;
  delete sim_entry;

  return;
}

void HMM::AllocateEMStructures( Data* data, Data* input_vars ) {
  /* Allocating structures needed for forward-backward algorithm */

  /* Index variables */
  long i,j,n;

  AllocateAlpha( data, input_vars );

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {
      /* Allocating betas */
      for( i=0; i<num_states; i++ )
	{ /* For each state */
	  beta[i]=new double*[data->num_seqs];
	  for( n=0; n<data->num_seqs; n++ )
	    /* For each sequence */
	    beta[i][n]=new double[data->sequence[n]->seq_length];
	} /* For each state */

      /* Bntij */
      for( i=0; i<num_states; i++ )
	for( j=0; j<num_states; j++ )
	  {
	    Bntij[i][j]=new double*[data->num_seqs];
	    for( n=0; n<data->num_seqs; n++ )
	      Bntij[i][j][n]=new double[data->sequence[n]->seq_length];
	  }
    }

  return;
}

void HMM::AllocateAlpha( Data* data, Data* input_vars ) {
  /* Allocating structures needed to calculate scaled alphas for forward-backward algorithm */

  /* Index variable(s) */
  long i,n,j;

  /* Allocation block */

  /* Allocating array of P(R_nt|S_nt=i) (or P(R_nt,X_nt|S_nt=i) ) */  
  for( i=0; i<num_states; i++ )
    { /* For each state */
      pb[i]=new double*[data->num_seqs];
      for( n=0; n<data->num_seqs; n++ )
	/* For each sequence */
	pb[i][n]=new double[data->sequence[n]->seq_length];
    } /* For each state */
  
  /* Allocating array of P(S_nt=i|R_nt,X_nt) */
  for( i=0; i<num_states; i++ )
    { /* For each state */
      Anti[i]=new double*[data->num_seqs];
      for( n=0; n<data->num_seqs; n++ )
	/* For each sequence */
	Anti[i][n]=new double[data->sequence[n]->seq_length];
    } /* For each state */

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {      
      /* Allocating alpha's, and scaling factors c's */
      for( i=0; i<num_states; i++ )
	{ /* For each state */
	  alpha[i]=new double*[data->num_seqs];
	  for( n=0; n<data->num_seqs; n++ )
	    alpha[i][n]=new double[data->sequence[n]->seq_length];
	} /* For each state */
      
      /* c's are independent of the state */
      c=new double*[data->num_seqs];
      for( n=0; n<data->num_seqs; n++ )
	/* For each sequence */
	c[n]=new double[data->sequence[n]->seq_length];
    }
      
  if( type==MODEL_TYPE_NHMM || type==MODEL_TYPE_NMIX )
    { /* Alocating structure necessary for NHMMs */
      /* Allocating the array of probabilities P(S_nt|S_n,t-1,X_nt) */
      for( i=0; i<num_states; i++ )
	if( type==MODEL_TYPE_NHMM )
	  for( j=0; j<num_states; j++ )
	    {
	      p_tr[i][j]=new double*[input_vars->num_seqs];
	      for( n=0; n<input_vars->num_seqs; n++ )
		/* For each sequence */
		p_tr[i][j][n]=new double[input_vars->sequence[n]->seq_length];
	    }
	else if( type==MODEL_TYPE_NMIX )
	  {
	    p_tr[i][0]=new double*[input_vars->num_seqs];
	    for( n=0; n<input_vars->num_seqs; n++ )
	      p_tr[i][0][n]=new double[input_vars->sequence[n]->seq_length];
	  }

      /* Allocating the array of log-unnormalized P(S_nt|S_n,t-1,X_nt) */
      for( i=1; i<num_states; i++ )
	if( type==MODEL_TYPE_NHMM )
	  for( j=0; j<num_states; j++ )
	    {
	      log_un_prob[i][j]=new double*[input_vars->num_seqs];
	      for( n=0; n<input_vars->num_seqs; n++ )
		/* For each sequence */
		log_un_prob[i][j][n]=new double[input_vars->sequence[n]->seq_length];
	    }
	else if( type==MODEL_TYPE_NMIX )
	  {
	    log_un_prob[i][0]=new double*[input_vars->num_seqs];
	    for( n=0; n<input_vars->num_seqs; n++ )
	      log_un_prob[i][0][n]=new double[input_vars->sequence[n]->seq_length];
	  }

    } /* Alocating structure necessary for NHMMs */

  return;
}

void HMM::EStep( Data* data, Data* input_var ) {

  /* Index variable(s) */
  long n,i;

  ll=0.0;
  for( n=0; n<data->num_seqs; n++ )
    switch( type ) {
    case MODEL_TYPE_HMM:
      ll+=log_likelihood_sequence( data->sequence[n], NULL, n, 0 );
      CalculateBetaSequence( data->sequence[n], n );
      break;
    case MODEL_TYPE_NHMM:
      ll+=log_likelihood_sequence( data->sequence[n], input_var->sequence[n], n, 0 );
      CalculateBetaSequence( data->sequence[n], n );
      break;
    case MODEL_TYPE_MIX:
      ll+=log_likelihood_sequence( data->sequence[n], NULL, n, 0 );
      break;
    case MODEL_TYPE_NMIX:
      ll+=log_likelihood_sequence( data->sequence[n], input_var->sequence[n], n, 0 );
      break;
    default:;
    }

  ll_normalized=ll/data->total_entries();

  return;
}

double HMM::log_likelihood_sequence( Sequence* data, Sequence* input_data, long n, long start_index ) {
  double ll=0.0;

  /* Index variable(s) */
  long i,j,t;

  /* Temporary variable(s) */
  double sum;

  /*** Calculating needed values (forward) ***/

  /* Calculating emission probabilities */
  CalculateProbEmissionsSequence( data, n, start_index );

  if( type==MODEL_TYPE_NHMM || type==MODEL_TYPE_NMIX )
    /* Calculating P(S_nt|S_n,t-1, X_nt) */
    CalculateProbSSequence( input_data, n, start_index );

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    /* Need to perform forward pass -- calculating scaled alpha values */
    CalculateAlphaSequence( data, n, start_index );
   
  /* Calculating the log-likelihood using scaled alphas */
  for( t=0; t<data->seq_length; t++ )
    switch( type ) {
    case MODEL_TYPE_HMM:
    case MODEL_TYPE_NHMM:
      ll-=log(c[n][t]);
      break;
    case MODEL_TYPE_MIX:
      sum=0.0;
      for( i=0; i<num_states; i++ )
	sum+=first[i]*pb[i][n][t];
      ll+=log(sum);
      break;
    case MODEL_TYPE_NMIX:
      sum=0.0;
      for( i=0; i<num_states; i++ )
	sum+=p_tr[i][0][n][t]*pb[i][n][t];
      ll+=log(sum);
      break;
    default:;
    }

  return( ll );
}

double HMM::complete_log_likelihood_sequence( Sequence* data, Sequence* input_data, long* state, long index ) {
  double ll=0.0;

  /* Index variable(s) */
  long t;

  /* !!! Assuming transition and emission probabilities are computed. !!! */
  
  /* t=0 */
  t=0;
  ll+=log(pb[state[t]][index][t]);
  switch( type ) {
  case MODEL_TYPE_MIX:
  case MODEL_TYPE_HMM:
    ll+=first[state[t]];
    break;
  case MODEL_TYPE_NMIX:
  case MODEL_TYPE_NHMM:
    ll+=p_tr[state[t]][0][index][t];
    break;
  default:;
  }
  
  for( t=0; t<data->seq_length; t++ )
    {
      ll+=log(pb[state[t]][index][t]);
      switch( type ) {
      case MODEL_TYPE_MIX:
	ll+=first[state[t]];
	break;
      case MODEL_TYPE_NMIX:
	ll+=p_tr[state[t]][0][index][t];
	break;
      case MODEL_TYPE_HMM:
	ll+=trans[state[t-1]][state[t]];
	break;
      case MODEL_TYPE_NHMM:
	ll+=p_tr[state[t]][state[t-1]][index][t];
	break;
      default:;
      }
    }

  return( ll );
}

void HMM::CalculateProbEmissionsSequence( Sequence* data, long n, long start_index ) {
  /* Index variable(s) */
  long i,t;

  /* Initializing array of P(R_nt|S_nt=i) */
  for( t=start_index; t<data->seq_length; t++ )
    if( t==0 )
      for( i=0; i<num_states; i++ )	
	pb[i][n][t]=emission[i]->prob(data->entry[t], NULL);
    else
      for( i=0; i<num_states; i++ )
	pb[i][n][t]=emission[i]->prob(data->entry[t], data->entry[t-1]);
}

void HMM::CalculateProbSSequence( Sequence* input_var, long n, long start_index ) {
  /* Index variable(s) */
  long i,j,t;

  /* Temporary variable(s) */
  double sum;
  double max;

  /* Calculating P(S_t|S_{t-1}, X_t ) */

  /* !!! Hack speeding up the calculation of transition probabilities !!! */
  if( start_index==0 )
    {
      /* Calculating P(S_n1|X_n1) */
      max=0.0;
      for( i=1; i<num_states; i++ )
	{
	  log_un_prob[i][0][n][0]=	    
	    input->log_un_probLogistic( input_var->entry[0], i, -1 );
	  if( max<log_un_prob[i][0][n][0] )
	    max=log_un_prob[i][0][n][0];
	}

      sum=exp(-max);
      for( i=1; i<num_states; i++ )
	sum+=exp(log_un_prob[i][0][n][0]-max);	    
      
      p_tr[0][0][n][0]=exp(-max)/sum;
      for( i=1; i<num_states; i++ )
	p_tr[i][0][n][0]=exp(log_un_prob[i][0][n][0]-max)/sum;

      /* Updating start_index */
      start_index=1;
    }

  /* Calculating P(S_{nt}|S_{n,t-1},X_{nt}) */
  for( t=start_index; t<input_var->seq_length; t++ )
    if( input->type==DIST_LOGISTIC )
      {
	max=0.0;
	for( i=1; i<num_states; i++ )
	  {
	    log_un_prob[i][0][n][t]=
	      input->log_un_probLogistic( input_var->entry[t], i, -1 );
	    if( max<log_un_prob[i][0][n][t] )
	      max=log_un_prob[i][0][n][t];
	  }
	   
	sum=exp(-max);
	for( i=1; i<num_states; i++ )
	  sum+=exp(log_un_prob[i][0][n][t]-max);
	
	p_tr[0][0][n][t]=exp(-max)/sum;
	for( i=1; i<num_states; i++ )
	  p_tr[i][0][n][t]=exp(log_un_prob[i][0][n][t]-max)/sum;
      }
    else if( input->type==DIST_TRANSLOGISTIC )
      for( j=0; j<num_states; j++ )
	{
	  max=0.0;
	  for( i=1; i<num_states; i++ )
	    {
	      log_un_prob[i][j][n][t]=
		  input->log_un_probLogistic( input_var->entry[t], i, j );
	      if( max<log_un_prob[i][j][n][t] )
		max=log_un_prob[i][j][n][t];
	    }

	  sum=exp(-max);
	  for( i=1; i<num_states; i++ )
	    sum+=exp(log_un_prob[i][j][n][t]-max);
	  
	  p_tr[0][j][n][t]=exp(-max)/sum;
	  for( i=1; i<num_states; i++ )
	    p_tr[i][j][n][t]=exp(log_un_prob[i][j][n][t]-max)/sum;
	}
  return;
}

void HMM::CalculateAlphaSequence( Sequence* data, long n, long start_index ) {
  /* Calculating necessary values for the forward-backward algorithm */

  /* Index variables */
  long i,t,j;
  
  double sum;         // Temporary accumulator

  /* Implementing forward-backward algorithm with scaling */
  
  /* Calculating scaled alphas */

  if( start_index==0 )
    {
      /* First alpha */
      sum=0.0;
      for( i=0; i<num_states; i++ )
	{ 
	  if( type==MODEL_TYPE_NHMM )
	    alpha[i][n][0]=p_tr[i][0][n][0]*pb[i][n][0];
	  else if( type==MODEL_TYPE_HMM )
	    alpha[i][n][0]=first[i]*pb[i][n][0];
	  sum+=alpha[i][n][0];
	}
      
      /* Calculating first c */
      c[n][0]=1.0/sum;
      
      /* Rescaling first alpha */
      for( i=0; i<num_states; i++ )
	alpha[i][n][0]*=c[n][0];

      /* Updating start_index */
      start_index=1;
    }

  /* Calculating the rest of alphas */
  for( t=start_index; t<data->seq_length; t++ )
    {
      sum=0.0;
      for( i=0; i<num_states; i++ )
	{ /* Summing over S_t */
	  alpha[i][n][t]=0.0;
	  for( j=0; j<num_states; j++ )
	    /* Summing over S_{t-1} */
	    if( type==MODEL_TYPE_NHMM )
	      alpha[i][n][t]+=(alpha[j][n][t-1]*p_tr[i][j][n][t]);
	    else if( type==MODEL_TYPE_HMM )
	      alpha[i][n][t]+=(alpha[j][n][t-1]*trans[j][i]);
	  
	  alpha[i][n][t]*=pb[i][n][t];
	  sum+=alpha[i][n][t];
	} /* Summing over S_t */
      
      c[n][t]=1.0/sum;

      /* Rescaling alphas */
      for( i=0; i<num_states; i++ )
	alpha[i][n][t]*=c[n][t];
    }
  
  return;
}

void HMM::CalculateBetaSequence( Sequence* data, long n ) {
  /* Calculating necessary values for the forward-backward algorithm */

  /* Index variables */
  long i,t,j;
  
  /* Calculating scaled betas */

  /* Initializing beta_T */
  for( i=0; i<num_states; i++ )
    beta[i][n][data->seq_length-1]=c[n][data->seq_length-1];
      
  /* Calculating remaining betas */
  for( t=data->seq_length-2; t>=0; t-- )
    {
      for( i=0; i<num_states; i++ )
	{ /* Summing over S_t */
	  beta[i][n][t]=0.0;
	  for( j=0; j<num_states; j++ )
	    /* Summing over S_{t+1} */
	    if( type==MODEL_TYPE_NHMM )
	      beta[i][n][t]+=(beta[j][n][t+1]*p_tr[j][i][n][t+1]*pb[j][n][t+1]);
	    else if( type==MODEL_TYPE_HMM )
	      beta[i][n][t]+=(beta[j][n][t+1]*trans[i][j]*pb[j][n][t+1]);
	} /* Summing over S_t */
	  
      /* Rescaling betas */
      for( i=0; i<num_states; i++ )
	beta[i][n][t]*=c[n][t];
    }

  return;
}

void HMM::CalculateSummaries( Data* data ) {
  /* Index variable(s) */
  long i,j,n,t;

  /* Temporary variable(s) */
  double sum;

  /* Calculating P(S_nt=i|R_n,X_n) */
  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    /* Markov dependence */
    for( i=0; i<num_states; i++ )
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  Anti[i][n][t]=alpha[i][n][t]*beta[i][n][t]/c[n][t];

  if( type==MODEL_TYPE_MIX || type==MODEL_TYPE_NMIX )
    { /* Mixture model */
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  {      
	    sum=0.0;
	    for( i=0; i<num_states; i++ )
	      {
		if( type==MODEL_TYPE_MIX )
		  Anti[i][n][t]=first[i]*pb[i][n][t];
		else if( type==MODEL_TYPE_NMIX )
		  Anti[i][n][t]=p_tr[i][0][n][t]*pb[i][n][t];
		sum+=Anti[i][n][t];
	      }
	    
	    /* Normalizing */
	    for( i=0; i<num_states; i++ )
	      Anti[i][n][t]/=sum;
	  }
    } /* Mixture model */

  /* Calculating normalization constant for the outputs */
  for( i=0; i<num_states; i++ )
    {
      norm_const[i]=0.0;
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  norm_const[i]+=Anti[i][n][t];
    }

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {
      /* Bntij */
      for( i=0; i<num_states; i++ )
	for( j=0; j<num_states; j++ )
	  for( n=0; n<data->num_seqs; n++ )
	    for( t=1; t<data->sequence[n]->seq_length; t++ )
	      if( type==MODEL_TYPE_NHMM )
		Bntij[i][j][n][t]=alpha[j][n][t-1]*beta[i][n][t]*pb[i][n][t]*p_tr[i][j][n][t];
	      else if( type==MODEL_TYPE_HMM )
		Bntij[i][j][n][t]=alpha[j][n][t-1]*beta[i][n][t]*pb[i][n][t]*trans[j][i];	
    }

  return;
}

void HMM::DeallocateEMStructures( Data* data, Data* input_vars ) {
  /* Deallocating structures needed for forward-backward algorithm */

  /* Index variables */
  long i,j,n;

  DeallocateAlpha( data, input_vars );

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {
      /* Beta */
      for( i=0; i<num_states; i++ )
	{ /* For each state */
	  for( n=0; n<data->num_seqs; n++ )
	    /* For each sequence */
	    delete beta[i][n];
	  delete beta[i];
	} /* For each state */

      /* Bntij */
      for( i=0; i<num_states; i++ )
	for( j=0; j<num_states; j++ )
	  {
	    for( n=0; n<data->num_seqs; n++ )
	      delete Bntij[i][j][n];
	    delete Bntij[i][j];
	  }
    }
  
  return;
}

void HMM::DeallocateAlpha( Data* data, Data* input_vars ) {
  /* Deallocating arrays needed to calculate scaled alphas in forward-backward algorithm */
  
  /* Index variable(s) */
  long i,n,j;

  /* pb */
  for( i=0; i<num_states; i++ )
    { /* For each state */
      for( n=0; n<data->num_seqs; n++ )
	delete pb[i][n];
      delete pb[i];
    } /* For each state */

  /* Anti */
  for( i=0; i<num_states; i++ )
    { /* For each state */
      for( n=0; n<data->num_seqs; n++ )
	delete Anti[i][n];
      delete Anti[i];
    } /* For each state */

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {
      /* Alpha and c */
      for( i=0; i<num_states; i++ )
	{ /* For each state */
	  for( n=0; n<data->num_seqs; n++ )
	    delete alpha[i][n];
	  delete alpha[i];
	} /* For each state */
      
      for( n=0; n<data->num_seqs; n++ )
	delete c[n];
      delete c;
    }

  if( type==MODEL_TYPE_NHMM || type==MODEL_TYPE_NMIX )
    { /* Deallocating probability arrays */

      /* Normalized transition probability */
      for( i=0; i<num_states; i++ )
	if( type==MODEL_TYPE_NHMM )
	  for( j=0; j<num_states; j++ )
	    {
	      for( n=0; n<input_vars->num_seqs; n++ )
		delete p_tr[i][j][n];
	      delete p_tr[i][j];
	    }
	else if( type==MODEL_TYPE_NMIX )
	  {
	    for( n=0; n<input_vars->num_seqs; n++ )
	      delete p_tr[i][0][n];
	    delete p_tr[i][0];
	  }

      /* Log-unnormalized transition probability */
      for( i=1; i<num_states; i++ )
	if( type==MODEL_TYPE_NHMM )
	  for( j=0; j<num_states; j++ )
	    {
	      for( n=0; n<input_vars->num_seqs; n++ )
		delete log_un_prob[i][j][n];
	      delete log_un_prob[i][j];
	    }      
	else
	  {
	    for( n=0; n<input_vars->num_seqs; n++ )
	      delete log_un_prob[i][0][n];
	    delete log_un_prob[i][0];
	  }

    } /* Deallocating probability arrays */

  return;
}
