/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* eval.cpp */

/* Contains routines used in evaluation */

#include "struct.h"

double* Data::mean() {
  /* Calculates percentage of rain occurences for each station */

  /* !!! Searches only binary entries !!! */
  /* !!! Does not handle missing data !!! */

  long dim;
  double* output;
  long* num_hits;         /* Number of positive occurences per station */
  long* num_total;        /* Total number per station */

  /* Index variable(s) */
  long i,j,k;
  
  /* Initializing hits and total arrays */
  /* !!! Assumes same dimensionality for all entries !!! */
  dim=sequence[0]->entry[0]->ddim;
  num_hits=new long[dim];
  for( i=0; i<dim; i++ )
    num_hits[i]=0;

  num_total=new long[dim];
  for( i=0; i<dim; i++ )
    num_total[i]=0;

  for( i=0; i<num_seqs; i++ )
    for( j=0; j<sequence[i]->seq_length; j++ )
      for( k=0; k<dim; k++ )
	if( !is_missing( sequence[i]->entry[j]->ddata[k] ) )
	  {
	    num_total[k]++;
	    if( sequence[i]->entry[j]->ddata[k]==1 )
	      num_hits[k]++;
	  }

  /* Creating the array of outputs */
  output=new double[dim];
  for( i=0; i<dim; i++ )
    output[i]=(double)num_hits[i]/(double)num_total[i];

  /* Deallocating hits and total arrays */
  delete num_hits;
  delete num_total;

  return( output );
}

double** Data::uncov() {
  /* Calculating unnormalized covariance */
  /* !!! Searches only binary entries !!! */
  /* !!! Does not handle missing data !!! */
  double* mu;
  double ** cov;
  long dim;

  /* Index variable(s) */
  long i,j,k,l;

  /* First, need to know the mean */
  mu=mean();

  /* Initializing covariance matrix */
  /* !!! Assumes same dimensionality for all entries !!! */
  dim=sequence[0]->entry[0]->ddim;
  cov=new double*[dim];
  for( k=0; k<dim; k++ )
    cov[k]=new double[dim];
  
  for( k=0; k<dim; k++ )
    for( l=0; l<dim; l++ )
      cov[k][l]=0.0;
  
  for( i=0; i<num_seqs; i++ )
    for( j=0; j<sequence[i]->seq_length; j++ )
      for( k=0; k<dim; k++ )
	for( l=0; l<dim; l++ )
	  if( !is_missing( sequence[i]->entry[j]->ddata[k] ) && 
	      !is_missing( sequence[i]->entry[j]->ddata[l] ) )
	    /* !!! If no missing data !!! */
	    cov[k][l]+=((double)sequence[i]->entry[j]->ddata[k]-mu[k])*
	      ((double)sequence[i]->entry[j]->ddata[l]-mu[l]);

  delete mu;

  return( cov );
}

double** Data::persistence() {
  /* Calculating the persistence of values for each component */
  
  long dim;
  long** total;
  long** pers;
  double** pers_prob;

  /* Index variable(s) */
  long i,j,k;

  /* Allocating and initializing arrays */
  dim=sequence[0]->entry[0]->ddim;
  total=new long*[dim];
  for( i=0; i<dim; i++ )
    total[i]=new long[2]; /* !!! !!! */

  pers=new long*[dim];
  for( i=0; i<dim; i++ )
    pers[i]=new long[2]; /* !!! !!! */
  
  for( i=0; i<dim; i++ )
    for( j=0; j<2 /* !!! !!! */; j++ )
      {
	total[i][j]=0;
	pers[i][j]=0;
      }

  /* Calculating counts */
  for( i=0; i<num_seqs; i++ )
    for( j=1; j<sequence[i]->seq_length; j++ )
      for( k=0; k<dim; k++ )
	if( sequence[i]->entry[j-1]->ddata[k]==0 )
	  {
	    total[k][0]++;
	    if( sequence[i]->entry[j]->ddata[k]==0 )
	      pers[k][0]++;
	  }
	else if( sequence[i]->entry[j-1]->ddata[k]==1 )
	  {
	    total[k][1]++;
	    if( sequence[i]->entry[j]->ddata[k]==1 )
	      pers[k][1]++;
	  }

  /* Allocating persistence probability array */
  pers_prob=new double*[dim];
  for( i=0; i<dim; i++ )
    pers_prob[i]=new double[2]; /* !!! !!! */

  /* Calculating persistence probabilities */
  for( i=0; i<dim; i++ )
    for( j=0; j<2 /* !!! !!! */; j++ )
      pers_prob[i][j]=(double)pers[i][j]/(double)total[i][j];

  /* Deallocating counts */
  for( i=0; i<dim; i++ )
    delete total[i];
  delete total;

  for( i=0; i<dim; i++ )
    delete pers[i];
  delete pers;

  return( pers_prob );
}

long** Data::spell( long outcome ) {

  long dim;
  long** length_dist;
  long spell_length;

  /* Index variable(s) */
  long i,j,k;

  /* Allocating and initializing the output array */
  dim=sequence[0]->entry[0]->ddim;
  
  /* First dimension is the components, second is the length of the spell */
  /* !!! Second component max size is the length of the first sequence !!! */
  length_dist=new long*[dim];
  for( i=0; i<dim; i++ )
    length_dist[i]=new long[sequence[0]->seq_length];

  for( i=0; i<dim; i++ )
    for( j=0; j<sequence[0]->seq_length; j++ )
      length_dist[i][j]=0;

  for( i=0; i<dim; i++ )
    { /* For each component (station) */
      spell_length=0;

      for( j=0; j<num_seqs; j++ )
	{ /* For each sequence */
	  for( k=0; k<sequence[j]->seq_length; k++ )
	    /* For each entry in the sequence */
	    if( sequence[j]->entry[k]->ddata[i]==outcome )
	      /* Spell continues */
	      spell_length++;
	    else if( spell_length>0 )
	      { /* Spell ended */

		/* Updating the counts */
		length_dist[i][spell_length-1]++;

		/* Resetting the length of the current spell */
		spell_length=0;
	      } /* Spell ended */
	  
	  /* Checking whether the sequence has ended in a spell */
	  if( spell_length>0 )
	    /* Updating the counts */
	    length_dist[i][spell_length-1]++;

	  /* Resetting the length */
	  spell_length=0;

	} /* For each sequence */

    } /* For each component (station) */
  
  return( length_dist );
}

double** Data::mutual_information() {
  /* Calculating mutual information for each pair of dimensions */
  double**** counts;
  double** MI;
  long dim;

  /* Index variable(s) */
  long i,j,k,l;

  /* Temporary variable(s) */
  double sum;

  dim=sequence[0]->entry[0]->ddim;

  /* Allocating the arrays of counts */
  counts=new double***[dim];
  for( i=0; i<dim; i++ )
    {
      counts[i]=new double**[i+1];
      for( j=0; j<i+1; j++ )
	{
	  counts[i][j]=new double*[2]; /* !!! !!! */
	  for( k=0; k<2 /* !!! */; k++ )
	    counts[i][j][k]=new double[2]; /* !!! !!! */
	}	
    }

  /* Allocating the MI matrix */
  MI=new double*[dim];
  for( i=0; i<dim; i++ )
    MI[i]=new double[dim];
  
  /* Initializing the counts */
  for( i=0; i<dim; i++ )
    for( j=0; j<i+1; j++ )
      for( k=0; k<2 /* !!! !!! */; k++ )
	for( l=0; l<2 /* !!! !!! */; l++ )
	  counts[i][j][k][l]=0.0;

  /* Calculating the counts */
  for( k=0; k<num_seqs; k++ )
    for( l=0; l<sequence[k]->seq_length; l++ )
      for( i=0; i<dim; i++ )
	for( j=0; j<i+1; j++ )
	  if( !is_missing( sequence[k]->entry[l]->ddata[i] ) &&
	      !is_missing( sequence[k]->entry[l]->ddata[j] ) )
	    counts[i][j][sequence[k]->entry[l]->ddata[i]][sequence[k]->entry[l]->ddata[j]]+=1.0;

  /* Calculating mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i+1; j++ )
      {
	MI[i][j]=0.0;

	/* !!! Assuming binary data !!! */
	sum=0.0;
	for( k=0; k<2; k++ )
	  for( l=0; l<2; l++ )
	    sum+=counts[i][j][k][l];
	
	if( sum>0.0 )
	  {
	    /* Making probabilities out of the counts */
	    for( k=0; k<2; k++ )
	      for( l=0; l<2; l++ )
		counts[i][j][k][l]/=sum;
	    
	    /* Calculating MI */
	    MI[i][j]+=xlogx(counts[i][j][0][0]);
	    MI[i][j]+=xlogx(counts[i][j][0][1]);
	    MI[i][j]+=xlogx(counts[i][j][1][0]);
	    MI[i][j]+=xlogx(counts[i][j][1][1]);
	    MI[i][j]-=xlogx(counts[i][j][0][0]+counts[i][j][0][1]);
	    MI[i][j]-=xlogx(counts[i][j][1][0]+counts[i][j][1][1]);
	    MI[i][j]-=xlogx(counts[i][j][0][0]+counts[i][j][1][0]);
	    MI[i][j]-=xlogx(counts[i][j][0][1]+counts[i][j][1][1]);
	  }
	MI[j][i]=MI[i][j];
      }

  /* Deallocating the array of counts */
  for( i=0; i<dim; i++ )
    {
      for( j=0; j<i+1; j++ )
	{
	  for( k=0; k<2 /* !!! */; k++ )
	    delete counts[i][j][k];
	  delete counts[i][j];
	}	
      delete counts[i];
    }
  delete counts;

  return( MI );
}

double** Data::log_odds_ratio() {
  /* Calculating log-odds ratio for each pair of dimensions */
  double**** counts;
  double** lor;
  long dim;

  /* Index variable(s) */
  long i,j,k,l;

  dim=sequence[0]->entry[0]->ddim;

  /* Allocating the arrays of counts */
  counts=new double***[dim];
  for( i=0; i<dim; i++ )
    {
      counts[i]=new double**[i+1];
      for( j=0; j<i+1; j++ )
	{
	  counts[i][j]=new double*[2]; /* !!! !!! */
	  for( k=0; k<2 /* !!! */; k++ )
	    counts[i][j][k]=new double[2]; /* !!! !!! */
	}	
    }

  /* Allocating the MI matrix */
  lor=new double*[dim];
  for( i=0; i<dim; i++ )
    lor[i]=new double[dim];
  
  /* Initializing the counts */
  for( i=0; i<dim; i++ )
    for( j=0; j<i+1; j++ )
      for( k=0; k<2 /* !!! !!! */; k++ )
	for( l=0; l<2 /* !!! !!! */; l++ )
	  counts[i][j][k][l]=0.0;

  /* Calculating the counts */
  for( k=0; k<num_seqs; k++ )
    for( l=0; l<sequence[k]->seq_length; l++ )
      for( i=0; i<dim; i++ )
	for( j=0; j<i+1; j++ )
	  if( !is_missing( sequence[k]->entry[l]->ddata[i] ) && 
	      !is_missing( sequence[k]->entry[l]->ddata[j] ) )
	    counts[i][j][sequence[k]->entry[l]->ddata[i]][sequence[k]->entry[l]->ddata[j]]+=1.0;

  /* Calculating mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      {
	/* !!! Works only for binary data !!! */
	/* Log-odds ratio = log(11)+log(00)-log(01)-log(10) */
	lor[i][j]=log(counts[i][j][1][1])+log(counts[i][j][0][0])
	  -log(counts[i][j][0][1])-log(counts[i][j][1][0]);

#ifdef LINUX
	/* !!! Temporarily removing this section for non-linux compilations !!! */
	switch(  isinf(lor[i][j]) ) {
	case 1:
	  lor[i][j]=1e+100; /* Large positive number */
	  break;
	case -1:
	  lor[i][j]=-1e+100; /* Large negative number */
	  break;
	default:;
	}
#endif
	lor[j][i]=lor[i][j];
      }

  for( i=0; i<dim; i++ )
    /* Diagonal entries cannot be made sense of */
    lor[i][i]=0.0;

  /* Deallocating the array of counts */
  for( i=0; i<dim; i++ )
    {
      for( j=0; j<i+1; j++ )
	{
	  for( k=0; k<2 /* !!! */; k++ )
	    delete counts[i][j][k];
	  delete counts[i][j];
	}	
      delete counts[i];
    }
  delete counts;

  return( lor );
}
