/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* main.cpp */

#include "main.h"

main( int argc, char* argv[] ) {
  /* Variables */
  FILE* param_file;                  // Parameter file

  long seed;                         // Seed for random number generator

  Data* current_output_data;         // Current output data set
  Data* current_input_data;          // Current input data set

  StateData* best_states;            // Most likely sequences of states according to the model and the data

  long num_xval_sets;                // Number of cross-validation sets
  long num_seqs;                     // Number of sequences to be simulated

  /* Index variable(s) */
  long i,xval_index,current_n,n;

#ifdef MEASURE_TIME
  clock_t start_time,end_time;
  start_time=clock();
#endif

  /* Processing inputs */
  if( argc==1 )
    { /* Displaying help */ 
      printf( "Usage: mvnhmm <parameter file> [<seed>]\n" );
      test();
      return( 0 );
    } /* Displaying help */
  else
    { /* Processing parameters file */
      param_file=fopen( argv[1], "r" );
      if( !param_file )
	{ /* File not found */
	  fprintf( stderr, "Unable to open parameter file %s!  Aborting.\n", argv[1] );
	  exit( -1 );
	} /* File not found */

#ifdef INPUT_VERBOSE
      printf( "Parameter file %s successfully opened.\n", argv[1] );
#endif

      /* Reading the parameters */
      params=new Parameters;
      params->Read( param_file );

      /* Closing parameter file */
      fclose( param_file );

      if( argc<=2 )
	seed=0;
      else
	seed=atol( argv[2] );

      /* Initializing random number generator */
      srand48( seed );
    } /* Processing parameters file */

  /* Making sure parameters are ok */
  ProcessParameters();

  /* Determining the number of cross-validation sets */
  switch( params->xval_type ) {
  case XVAL_NONE:
    num_xval_sets=1;
    break;
  case XVAL_LEAVENOUT:
    switch( params->action ) {
    case ACTION_CODE_LEARN:
    case ACTION_CODE_VITERBI:
    case ACTION_CODE_LL:
    case ACTION_CODE_FILLING:
      num_xval_sets=output_data->num_seqs/params->number_out;
      break;
    case ACTION_CODE_SIM:
      num_xval_sets=params->num_models;
      break;
    case ACTION_CODE_ANALYZE:
      num_xval_sets=output_data->num_seqs/(params->num_simulations*params->number_out);
      break;
    case ACTION_CODE_DEBUG:
      num_xval_sets=output_data->num_seqs/params->number_out;      
      break;
    default:;
    }
    break;
  default:
    num_xval_sets=1;
  }

  /* Allocating data structures */
  current_input_data=NULL;
  current_output_data=NULL;
  switch( params->xval_type ) {
  case XVAL_NONE:
    switch( params->action ) {
    case ACTION_CODE_LEARN:
    case ACTION_CODE_VITERBI:
    case ACTION_CODE_LL:
    case ACTION_CODE_FILLING:
      current_output_data=new Data( output_data->num_seqs );
      if( input_data )
	current_input_data=new Data( input_data->num_seqs );
      else
	current_input_data=NULL;
      break;
    case ACTION_CODE_SIM:
      if( input_data )
	current_input_data=new Data( input_data->num_seqs );
      else
	current_input_data=NULL;
      num_seqs=params->num_data_seqs;
      break;
    case ACTION_CODE_ANALYZE:
      current_output_data=new Data( output_data->num_seqs );
      break;
    case ACTION_CODE_DEBUG:
      current_output_data=new Data( output_data->num_seqs );
      if( input_data )
	current_input_data=new Data( input_data->num_seqs );
      else
	current_input_data=NULL;
      break;      
    default:;
    }
    break;
  case XVAL_LEAVENOUT:
    switch( params->action ) {
    case ACTION_CODE_LEARN:
      current_output_data=new Data( output_data->num_seqs-params->number_out );
      if( input_data )
	current_input_data=new Data( input_data->num_seqs-params->number_out );
      else
	current_input_data=NULL;
      break;
    case ACTION_CODE_VITERBI:
    case ACTION_CODE_LL:
    case ACTION_CODE_FILLING:
      current_output_data=new Data( params->number_out );
      if( input_data )
	current_input_data=new Data( params->number_out );
      else
	current_input_data=NULL;
      break;
    case ACTION_CODE_SIM:
      if( input_data )
	current_input_data=new Data( params->number_out );
      else
	current_input_data=NULL;
      num_seqs=params->number_out;
      break;
    case ACTION_CODE_ANALYZE:
      current_output_data=new Data( params->num_simulations*params->number_out );
      break;
    case ACTION_CODE_DEBUG:
      current_output_data=new Data( params->number_out );
      if( input_data )
	current_input_data=new Data( params->number_out );
      else
	current_input_data=NULL;
      break;
    default:;
    }
    break;
  default:;
  }


  /* Looping over cross-validation sets */
  for( xval_index=0; xval_index<num_xval_sets; xval_index++ )
    {
      /* Initializing sets */
      switch( params->xval_type ) {
      case XVAL_NONE:
	switch( params->action ) {
	case ACTION_CODE_LEARN:
	  for( n=0; n<output_data->num_seqs; n++ )
	    {
	      current_output_data->sequence[n]=output_data->sequence[n];
	      if( input_data )
		current_input_data->sequence[n]=input_data->sequence[n];
	    }
	  break;
	case ACTION_CODE_VITERBI:
	case ACTION_CODE_LL:
	case ACTION_CODE_FILLING:
	  for( n=0; n<output_data->num_seqs; n++ )
	    {
	      current_output_data->sequence[n]=output_data->sequence[n];
	      if( input_data )
		current_input_data->sequence[n]=input_data->sequence[n];
	    }
	  break;
	case ACTION_CODE_SIM:
	  if( input_data )
	    for( n=0; n<input_data->num_seqs; n++ )
	      current_input_data->sequence[n]=input_data->sequence[n];
	  break;
	case ACTION_CODE_ANALYZE:
	  for( n=0; n<output_data->num_seqs; n++ )
	    current_output_data->sequence[n]=output_data->sequence[n];
	  break;
	case ACTION_CODE_DEBUG:
	  for( n=0; n<output_data->num_seqs; n++ )
	    {
	      current_output_data->sequence[n]=output_data->sequence[n];
	      if( input_data )
		current_input_data->sequence[n]=input_data->sequence[n];
	    }
	  break;	  
	default:;
	}
	break;
      case XVAL_LEAVENOUT:
	switch( params->action ) {
	case ACTION_CODE_LEARN:
	  fprintf( output_file, "%c Best model for set without years %ld-%ld\n", 
		   COMMENT_SYMBOL, xval_index*params->number_out+1, (xval_index+1)*params->number_out );
	  
#ifdef VERBOSE_BRIEF
	  printf( "Set #%ld\n", xval_index+1 );
#endif	  

	  current_n=0;

	  for( n=0; n<output_data->num_seqs; n++ )
	    if( n<xval_index*params->number_out || n>=(xval_index+1)*params->number_out )
	      {
		current_output_data->sequence[current_n]=output_data->sequence[n];
		if( input_data )
		  current_input_data->sequence[current_n]=input_data->sequence[n];
		
		current_n++;
	      }
	  break;
	case ACTION_CODE_VITERBI:
	case ACTION_CODE_LL:
	case ACTION_CODE_FILLING:
	  for( n=0; n<params->number_out; n++ )
	    {
	      current_output_data->sequence[n]=output_data->sequence[xval_index*params->number_out+n];
	      if( input_data )
		current_input_data->sequence[n]=input_data->sequence[xval_index*params->number_out+n];
	    }
	  break;
	case ACTION_CODE_SIM:
	  for( n=0; n<params->number_out; n++ )
	    if( input_data )
	      current_input_data->sequence[n]=input_data->sequence[xval_index*params->number_out+n];
	  break;
	case ACTION_CODE_ANALYZE:
	  for( n=0; n<params->num_simulations*params->number_out; n++ )
	    current_output_data->sequence[n]=
	      output_data->sequence[xval_index*params->num_simulations*params->number_out+n];
	  break;
	case ACTION_CODE_DEBUG:
	  for( n=0; n<params->number_out; n++ )
	    {
	      current_output_data->sequence[n]=output_data->sequence[xval_index*params->number_out+n];
	      if( input_data )
		current_input_data->sequence[n]=input_data->sequence[xval_index*params->number_out+n];
	    }
	  break;
	default:;
	}
	break;
      default:;
      }
      
#ifdef XVAL_VERBOSE
      fprintf( stdout, "Running cross-validation run %ld.\n", xval_index+1 );
      if( current_output_data )
	fprintf( stdout, "Output data consists of %ld sequences of length %ld each.\n", 
		 current_output_data->num_seqs, current_output_data->sequence[0]->seq_length );
      else
	fprintf( stdout, "No output data created\n" );
      
      if( current_input_data )
	fprintf( stdout, "Input data consists of %ld sequences of length %ld each.\n", 
		 current_input_data->num_seqs, current_input_data->sequence[0]->seq_length );
      else
	fprintf( stdout, "No input data created\n" );

      fflush( stdout );
#endif

      /* Running model */
      switch( params->action ) {
      case ACTION_CODE_LEARN:
	RunLearnParameters( current_output_data, 
			    current_input_data, 
			    theta, 
			    passed_theta,
			    params,
			    output_file );
	break;
      case ACTION_CODE_VITERBI:
	/* Calculating the most likely sequences of states */
	best_states=passed_theta[xval_index]->viterbi( current_output_data, current_input_data, 1 );
	best_states->WriteToFile( output_file );
	
	/* Deallocating the sequences of states */
	delete best_states;	  
	break;
      case ACTION_CODE_LL:
	RunLogLikelihoodData( current_output_data, current_input_data, passed_theta[xval_index], output_file );
	break;
      case ACTION_CODE_SIM:
	for( i=0; i<params->num_simulations; i++ )
	  RunSimulateData( current_input_data, 
			   passed_theta[xval_index], 
			   params, 
			   state_output_file, 
			   output_file,
			   num_seqs );	
	break;
      case ACTION_CODE_ANALYZE:
 	RunEvaluate( current_output_data, params->analysis_type, output_file );
	break;
     case ACTION_CODE_FILLING:
       RunHoleData( current_output_data, current_input_data, passed_theta[xval_index], output_file, params->poking_type );
       break;
      case ACTION_CODE_DEBUG:
	TestDebug( current_output_data, current_input_data, passed_theta[xval_index], output_file );
	break;
      default:
	fprintf( stderr, "Action is unidentified or unspecified.  Aborting.\n" );
	return( -1 );
      }
    }
  
  /* Deallocating data structures */
  switch( params->action ) {
  case ACTION_CODE_LEARN:
  case ACTION_CODE_VITERBI:
  case ACTION_CODE_LL:
  case ACTION_CODE_FILLING:
    delete current_output_data;
    if( input_data )
      {
	delete current_input_data;
	delete input_data;
      }
    break;
  case ACTION_CODE_SIM:
    if( input_data )
      {
	delete current_input_data;
	delete input_data;
      }
    break;
  case ACTION_CODE_ANALYZE:
    delete current_output_data;
    break;
  case ACTION_CODE_DEBUG:
    delete current_output_data;
    if( input_data )
      {
	delete current_input_data;
	delete input_data;
      }
    break;
  default:;
  }

  /* Deallocating passed models */
  if( passed_theta )
    {
      for( i=0; i<params->num_models; i++ )
	delete passed_theta[i];
      delete passed_theta;
    }
  
  /* Deallocating the model */
  if( theta )
    delete theta;
  
  /* Closing output file */
  if( output_file!=stdout )
    fclose( output_file );
    
  /* Deallocating the structure with data */
  if( output_data )
    delete output_data;

  /* Deallocating the structure with parameters */
  delete params;

#ifdef MEASURE_TIME
  end_time=clock();
  printf( "Elapsed CPU time: %lf\n", (double)(end_time-start_time)/(double)CLOCKS_PER_SEC );
#endif

  return( 0 );
}

Parameters::Parameters() {
  /* Filenames */
  output_data_filename=NULL;
  input_filename=NULL;
  model_filename=NULL;
  output_filename=NULL;

  state_filename=NULL;

  /* General variables */
  num_states=2;
  model_type=MODEL_TYPE_HMM;
  action=ACTION_CODE_UNKNOWN;
  emission=NULL;
  num_ddata_components=0;
  num_rdata_components=0;
  num_data_seqs=1;
  length_data_seqs=1;
  input_dist=NULL;
  num_dinput_components=0;
  num_rinput_components=0;
  num_input_seqs=1;
  length_input_seqs=1;
  xval_type=XVAL_NONE;
  number_out=1;
  num_models=1;

  /* Specific variables and options */
  /* Learning */
  num_restarts=1;
  em_precision=EM_EPSILON;
  initialization_type=INIT_RANDOM;
  num_simulations=1;
  analysis_type=ANALYSIS_MEAN;
  poking_type=POKING_TYPE_LOGP;

  cg_epsilon=CONJ_GRAD_EPSILON;
  maxent_epsilon=MIN_ME_LL_CHANGE;

  /* Other */
  output_datum=NULL;
  input_datum=NULL;
  
#ifdef TEMPDISABLED
  chowliu_MI_threshold=MI_THRESHOLD;
#endif
}

Parameters::~Parameters() {
  /* Index variable(s) */
  long i;

  /* Filenames */
  if( output_data_filename )
    delete output_data_filename;
  if( input_filename )
    delete input_filename;
  if( model_filename )
    delete model_filename;
  if( output_filename )
    delete output_filename;
  if( state_filename )
    delete state_filename;

  if( output_datum )
    delete output_datum;
  if( input_datum )
    delete input_datum;
  if( input_dist )
    delete input_dist;
  if( emission )
    delete emission;
}

void ProcessParameters( void ) {
  /* Making sure that passed parameters are proper, that all necessary
     files are passed and contain needed information */

  /* Index variables */
  long i;

  /* Creating templates for data points in data and input files */
  params->output_datum=new DataPoint( params->num_ddata_components, 
				      params->num_rdata_components );
  params->input_datum=new DataPoint( params->num_dinput_components, 
				     params->num_rinput_components );

  /* !!! Hack !!! */
  cg_epsilon=params->cg_epsilon;
  maxent_epsilon=params->maxent_epsilon;

  /* Reading in the output file if passed */
  if( params->output_data_filename )
    { 
      data_file=fopen( params->output_data_filename, "r" );
      if( !data_file )
	{
	  fprintf( stderr, "Unable to open data file %s for reading. Aborting\n", 
		   params->output_data_filename );
	  exit( -1 );
	}
      else
	{
#ifdef INPUT_VERBOSE
	  printf( "Data file %s successfully opened for reading.\n",
		  params->output_data_filename );
#endif
	  output_data=new Data( params->num_data_seqs );
	  output_data->ReadData( data_file, 0, params->num_data_seqs, 
				 params->length_data_seqs, params->output_datum );
	  
	  /* Closing the data file */
	  fclose( data_file );

	  /* Determining the number of missing entries */
	  for( i=0; i<output_data->num_seqs; i++ )
	    output_data->sequence[i]->CatalogueMissingEntries();
	}
    }      

  /* Reading in the input file if passed */
  if( params->input_filename )
    {
      input_file=fopen( params->input_filename, "r" );
      if( !input_file )
	{
	  fprintf( stderr, "Unable to open input file %s for reading.\n", 
		   params->input_filename );
	}
      else
	{
#ifdef INPUT_VERBOSE
	  printf( "Input variable file %s successfully opened for reading.\n", 
		  params->input_filename );
#endif	
	  input_data=new Data( params->num_input_seqs );
	  input_data->ReadInputData( input_file, 0, params->num_input_seqs, 
				     params->length_input_seqs, params->input_datum );
	      
	  /* Closing the input file */
	  fclose ( input_file );
	}
    }

  if( params->input_dist && !input_data )
    {
      fprintf( stderr, "Need input filename to work with input distribution.  Aborting.\n" );
      exit( -1 );
    }

  /* Output file */
  if( params->output_filename )
    { /* Output file listed */
      output_file=fopen( params->output_filename, "w" );
      if( !output_file )
	{
	  fprintf( stderr, "Unable to open file %s for writing.  Outputting to screen.\n",
		   params->output_filename );
	  output_file=stdout;
	}
    } /* Output file listed */
  else
    { /* No output file provided -- outputting results to the screen */ 
      output_file=stdout;
    } /* No output file provided -- outputting results to the screen */

  if( params->action!=ACTION_CODE_ANALYZE )
    { /* HMM model is defined */
      /* Initializing the model structure */
      if( !params->emission )
	{
	  fprintf( stderr, "Need to have proper emission distribution defined!  Aborting.\n" );
	  exit( -1 );
	}

      theta=new HMM( params->model_type, params->num_states, params->emission, params->input_dist );
      
      /* Reading in the model */
      if( params->model_filename )
	{
	  model_file=fopen( params->model_filename, "r" );
	  
	  /* Correcting the number of models in the file */
	  if( model_file )
	    { /* Reading in the models */
	      switch( params->xval_type ) {
	      case XVAL_NONE:
		switch( params->action ) {
		case ACTION_CODE_LEARN:
		  /* Number of models is as specified */
		  break;
		case ACTION_CODE_VITERBI:
		case ACTION_CODE_LL:
		case ACTION_CODE_SIM:
		case ACTION_CODE_FILLING:
		  /* Only one model in the file */
		  params->num_models=1;
		  break;
		case ACTION_CODE_DEBUG:
		  /* Only one model in the file */
		  params->num_models=1;
		  break;
		default:;
		}
		break;
	      case XVAL_LEAVENOUT:
		/* Determining whether the number of examples to be left out is proper */
		if( input_data )
		  {
		    if( params->number_out>(long)ceil((double)params->num_input_seqs/2.0) )
		      {
			fprintf( stderr, "Number of examples to leave out is too large.  Setting it to 1.\n" );
			params->number_out=1;
		      }
		  }
		else
		  if( params->number_out>(long)ceil((double)params->num_data_seqs/2.0) )
		    {
		      fprintf( stderr, "Number of examples to leave out is too large.  Setting it to 1.\n" );
		      params->number_out=1;
		    }

		/* Number of models is specified by the number of cross-validated sets */
		switch( params->action ) {
		case ACTION_CODE_VITERBI:
		case ACTION_CODE_LL:
		case ACTION_CODE_SIM:
		case ACTION_CODE_FILLING:
		  if( input_data )
		    params->num_models=params->num_input_seqs/params->number_out;
		  else
		    params->num_models=params->num_data_seqs/params->number_out;
		  break;
		case ACTION_CODE_DEBUG:
		  if( input_data )
		    params->num_models=params->num_input_seqs/params->number_out;
		  else
		    params->num_models=params->num_data_seqs/params->number_out;
		  break;		  
		default:;
		}
		break;
	      default:;
	      }
	      
	      /* Reading in the model */
	      passed_theta=new HMM*[params->num_models];
	      for( i=0; i<params->num_models; i++ )
		{
		  passed_theta[i]=theta->copy();
		  passed_theta[i]->ReadParameters( model_file );
		}
	      
	      /* Closing the file */
	      fclose( model_file );
	    } /* Reading in the models */
	  else
	    {
	      fprintf( stderr, "Unable to open model file %s for reading.  Ignoring.\n",
		       params->model_filename );
	    }
	}
    } /* HMM model is defined */
  
  /* Determining whether all information is passed for specific action */
  switch( params->action ) {
  case ACTION_CODE_LEARN:
    if( !output_data )
      { /* No data provided */
	fprintf( stderr, "No data sequences provided. Aborting.\n" );
	exit( -1 );
      } /* No data provided */
   
    if( input_data )
      if( params->num_data_seqs!=params->num_input_seqs || 
	  params->length_data_seqs!=params->length_input_seqs )
	{
	  /* !!! To be changed later !!! */
	  fprintf( stderr, "The dimensions of input and output data are not the same. Aborting.\n" );
	  exit( -1 );
	}
   
    break;
  case ACTION_CODE_VITERBI:
    if( !output_data )
      { /* No data provided */
	fprintf( stderr, "No data sequences provided. Aborting.\n" );
	exit( -1 );
      } /* No data provided */
    
    if( input_data )
      {
	if( params->num_data_seqs!=params->num_input_seqs || 
	    params->length_data_seqs!=params->length_input_seqs )
	  {
	    /* !!! To be changed later !!! */
	    fprintf( stderr, "The dimensions of input and output data are not the same. Aborting.\n" );
	    exit( -1 );
	  }
      }
    
    if( !passed_theta )
      {
	fprintf( stderr, "Action 'viterbi' requires models to be passed to it. Aborting.\n" );
	exit( -1 );
      }
    break;
  case ACTION_CODE_LL:
    if( !output_data )
      { /* No data provided */
	fprintf( stderr, "No data sequences provided. Aborting.\n" );
	exit( -1 );
      } /* No data provided */

    if( input_data )
      {
	if( params->num_data_seqs!=params->num_input_seqs || 
	    params->length_data_seqs!=params->length_input_seqs )
	  {
	    /* !!! To be changed later !!! */
	    fprintf( stderr, "The dimensions of input and output data are not the same. Aborting.\n" );
	    exit( -1 );
	  }
      }
    
    if( !passed_theta )
      {
	fprintf( stderr, "Action 'll' requires models to be passed to it. Aborting.\n" );
	exit( -1 );
      }
    break;
  case ACTION_CODE_SIM:
    if( !passed_theta )
      {
	fprintf( stderr, "Action 'simulation' requires models to be passed to it. Aborting.\n" );
	exit( -1 );
      }

    if( params->input_dist )
      {
	if( params->num_data_seqs!=params->num_input_seqs )
	  {
	    fprintf( stderr, "Changing the number of data sequences per simulated sets to %ld.\n",
		     params->num_input_seqs );
	    params->num_data_seqs=params->num_input_seqs;
	  }

	if( params->length_data_seqs!=params->length_input_seqs )
	  {
	    fprintf( stderr, "Changing the length of data sequences of simulated sets to %ld.\n",
		     params->length_input_seqs );
	    params->length_data_seqs=params->length_input_seqs;
	  }
      }

    /* Checking whether to open a file for the hidden state sequences */
    if( params->state_filename )
      { 
	state_output_file=fopen( params->state_filename, "w" );
	if( !state_output_file )
	  {
	    fprintf( stderr, "Unable to open file %s for writing. Not displaying the states\n",
		     params->state_filename );
	  }
      } 
   
    /* No output data is needed */
    if( output_data )
      {
	delete output_data;
	output_data=NULL;
      }	
    break;
  case ACTION_CODE_FILLING:
    if( !output_data )
      { /* No data provided */
	fprintf( stderr, "No data sequences provided. Aborting.\n" );
	exit( -1 );
      } /* No data provided */

    if( input_data )
      {
	if( params->num_data_seqs!=params->num_input_seqs || 
	    params->length_data_seqs!=params->length_input_seqs )
	  {
	    /* !!! To be changed later !!! */
	    fprintf( stderr, "The dimensions of input and output data are not the same. Aborting.\n" );
	    exit( -1 );
	  }
      }
    
    if( !passed_theta )
      {
	fprintf( stderr, "Action 'poking' requires models to be passed to it. Aborting.\n" );
	exit( -1 );
      }
    break;
  case ACTION_CODE_DEBUG:
    if( !output_data )
      { /* No data provided */
	fprintf( stderr, "No data sequences provided. Aborting.\n" );
	exit( -1 );
      } /* No data provided */

    if( input_data )
      {
	if( params->num_data_seqs!=params->num_input_seqs || 
	    params->length_data_seqs!=params->length_input_seqs )
	  {
	    /* !!! To be changed later !!! */
	    fprintf( stderr, "The dimensions of input and output data are not the same. Aborting.\n" );
	    exit( -1 );
	  }
      }
    
    if( !passed_theta )
      {
	fprintf( stderr, "Action 'debug' requires models to be passed to it. Aborting.\n" );
	exit( -1 );
      }
    break;
  default:;
  }
  
#ifdef FINAL_CHECK
  extern long em_verbose;

  printf( "Number of states: %ld\n", params->num_states );
  
  printf( "Type of model: " );
  switch( params->model_type ) {
  case MODEL_TYPE_HMM:
    printf( "HMM\n" );
    break;
  case MODEL_TYPE_NHMM:
    printf( "NHMM\n" );
    break;
  case MODEL_TYPE_MIX:
    printf( "mixture\n" );
    break;
  case MODEL_TYPE_NMIX:
    printf( "non-homogeneous mixture\n" );
    break;
  case MODEL_TYPE_SL:
    printf( "stateless\n" );
    break;
  default:
    printf( "unknown\n" );
  }
 
  printf( "Action type: " );
  switch( params->action ) {
  case ACTION_CODE_LEARN:
    printf( "parameter estimation\n" );
    break;
  case ACTION_CODE_VITERBI:
    printf( "best sequence calculation\n" );
    break;
  case ACTION_CODE_LL:
    printf( "log-likelihood calculation\n" );
    break;
  case ACTION_CODE_SIM:
    printf( "simulation\n" );
    break;
  case ACTION_CODE_ANALYZE:
    printf( "data analysis\n" );
    break;
  case ACTION_CODE_FILLING:
    printf( "data analysis with hole filling\n" );
    break;
  case ACTION_CODE_DEBUG:
    printf( "debugging\n" );
    break;
  default:
    printf( "unknown\n" );
  }

  if( params->emission )
    printf( "Emission distribution defined\n" );
  if( params->input_dist )
    printf( "Input distribution defined\n" );

  if( output_data )
    printf( "Output data filename: %s\n", params->output_data_filename );
  else
    printf( "No output data provided\n" );

  printf( "Number of sequences: %ld\n", params->num_data_seqs );
  printf( "Length of each sequence: %ld\n", params->length_data_seqs );
  printf( "Number of discrete-valued components: %ld\n", params->num_ddata_components );
  printf( "Number of real-valued components: %ld\n", params->num_rdata_components );

  if( input_data )
    {
      printf( "Input data filename: %s\n", params->input_filename );
      printf( "Number of sequences: %ld\n", params->num_input_seqs );
      printf( "Length of each sequence: %ld\n", params->length_input_seqs );
      printf( "Number of discrete-valued components: %ld\n", params->num_dinput_components );
      printf( "Number of real-valued components: %ld\n", params->num_rinput_components );      
    }
  else
    printf( "No input data provided\n" );
  
  printf( "Cross-validation type: " );
  switch( params->xval_type ) {
  case XVAL_NONE:
    printf( "none\n" );
    break;
  case XVAL_LEAVENOUT:
    printf( "leave-n-out\n" );
    printf( "Number of examples to leave out: %ld\n", params->number_out );
    break;
  default:
    printf( "unknown\n" );
  }

  if( passed_theta ) 
    {
      printf( "Model filename: %s\n", params->model_filename );
      printf( "Number of models: %ld\n", params->num_models );
    }
  else
    printf( "No models passed\n" );

  if( output_file!=stdout )
    printf( "Output file: %s\n", params->output_filename );
  else
    printf( "Outputting to the screen\n" );

  if( state_output_file )
    printf( "Output file for the simulated states: %s\n", params->state_filename );

  printf( "Number of random restarts: %ld\n", params->num_restarts );
  printf( "EM threshold constant: %0.12le\n", params->em_precision );
  if( em_verbose )
    printf( "Outputting intermediate EM messages to the screen\n" );
  else
    printf( "Not outputting intermediate EM messages to the screen\n" );
  printf( "Learning initialization type: " );
  switch( params->initialization_type ) {
  case INIT_RANDOM:
    printf( "random\n" );
    break;
  case INIT_EM:
    printf( "EM without inputs\n" );
    break;
  default:
    printf( "unknown\n" );
  }

  if( robust_first_state )
    printf( "Calculating first state probabilities from all observations\n" );
  else
    printf( "Calculating first state probabilities from first observations only\n" );

  printf( "Number of simulations: %ld\n", params->num_simulations );
  if( params->state_filename )
    printf( "Printing simulated hidden state sequences in file %s\n", params->state_filename );
  else
    printf( "No simulated state files\n" );
  
  printf( "Analysis type: " );
  switch( params->analysis_type ) {
  case ANALYSIS_MEAN:
    printf( "mean\n" );
    break;
  case ANALYSIS_CORR:
    printf( "correlation\n" );
    break;
  case ANALYSIS_PERS:
    printf( "persistence\n" );
    break;
  case ANALYSIS_DRY:
    printf( "dry spell length\n" );
    break;
  case ANALYSIS_WET:
    printf( "wet spell length\n" );
    break;
  case ANALYSIS_MI:
    printf( "mutual information\n" );
    break;
  case ANALYSIS_LO:
    printf( "log-odds ratio\n" );
    break;
  default:
    printf( "unknown\n" );
  }

  printf( "Missing value filling analysis type: " );
  switch( params->poking_type ) {
  case POKING_TYPE_LOGP:
    printf( "log-probability\n" );
    break;
  case POKING_TYPE_PRED:
    printf( "number of correct predictions\n" );
    break;
  case MISSING_IND_PROB:
    printf( "marginal probabilities for missing values\n" );
    break;
  case HIDDEN_STATE_PROB:
    printf( "probabilities for each value of hidden states\n" );
    break;
  default:
    printf( "unknown\n" );
  }
#endif
  return;
}

void test( void ) {
  srand48( 0 );
  printf( "Missing value = %lf\n", missing_value( (double)0.0 ) );
  return;
}


void TestDebug( Data* data, Data* input_data, HMM* theta, FILE* output ) {
  long** sim_state;
  long** sim_entry;
  long num_sims=1000;
  long burnin=1000;

  /* Index variable(s) */
  long i,t;

  long* temp_entry;
  long* temp_state;

  /* Cataloging missing values */
  data->sequence[2]->CatalogueMissingEntries();

  /* Allocating necessary structures */
  theta->AllocateEMStructures( data, input_data );  

  /* Initial fill-in of missing values */
  theta->InitializeMissingEntriesSequence( data->sequence[2] );

  for( i=0; i<burnin; i++ )
    {
      temp_state=theta->simulate_state_sequence_gibbs( data->sequence[2], NULL, 0, 1);
      temp_entry=theta->simulate_missing_discrete_sequence_gibbs( data->sequence[2], temp_state );
      delete temp_state;
      delete temp_entry;
    }


  sim_state=new long*[num_sims];
  sim_entry=new long*[num_sims];

  sim_state[0]=theta->simulate_state_sequence_gibbs( data->sequence[2], NULL, 0, 1 );
  sim_entry[0]=theta->simulate_missing_discrete_sequence_gibbs( data->sequence[2], sim_state[0] );

  for( i=1; i<num_sims; i++ )
    {
    }

#ifdef BLAH
  if( input_data )
    {
      sim_state[0]=theta->simulate_state_sequence_gibbs( data->sequence[0], input_data->sequence[0], 0, 1 );
      for( i=1; i<num_sims; i++ )
	sim_state[i]=theta->simulate_state_sequence_gibbs( data->sequence[0], input_data->sequence[0], 0, 0 );
    }
  else
    {
      sim_state[0]=theta->simulate_state_sequence_gibbs( data->sequence[0], NULL, 0, 1 );
      for( i=1; i<num_sims; i++ )
	sim_state[i]=theta->simulate_state_sequence_gibbs( data->sequence[0], NULL, 0, 0 );
    }

  /* Outputting the outputs */
  for( i=0; i<num_sims; i++ )
    {
      for( t=0; t<data->sequence[0]->seq_length; t++ )
	fprintf( output, "%ld ", sim_state[i][t] );
      fprintf( output, "\n" );
    }

  theta->CalculateBetaSequence( data->sequence[0], 0 );
  theta->CalculateSummaries( data );
  
  /* Outputting summaries of the forward backward pass */
  for( t=0; t<data->sequence[0]->seq_length; t++ )
    fprintf( output, "%0.12le ", theta->Anti[0][0][t] );
  fprintf( output, "\n" );
  for( t=0; t<data->sequence[0]->seq_length; t++ )
    fprintf( output, "%0.12le ", theta->Anti[1][0][t] );
  fprintf( output, "\n" );

  /* Deallocating arrays */
  theta->DeallocateEMStructures( data, input_data );

  for( i=0; i<num_sims; i++ )
    delete sim_state[i];
  delete sim_state;
#endif

  theta->DeallocateEMStructures( data, input_data );

  /* Displaying the data */
  for( t=0; t<data->sequence[2]->seq_length; t++ )
    fprintf( output, "%ld ", sim_state[0][t] );
  fprintf( output, "\n" );
  for( i=0; i<data->sequence[2]->num_missing_discrete_entries; i++ )
    fprintf( output, "%ld\t%ld\t%ld\n", 
	     data->sequence[2]->missing_discrete[i][0],
	     data->sequence[2]->missing_discrete[i][1],
	     data->sequence[2]->entry[data->sequence[2]->missing_discrete[i][0]]->
	     ddata[data->sequence[2]->missing_discrete[i][1]] );

  delete sim_state[0];
  delete sim_state;
  delete sim_entry[0];
  delete sim_entry;

  return;
}
