/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* model.cpp */
/* Declarations for class HMM */
#include "struct.h"

HMM::HMM( long t, long num, Distribution* e_dist, Distribution* input_dist ) {

  /* Index variable(s) */
  long i,j;
	
  /* Initializing the number of states */
  num_states=num;

  if( e_dist )
    { /* Emission distribution passed */
      emission=new Distribution*[num_states];
      for( i=0; i<num_states; i++ )
	emission[i]=e_dist->copy();
    } /* Emission distribution passed */
  else
    emission=NULL;

  if( input_dist )
    /* Input distribution passed */
    input=input_dist->copy();
  else 
    input=NULL;

  /* Initializing the type of the model */
  type=t;

  if( type==MODEL_TYPE_HMM )
    { /* Initializing the transition matrix */
      trans=new double*[num_states];
      for( i=0; i<num_states; i++ )
	trans[i]=new double[num_states];

      for( i=0; i<num_states; i++ )
	for( j=0; j<num_states; j++ )
	  trans[i][j]=0.0;
    } /* Initializing the transition matrix */

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_MIX )
    { /* Initializing first state probability vector */
      first=new double[num_states];
      
      for( i=0; i<num_states; i++ )
	first[i]=0.0;
    } /* Initializing first state probability vector */

  /* Initializing EM arrays */
  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {
      alpha=new double**[num_states];
      beta=new double**[num_states];
    }
  pb=new double**[num_states];
  norm_const=new double[num_states];
  Anti=new double**[num_states];

  if( type==MODEL_TYPE_NHMM || type==MODEL_TYPE_NMIX )
    {
      /* Initializing the transition probability array */
      p_tr=new double***[num_states];
      for( i=0; i<num_states; i++ )
	p_tr[i]=new double**[num_states];

      /* Initializing log-unnormalized transition probability array */
      log_un_prob=new double***[num_states];
      for( i=0; i<num_states; i++ )
	log_un_prob[i]=new double**[num_states];
    }
  
  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {
      Bntij=new double***[num_states];
      for( i=0; i<num_states; i++ )
	Bntij[i]=new double**[num_states];
    }

  ll=-1E+100;
  ll_normalized=-1E+100;
}

HMM::~HMM() {
  /* Index variable(s) */
  long i;

  /* Deallocating emission structure */
  if( emission )
    {
      for( i=0; i<num_states; i++ )
	delete emission[i];
      delete emission;
    }

  /* Deallocating input structure */
  if( input )
    delete input;

  if( type==MODEL_TYPE_HMM )
    { /* Deallocating the transition matrix */
      for( i=0; i<num_states; i++ )
	delete trans[i];
      delete trans;
    } /* Deallocating the transition matrix */

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_MIX )
    /* Deallocating first state probability vector */
    delete first;
    
  /* Deallocating EM arrays */
  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {
      delete alpha;
      delete beta;
    }
  delete pb;
  delete Anti;

  /* Normalization constants */
  delete norm_const;

  if( type==MODEL_TYPE_NHMM || type==MODEL_TYPE_NMIX )
    {
      /* Deallocating the probability arrays */
      for( i=0; i<num_states; i++ )
	delete p_tr[i];
      delete p_tr;
      
      for( i=0; i<num_states; i++ )
	delete log_un_prob[i];
      delete log_un_prob;
    }

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
    {
      for( i=0; i<num_states; i++ )
	delete Bntij[i];
      delete Bntij;
    }
}

HMM* HMM::copy() {
  /* Copy constructor */	
  HMM* model;

  /* Index variable(s) */
  long i,j;

  model=new HMM( type, num_states, NULL, NULL );

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_MIX )
    /* Copying values for the first state distribution */
    for( i=0; i<num_states; i++ )
      model->first[i]=first[i];

  if( type==MODEL_TYPE_HMM )
    /* Copying values for the transition matrix */
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	model->trans[i][j]=trans[i][j];

  /* Copying type */
  model->type=type;
  
  /* Copying emissions */
  if( emission )
    {
      model->emission=new Distribution*[num_states];
      for( i=0; i<num_states; i++ )
	model->emission[i]=emission[i]->copy();
    }

  /* Copying input distributions */
  if( input )
    model->input=input->copy();
    
  model->ll=ll;
  model->ll_normalized=ll_normalized;

  return( model );
}

void HMM::Initialize( Data* data, Data* input_vars ) {
  /* Index variable(s) */
  long i,j;
  double sum;
  
  /* Initializing parameters for the HMM */
  if( type==MODEL_TYPE_HMM )
    /* Transition matrix */
    for( i=0; i<num_states; i++ )
      {
	sum=0.0;
	for( j=0; j<num_states; j++ )
	  {
	    trans[i][j]=drand48();
	    sum+=trans[i][j];
	  }
	
	/* Renormalizing */
	for( j=0; j<num_states; j++ )
	  trans[i][j]/=sum;    
      }
	
  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_MIX )
    { /* First state distribution */
      sum=0.0;
      for( j=0; j<num_states; j++ )
	{
	  first[j]=drand48();
	  sum+=first[j];
	}
      
      /* Renormalizing */
      for( j=0; j<num_states; j++ )
	first[j]/=sum;
    } /* First state distribution */
     
  /* Initializing emission distribution */
  if( emission )
    for( i=0; i<num_states; i++ )
      emission[i]->Initialize( data );
  
  if( input )
    /* Initializing input distributions */
    input->Initialize( input_vars );

  return;
}

void HMM::InitializeByEM( Data* data, Data* input_vars ) {
  /* Initializes the model by estimating the parameters from the data according to
     a simpler sub-model */

  Distribution* input_dist;  // A copy of the input distribution (for expansion)
  extern long num_failed;    // Number of failed runs so far -- should not change in the end
  long num_prev_failed;      // Temporary value of num_failed;

  if( type==MODEL_TYPE_NHMM || type==MODEL_TYPE_NMIX )
    { /* Models with complex input layer */
      input_dist=input->copy();

      /* Simplifying the model by dropping input structures */
      Project();

      num_prev_failed=num_failed;
      do
	{
	  /* Initializing the compressed model */
	  Initialize( data, input_vars );

	  EM( data, input_vars, EM_EPSILON );
	  if( num_failed!=num_prev_failed )
	    /* Run failed -- setting the counter of failed EM runs back to the previous value */
	    num_failed=num_prev_failed;
	  else
	    /* Run is ok -- making the counters differ */
	    num_prev_failed--;
	}
      while( num_failed==num_prev_failed );

      /* Expanding the model to the original type */
      Expand( input_dist );
    } /* Models with complex input layer */

  return;
}

double HMM::log_likelihood( Data* data, Data* input_vars ) {

  /* Index variable(s) */
  long n;

  ll=0.0;
  for( n=0; n<data->num_seqs; n++ )
    if( input_vars )
      ll+=log_likelihood_sequence( data->sequence[n], input_vars->sequence[n], n, 0 );
    else
      ll+=log_likelihood_sequence( data->sequence[n], NULL, n, 0 );

  ll_normalized=ll/data->total_entries();

  return( ll );
}

double HMM::hole_score( Data* data, Data* input_data, long analysis_type ) {
  double score=0.0;

  /* Index variable(s) */
  long i;
  
  /* Sequences are assumed iid */
  for( i=0; i<data->num_seqs; i++ )
    if( input_data )
      score+=hole_score_sequence( data->sequence[i], input_data->sequence[i], analysis_type, i );
    else
      score+=hole_score_sequence( data->sequence[i], NULL, analysis_type, i );

  return( score );
}

double HMM::hole_score_sequence( Sequence* data, Sequence* input_data, long analysis_type, long index ) {
  double score=0.0;            // Per bit log-likelihood score of the data
  double ll;                   // Log-likelihood of the unaltered sequence

  /* Index variable(s) */
  long t,m;

  /* Temporary variable(s) */
  long covered_value;
  long entry_dim;
  double current_ll;

  /* Calculating the log-likelihood of the sequence */
  ll=log_likelihood_sequence( data, input_data, index, 0 );

#ifdef HOLE_VERBOSE
  printf(" Log-likelihood for sequence %ld = %0.6le\n", index+1, ll );
#endif

  /*** Calculating the scores ***/
  /* !!! Determining the dimension of the discrete-valued entries !!! */
  entry_dim=data->entry[0]->ddim;

  /* Starting changing values in the end of the sequence and moving towards the front */
  for( t=0; t<data->seq_length; t++ )
    {
      /* !!! Hack -- should be replaced by something modular !!! */
      for( m=0; m<entry_dim; m++ )
	{
	  /* !!! Assuming two possible values (0 and 1) !!! */

	  /* Making part of the entry different from before */
	  covered_value=data->entry[t]->ddata[m];
	  if( covered_value==0 )
	    data->entry[t]->ddata[m]=1;
	  if( covered_value==1 )
	    data->entry[t]->ddata[m]=0;
	  
	  if( !is_missing( covered_value ) )
	    {
	      /* Calculating the score for the altered sequence */
	      current_ll=log_likelihood_sequence( data, input_data, index, t );
#ifdef HOLE_VERBOSE
	      printf( "%0.2le ", current_ll );
#endif      
	      /* Updating the score */
	      switch( analysis_type ) {
	      case POKING_TYPE_LOGP:
		/* Log-probability of the correct prediction */
		score+=-log(exp(current_ll-ll)+1.0);
		break;
	      case POKING_TYPE_PRED:
		/* Number of correct predictions */
		if( ll>current_ll )
		  score+=1.0;
		break;
	      default:;
	      }
	    }
	  else
	    {
#ifdef HOLE_VERBOSE
	      printf( "-1 " );
#endif
	    }

	  /* Putting the original value back */
	  data->entry[t]->ddata[m]=covered_value;
	}
#ifdef HOLE_VERBOSE
      printf( "\n" );
#endif
      /* Resetting intermediate values need to compute the log-likelihood */
      log_likelihood_sequence( data, input_data, index, 0 );
    }

  return( score );
}

void HMM::MissingIndividualMarginals( Data* data, Data* input_data, FILE* output ) {
  /* Computing and saving to a file marginal probabilities for all missing entries */

  long num_values;           // Number of possible values for the current entry
  double prob[MAXNUMVALUES]; // Marginal probability for the current missing value
  
  double baseline_ll;        // log-likelihood of a sequence in question

  long prev_t=-1;

  /* Index variable(s) */
  long index,n,t,i,b,k;

  /* Temporary variable(s) */
  double sum;

  /* !!! Assuming all variables have the same number of values -- binary !!! */
  num_values=2;

  for( n=0; n<data->num_seqs; n++ )
    data->sequence[n]->CatalogueMissingEntries();

  EStep( data, input_data );
  CalculateSummaries( data );

  /* Computing the probabilities */
  for( n=0; n<data->num_seqs; n++ )
    {
#ifdef BLAH
      /* Starting a new sequence */
      if( input_data )
	baseline_ll=log_likelihood_sequence( data->sequence[n], input_data->sequence[n],
					     n, 0 );
      else
	baseline_ll=log_likelihood_sequence( data->sequence[n], NULL,
					     n, 0 );
#endif
      prev_t=-1;
      
      for( index=0; index<data->sequence[n]->num_missing_discrete_entries; index++ )
	{
	  t=data->sequence[n]->missing_discrete[index][0];
	  i=data->sequence[n]->missing_discrete[index][1];
#ifdef BLAH	  
	  if( prev_t!=t && prev_t!=-1)
	    if( input_data )
	      baseline_ll=log_likelihood_sequence( data->sequence[n], input_data->sequence[n],
						   n, prev_t );
	    else
	      baseline_ll=log_likelihood_sequence( data->sequence[n], NULL,
						   n, prev_t );
#endif
	  prev_t=t;
	  
	  sum=0.0;
	  for( b=0; b<num_values-1; b++ )
	    { /* Looping over the value of current missing entry */
	      /* Setting the value */
	      data->sequence[n]->entry[t]->ddata[i]=b;
#ifdef BLAH
	      if( input_data )
		prob[b]=log_likelihood_sequence( data->sequence[n], input_data->sequence[n], n, t );
	      else
		prob[b]=log_likelihood_sequence( data->sequence[n], NULL, n, t );

	      prob[b]=exp(prob[b]-baseline_ll );
#endif
	      prob[b]=0.0;
	      for( k=0; k<num_states; k++ )
		if( t==0 )
		  prob[b]+=Anti[k][n][t]*emission[k]->prob( data->sequence[n]->entry[t], NULL )
		    /pb[k][n][t];
		else
		  prob[b]+=Anti[k][n][t]*emission[k]->prob( data->sequence[n]->entry[t],
							    data->sequence[n]->entry[t-1] )
		    /pb[k][n][t];

	      sum+=prob[b];
	    } /* Looping over the value of current missing entry */
	  
	  /* Computing the value for the remaining entry */
	  prob[num_values-1]=1.0-sum;
	  
	  /* Placing missing value back */
	  data->sequence[n]->entry[t]->ddata[i]=missing_value( (long)0 );
	  
	  /* Outputting the value */
	  //	  fprintf( output, "%ld %ld %ld", n, t, i );
	  /* !!! Hack -- only outputting the probability of 1 !!! */
	  //	  for( b=0; b<num_values; b++ )
	  //	    fprintf( output, "\t%0.12le", prob[b] );
	  fprintf( output, "%0.12le", prob[1] );
	  fprintf( output, "\n" );
	}
    }  

  return;
}

void HMM::StateProbabilities( Data* data, Data* input_data, FILE* output ) {
  /* Computing and writing to a file P(S_nt|R_n,X_n) */

  /* Index variable(s) */
  long n,t,i;

  EStep( data, input_data );
  CalculateSummaries( data );

  /* Computing the probabilities */
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      {
	for( i=0; i<num_states; i++ )
	  fprintf( output, "%0.12le\t", Anti[i][n][t] );
	fprintf( output, "\n" );
      }

  return;
}

void HMM::WriteToFile( FILE* output ) {
  /* Records model parameters into a file as 12-digit doubles */

  /* Index variable(s) */
  long i,j;

  /* File clarification */
  fprintf( output, "%c This file contains the parameters of the HMM\n", COMMENT_SYMBOL );

  if( type==MODEL_TYPE_HMM )
    fprintf( output, "%c First state probabilities:\n", COMMENT_SYMBOL ); 
  else if( type==MODEL_TYPE_MIX )
    fprintf( output, "%c Mixing probabilities:\n", COMMENT_SYMBOL );

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_MIX )
    { /* First state or mixing probability distributions */
      for( i=0; i<num_states; i++ )
	fprintf( output, "\t%0.12le", first[i] );
      fprintf( output, "\n\n" );
    } /* First state or mixing probability distributions */

  if( type==MODEL_TYPE_HMM )
    { /* Transition matrix */
      fprintf( output, "%c Transition matrix:\n", COMMENT_SYMBOL );
      for( i=0; i<num_states; i++ )
	{
	  for( j=0; j<num_states; j++ )
	    fprintf( output, "\t%0.12le", trans[i][j] );
	  fprintf( output, "\n" );
	}
      fprintf( output, "\n" );
    } /* Transition matrix */

  /* Emission parameters */
  fprintf( output, "%c Parameters for the emission distributions for different possible hidden states:\n",
	   COMMENT_SYMBOL );
  for( i=0; i<num_states; i++ )
    {
      fprintf( output, "%c State #%ld:\n", COMMENT_SYMBOL, i+1 );
      emission[i]->WriteToFile( output );
      fprintf( output, "\n" );
    }
  fprintf( output, "\n" );

  /* Displaying input distribution parameters (if exist) */
  if( input )
    {
      fprintf( output, "%c Parameters for the distributions on the inputs:\n", COMMENT_SYMBOL );
      input->WriteToFile( output );
      fprintf( output, "\n" );
    }

  /* Displaying the log-likelihood for the data set */
  fprintf( output, "%c Log-likelihood of the data set: %0.12le\n", COMMENT_SYMBOL, ll );

  /* Displaying the per-dim log-likelihood for the data set */
  fprintf( output, "%c Per-dimension log-likelihood of the data set: %0.12le\n", 
	   COMMENT_SYMBOL, ll_normalized );

  /* Displaying BIC score for the model and set */
  fprintf( output, "%c Bayes Information Criterion score: %0.12le\n", COMMENT_SYMBOL, bic ); 

  fflush( output );

  return;
}

long HMM::num_params() {
  /* Calculating the number of free parameters for the model */
  long num=0;

  /* Index variables */
  long i;

  if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_MIX )
    /* Accounting for probability of the first state or mixing probabilities */
    num+=num_states-1;
  
  if( type==MODEL_TYPE_HMM )
    /* Accounting for transition matrix */
    num+=num_states*(num_states-1);

  /* Accounting for output layer */
  for( i=0; i<num_states; i++ )
    num+=emission[i]->num_params();

  /* Accounting for input layer */
  if( input )
    num+=input->num_params();

  return( num );
}

double HMM::BIC( long num_points ) {
  /* Calculating Bayes Inference Criterion score */
  
  return( log(double(num_points))*double(num_params())-2*ll );
}

void HMM::PostProcess() {
  /* Post training rearrangement of parameters */

  /* Index variable(s) */
  long i;
  
  /* So far, may need to rearrange only the emission probabilities */
  for( i=0; i<num_states; i++ )
    emission[i]->PostProcess();

  return;
}

void HMM::Project() {
  /* Converting HMM models with inputs to models without inputs */

  /* Index variable(s) */
  long i,j;

  /* Temporary variable(s) */
  double sum;

  /* Converting the structures */
  switch( type ) {
  case MODEL_TYPE_HMM:
  case MODEL_TYPE_MIX:
    /* No change */
    break;
  case MODEL_TYPE_NHMM:
    /* Allocating transition terms */
    trans=new double*[num_states];
    for( i=0; i<num_states; i++ )
      trans[i]=new double[num_states];

    /* Retrieving the values of transitions for sigmas 
       of the transitional logistic distribution */
    for( i=0; i<num_states; i++ )
      { /* Previous state i */
	sum=0.0;
	for( j=0; j<num_states; j++ )
	  {
	    trans[i][j]=exp(input->sigma[i][j]);
	    sum+=trans[i][j];
	  }

	/* Normalizing */
	for( j=0; j<num_states; j++ )
	  trans[i][j]/=sum;
      } /* Previous state i */

  case MODEL_TYPE_NMIX:
    /* Allocating first state term */
    first=new double[num_states];

    /* Retrieving the values of the first state probabilities
       from lambdas of the logistic distribution */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      {
	/* !!! !!! */
	first[i]=exp(input->lambda[i]);
	sum+=first[i];
      }

    /* Normalizing */
    for( i=0; i<num_states; i++ )
      first[i]/=sum;

    /* Deallocating transition probability and unnormalized probability arrays */
    for( i=0; i<num_states; i++ )
      delete p_tr[i];
    delete p_tr;
    
    for( i=0; i<num_states; i++ )
      delete log_un_prob[i];
    delete log_un_prob;    

    /* Deleting input distribution */
    delete input;
    input=NULL;

    break;
  default:;
  }

  /* Updating the type */
  switch( type ) {
  case MODEL_TYPE_HMM:
  case MODEL_TYPE_MIX:
    /* No change */
    break;
  case MODEL_TYPE_NHMM:
    type=MODEL_TYPE_HMM;
    break;
  case MODEL_TYPE_NMIX:
    type=MODEL_TYPE_MIX;
    break;
  default:;
  }
  
  return;
}

void HMM::Expand( Distribution* input_dist ) {
  /* Expands a model without inputs to a model with inputs of dimension d */

  /* Index variable(s) */
  long i,j;

  /* Temporary variable(s) */
  double temp;

  if( !input )
    /* Adding input distribution */
    input=input_dist;

  /* Converting the structures */
  switch( type ) {
  case MODEL_TYPE_NHMM:
  case MODEL_TYPE_NMIX:
    /* No change */
    break;
  case MODEL_TYPE_HMM:
    /* Calculating sigmas from transition probabilities */

    for( i=0; i<num_states; i++ )
      {
	/* Making sure the entries are such that the entries for the first state are zero */
	temp=log(trans[i][0]);
	input->sigma[i][0]=0.0;
	for( j=1; j<num_states; j++ )
	  input->sigma[i][j]=log(trans[i][j])-temp;
      }

    /* Deallocating transition probabilities */
    for( i=0; i<num_states; i++ )
      delete trans[i];
    delete trans;

  case MODEL_TYPE_MIX:
    /* Calculating lambdas from first state probabilities */

    /* Making sure the entries are such that the entries for the first state are zero */
    temp=log(first[0]);
    for( i=1; i<num_states; i++ )
      input->lambda[i]=log(first[i])-temp;

    /* Zeroing out the linear term */
    for( i=0; i<num_states; i++ )
      for( j=0; j<input->dim; j++ )
	input->rho[i][j]=0.0;

    /* Deallocating first state probabilities */
    delete first;

    /* Initializing the transition probability array */
    p_tr=new double***[num_states];
    for( i=0; i<num_states; i++ )
      p_tr[i]=new double**[num_states];
    
    /* Initializing log-unnormalized transition probability array */
    log_un_prob=new double***[num_states];
    for( i=0; i<num_states; i++ )
      log_un_prob[i]=new double**[num_states];

    break;
  default:;
  }

  /* Updating the type */
  switch( type ) {
  case MODEL_TYPE_NHMM:
  case MODEL_TYPE_NMIX:
    /* No change */
    break;
  case MODEL_TYPE_HMM:
    type=MODEL_TYPE_NHMM;
    break;
  case MODEL_TYPE_MIX:
    type=MODEL_TYPE_NMIX;
    break;
  default:;
  }

  return;
}
