/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* simulation.cpp */

#include "struct.h"

static long is_first_normal_value=1;

long** HMM::simulate_states( Data* input_var, long num_seqs, long num_entries ) {
  /* Simulating sequences of hidden states according to the model theta */

  /* Variables */
  long** states;                // Output structure

  double* prob;                 // Probability distribution for S_nt given previous S_n,t-1 and Xs

  /* Index variable(s) */
  long n,t;

  /* Allocating the structure for the state sequences */
  states=new long*[num_seqs];

  if( input )
    /* Allocating probability distribution for P(S_nt|S_n,t-1,Xs) */
    prob=new double[num_states];

  for( n=0; n<num_seqs; n++ )
    { /* Generating state sequence n */
      
      /* Allocating states[i] */
      if( input )
	num_entries=input_var->sequence[n]->seq_length;

      states[n]=new long[num_entries];

      for( t=0; t<num_entries; t++ )
	{ /* Generating state S_nt */	  
	
	  /* Resetting probability distribution for the state */
	  switch( type ) {
	  case MODEL_TYPE_HMM:
	    if( t==0 )
	      prob=first;
	    else
	      prob=trans[states[n][t-1]];
	    break;
	  case MODEL_TYPE_NHMM:
	    if( t==0 )
	      input->probLogistic( input_var->sequence[n]->entry[t], -1, prob );
	    else
	      input->probLogistic( input_var->sequence[n]->entry[t], states[n][t-1], prob );
	    break;
	  case MODEL_TYPE_MIX:
	    prob=first;
	    break;
	  case MODEL_TYPE_NMIX:
	    input->probLogistic( input_var->sequence[n]->entry[t], -1, prob );
	    break;
	  default:;
	  }

	  /* Simulating state S_nt */
	  states[n][t]=generateBernoulli( prob, num_states );
	} /* Generating state S_nt */
      
    } /* Generating state sequence n */

      /* Deallocating structures */
  if( input )
    /* Deallocating distribution P(S_nt|S_n,t-1,Xs) */
    delete prob;

  return( states );
}

long* HMM::simulate_state_sequence_gibbs( Sequence* data, Sequence* input_data, long index, long do_init ) {
  /* Simulating a sequence of hidden states according to the model and the 
     COMPLETE input and output sequence */

  /* Variables */
  long* state;                 // Output structure
  double* prob;                // Probability distribution for S_t given everyting instanstiated

  /* Index variable(s) */
  long t,i;

  /* Temporary variable(s) */
  double sum;
  
  /* Allocating the structure for the state sequences */
  state=new long[data->seq_length];
  
  /* Allocating the array of probabilities */
  prob=new double[num_states];

  if( do_init ) 
    {
    
      /* Calculating emission probabilities */
      CalculateProbEmissionsSequence( data, index, 0 );
      
      switch( type ) {
      case MODEL_TYPE_NMIX:
	CalculateProbSSequence( input_data, index, 0 );
	break;
      case MODEL_TYPE_NHMM:
	CalculateProbSSequence( input_data, index, 0 );
	CalculateAlphaSequence( data, index, 0 );
	break;
      case MODEL_TYPE_HMM:
	CalculateAlphaSequence( data, index, 0 );
	break;
      default:;
      }
    }

  for( t=data->seq_length-1; t>=0; t-- )
    { /* Simulating hidden state S_t */
      sum=0.0;
      if( type==MODEL_TYPE_MIX )
	/* Homogeneous mixture model */
	for( i=0; i<num_states; i++ )
	  {
	    prob[i]=pb[i][index][t]*first[i];
	    sum+=prob[i];
	  }
      else if( type==MODEL_TYPE_NMIX )
	/* Nonhomogeneous mixture model */
	for( i=0; i<num_states; i++ )
	  {
	    prob[i]=pb[i][index][t]*p_tr[i][0][index][t];
	    sum+=prob[i];
	  }
      else if( type==MODEL_TYPE_HMM || type==MODEL_TYPE_NHMM )
	{ /* Hidden Markov model */
	  if( t==data->seq_length-1 )
	    /* Last hidden variable */
	    for( i=0; i<num_states; i++ )
	      {
		prob[i]=alpha[i][index][t];
		sum+=prob[i];
	      }
	  else
	    /* Not the last hidden variable */
	    for( i=0; i<num_states; i++ )
	      {
		if( type==MODEL_TYPE_NHMM )
		  prob[i]=alpha[i][index][t]*p_tr[state[t+1]][i][index][t+1];
		else if( type==MODEL_TYPE_HMM )
		  prob[i]=alpha[i][index][t]*trans[i][state[t+1]];
		sum+=prob[i];
	      }
	} /* Hidden Markov model */
	
      /* Normalizing probabilities */
      for( i=0; i<num_states; i++ )
	prob[i]/=sum;

      /* Simulating state t */
      state[t]=generateBernoulli( prob, num_states );
    } /* Simulating hidden state S_t */

  delete prob;

  return( state );
}

long* HMM::simulate_missing_discrete_sequence_gibbs( Sequence* data, long* state ) {
  long* missing;
  double* prob;

  /* Index variables */
  long n,t,i,b,e,m1,m2,mi;

  /* Temporary variable(s) */
  double sum;
  Distribution* dist;

  if( data->num_missing_discrete_entries>0 )
    /* Allocating the array for new values */
    missing=new long[data->num_missing_discrete_entries];
  else
    missing=NULL;

  /* !!! Assuming Conditional Chow-Liu Forests for emissions !!! */

  for( n=0; n<data->num_missing_discrete_entries; n++ )
    { /* Looping over missing values */
      t=data->missing_discrete[n][0];
      i=data->missing_discrete[n][1];
      
      //      printf( "Missing value at time %ld at component %ld\n", t, i );

      dist=emission[state[t]];
      prob=new double[dist->num_states];
      
      /* Since all interactions are pairwise, can some over all edges
	 with R_t^i as a vertex. */
      /* Processing P(R_t|R_t-1,S_t) */
      //      printf( "S(%ld)=%ld\n", t, state[t] );
      if( dist->type==DIST_CONDCHOWLIU )
	{
	  /* Locating the index corresponding to i */
	  for( mi=0; mi<dist->dim; mi++ )
	    if( dist->dim_index[mi]==i )
	      break;

	  //	  printf( "Corresponding index=%ld\n", mi );
	  //	  printf( "Considering auto-regressive transition %ld->%ld\n", t-1, t );

	  for( b=0; b<dist->num_states; b++ )
	    prob[b]=dist->ind_prob[mi][b];
	  
	  for( e=0; e<dist->num_edges; e++ )
	    {
	      m1=dist->edge[e][0];
	      m2=dist->edge[e][1];
	      if( m1>=dist->dim && m1-dist->dim==mi )
		{ /* i is the first node for edge e */
		  if( m2>=dist->dim )
		    //		    {
		    //		  printf( "Edge connects %ld and %ld\n", m1, m2 );

		    /* Both nodes are in the current time slice */
		    for( b=0; b<dist->num_states; b++ )
		      prob[b]*=
			dist->edge_prob[e][b][data->entry[t]->ddata[dist->dim_index[m2-dist->dim]]]/
			dist->ind_prob[m1][b];
		  //		    }
		  else if( t>0 )
		    //		    {
		    //		  printf( "Edge connects %ld and %ld\n", m1, m2 );

		    /* The second node is in the previous time slice */
		    for( b=0; b<dist->num_states; b++ )
		      prob[b]*=
			dist->edge_prob[e][b][data->entry[t-1]->ddata[dist->dim_index[m2]]]/
			dist->ind_prob[m1][b];
		  //	    }			  
		} /* i is the first node for edge e */
	      else if( m2>=dist->dim && m2-dist->dim==mi )
		{ /* i is the second node for edge e */
		  if( m1>=dist->dim )
		    //  {
		    //	  printf( "Edge connects %ld and %ld\n", m1, m2 );
		    /* Both nodes are in the current time slice */
		    for( b=0; b<dist->num_states; b++ )
		      prob[b]*=
			dist->edge_prob[e][data->entry[t]->ddata[dist->dim_index[m1-dist->dim]]][b]/
			dist->ind_prob[m2][b];
		  //	    }
		  else if( t>0 )
		    //	    {
		    //	  printf( "Edge connects %ld and %ld\n", m1, m2 );
		    /* The first node is in the previous time slice */
		    for( b=0; b<dist->num_states; b++ )
		      prob[b]*=
			dist->edge_prob[e][data->entry[t-1]->ddata[dist->dim_index[m1]]][b]/
			dist->ind_prob[m2][b];
		  //	    }
		} /* i is the second node for edge e */	    
	    }
	}
	 
      /* Processing P(R_{t+1}|R_t,S_t) */
      if( t<data->seq_length-1 )
	{
	  //	  printf( "Considering auto-regressive transition %ld->%ld\n", t, t+1 );
	  //	  printf( "S(%ld)=%ld\n", t+1, state[t+1] );

	  dist=emission[state[t+1]];
	  if( dist->type==DIST_CONDCHOWLIU )
	    for( e=0; e<dist->num_edges; e++ )
	      {
		m1=dist->edge[e][0];
		m2=dist->edge[e][1];
		if( m1<dist->dim && m1==mi )
		  //  {
		  //   printf( "Edge connects %ld and %ld\n", m1, m2 );
		  /* i is the first node for edge e */
		  /* m2 must be from the next time slice */
		  for( b=0; b<dist->num_states; b++ )
		    prob[b]*=
		      dist->edge_prob[e][b][data->entry[t+1]->ddata[dist->dim_index[m2-dist->dim]]]/
		      dist->ind_prob[m1][b];
		//	  }
		else if( m2<dist->dim && m2==mi )
		  //	  {
		  //	    printf( "Edge connects %ld and %ld\n", m1, m2 );
		  /* i is the second node for edge e */
		  /* m1 must be from the next time slice */
		  
		  for( b=0; b<dist->num_states; b++ )
		    prob[b]*=
		      dist->edge_prob[e][data->entry[t+1]->ddata[dist->dim_index[m1-dist->dim]]][b]/
		      dist->ind_prob[m2][b];
		//	  }
	      } /* i is the second node for edge e */	    
	}
 
      /* Normalizing the probability */
      sum=0.0;
      for( b=0; b<dist->num_states; b++ )
	sum+=prob[b];
      for( b=0; b<dist->num_states; b++ )
	prob[b]/=sum;

      missing[n]=generateBernoulli( prob, dist->num_states );

      delete prob;      
    } /* Looping over missing values */
  
  return( missing );
}
						    
void HMM::InitializeMissingEntriesSequence( Sequence* data ) {
  /* !!! Assinging initial values uniformly at random */
  /* !!! Assuming binary values */
						
  /* Index variables */
  long m,t,i;

  for( m=0; m<data->num_missing_discrete_entries; m++ )
    {
      t=data->missing_discrete[m][0];
      i=data->missing_discrete[m][1];

      if( drand48()<0.5 )
	data->entry[t]->ddata[i]=0;
      else
	data->entry[t]->ddata[i]=1;
    }

  return;
}

long generateBernoulli( double* dist, long dim ) {
  /* Generates a state (1 through dim) accrding to a Bernoulli distribution with dim states and distribution
     dist for the states.  Assumes that the total probability mass in dist adds up to 1. */

  long state=0;     // Generated state
  double sum=0.0;   // Probability mass counter
  double dice_roll; // Random number

  dice_roll=drand48();
  while( sum<dice_roll )
    {
      sum+=dist[state];
      state++;
    }

  if( state==0 )
    return( 0 );
  else
    return( state-1 );
}

double generateUnitNormal( void ) {
  /* Generating a univariate normally distributed value with zero mean and unit variance */ 
  /* Box-Muller */

  /* Following "Numerical Recipes in C", 2nd edition */

  /* Two generated random variables */
  double first_value;              // First
  static double second_value;      // Second value (to be carried over for the next run)

  double v1, v2;                   // Uniformly distributed in (-1,1)
  double d;                        // Length of (v1,v2)
  double fac;                      // Multiplication factor

  if( is_first_normal_value )
    { /* Simulating two normal values */
      do 
	{
	  v1=2.0*drand48()-1.0;
	  v2=2.0*drand48()-1.0;
	  d=v1*v1+v2*v2;
	} 
      while( d>=1.0 || d==0.0 );

      fac=sqrt(-2.0*log(d)/d);
      first_value=v1*fac;
      second_value=v2*fac;
      is_first_normal_value=0;

      return( first_value );
    } /* Simulating two normal values */
  else
    { /* Using left-over value */
      is_first_normal_value=1;
      return( second_value );
    } /* Using left-over value */
}

double generateExponential( double lambda ) {
  /* Generating a sample from a univariate exponential distribution f(x)=l*exp(-l*x) */
  return( -log(1.0-drand48())/lambda );
}

double generateGamma( double a, double b ) {
  /* Generating a sample from a Gamma distribution f(x)=x^(a-1)exp(-b*x)*b^a/G(a) */

  double x=0.0;
  double u,y,v,c,p;

  long done;

  /* First sampling from f(x)=x^(a-1)exp(-x)/G(a) */

  /* The method implmented below is based on rejection method as sugested by 
     Ahrens and summarized in "The Art of Computer Programming" by D. Knuth */

  done=0;

  /* Cases of a>1 and 0<a<=1 are treated separately */
  if( a>1.0 )
    while( !done )
      {
	u=drand48();
	y=tan(M_PI*u);
	x=sqrt(2.0*a-1.0)*y+a-1.0;
	if( x>0 )
	  {
	    /* Rejection in action */
	    v=drand48();
	    if( v<=(1.0+y*y)*exp((a-1.0)*log(x/(a-1.0))-sqrt(2.0*a-1.0)*y) )
	      /* Value under the curve -- accepting */
	      done=1;
	  }
      }
  else if( a>0.0 )
    {
      c=(exp(1.0)+a)/exp(1.0);
      while( !done )
	{
	  u=drand48();
	  p=c*u;
	  if( p>1.0 )
	    {
	      x=-log((c-p)/a);
	      v=drand48();
	      if( v<=pow(x,a-1.0) )
		done=1;            
	    }
	  else
	    {
	      x=pow(p,1.0/a);
	      u=drand48();
	      if( u<=exp(-x) )
		done=1;
	    }
	}
    }

  return( x/b );
}

void SimulateNormal( double* mu, double** Sigma, double* x, long dim ) {
  /* Simulating x according to the multivariate normal distribution with mean mu
     and upper-triangle Cholesky decomposition Sigma of the covariance matrix */

  double* unitnormal;

  /* Index variable(s) */
  long i,j;

  if( dim!=1 )
    { /* Multivariate normal */
      unitnormal=new double[dim];

      for( i=0; i<dim; i++ )
	{
	  unitnormal[i]=generateUnitNormal();
	  x[i]=mu[i];
	  for( j=0; j<=i; j++ )
	    x[i]+=unitnormal[j]*Sigma[j][i];
	}

      /* Applying the covariance structure */

      delete unitnormal;
    } /* Multivariate normal */
  else
    x[0]=mu[0]+Sigma[0][0]*generateUnitNormal();

  return;
}

void Data::Simulate( Distribution** dist, long** states, long num_entries, DataPoint* datum ) {
  /* Simulating sequences of data according to distribution dist and already generated hidden states */

  /* Index variable(s) */
  long n;
  
  for( n=0; n<num_seqs; n++ )
    {
      /* Initializing sequence n */
      sequence[n]=new Sequence( num_entries );
            
#ifdef SIM_VERBOSE
      printf( "Sequence %ld", n );
      fflush( stdout );
#endif 
      /* Simulating sequence n */
      sequence[n]->Simulate( dist, states[n], datum );
#ifdef SIM_VERBOSE
      printf( "Done\n" );
#endif
    }

  is_initialized=1;

  return;
}

void Sequence::Simulate( Distribution** dist, long* states, DataPoint* datum ) {
  /* Simulating a sequence of data according to distribution dist and already generated hidden states */

  /* Index variable(s) */
  long t;
  
  for( t=0; t<seq_length; t++ )
    {
      /* Initializing observation t */
      entry[t]=new DataPoint( datum->ddim, datum->rdim );

      /* Simulating observation t */
      if( t==0 )
	dist[states[t]]->Simulate( entry[t], NULL );
      else
	dist[states[t]]->Simulate( entry[t], entry[t-1] );
#ifdef SIM_VERBOSE
      printf( "." );
#endif
    }  

  is_initialized=1;

  return;
}

void Distribution::Simulate( DataPoint* sim, DataPoint* prev_datum ) {
  /* Simulating an observation according to distribution */

  long* generated;
  double* prob_vector;
  long num_gibbs_sims;

  /* Index variable(s) */
  long i,j,e,sim_index;

  /* Temporary variables */
  long temp_value;
  long* ddata;
  double* rdata;
  double sum;

  switch( type ) {
  case DIST_FACTOR:
    for( i=0; i<dim; i++ )
      subdist[i]->Simulate( sim, prev_datum );
    break;
  case DIST_MIXTURE:
    /* Picking a mixture component */
    i=generateBernoulli( state_prob, num_states );

    /* Simulating a data point from the component */
    subdist[i]->Simulate( sim, prev_datum );

    break;
  case DIST_NESTEDDIVIDE:
    for( i=0; i<num_dists; i++ )
      subdist[i]->Simulate( sim, prev_datum );
    break;
  case DIST_BERNOULLI:
    if( dim==1 )
      sim->ddata[dim_index[0]]=generateBernoulli( state_prob, num_states );
    else
      {
	fprintf( stderr, "Multi-dimensional Bernoulli distribution simulation is not yet implemented!\n" );
	exit( -1 );
      }
    break;
  case DIST_CONDBERNOULLI:
    if( dim==1 )
      if( prev_datum )
	/* Not the first sequence entry */
	sim->ddata[dim_index[0]]=generateBernoulli( cond_state_prob[prev_datum->ddata[dim_index[0]]], num_states );
      else
	/* First sequence entry */
	sim->ddata[dim_index[0]]=generateBernoulli( state_prob, num_states );
    else
      {
	fprintf( stderr, "Multi-dimensional conditional Bernoulli distribution simulation is not yet implemented!\n" );
	exit( -1 );
      }
    break;
  case DIST_CHOWLIU:
    /* When generating the model, assuming that postprocessing step is run as well.
       Generating the nodes in the order in which they appear in the edges. */

    ddata=new long[dim];

    generated=new long[dim];
    for( i=0; i<dim; i++ )
      generated[i]=0;

    prob_vector=new double[num_states];

    for( e=0; e<num_edges; e++ )
      {
	/* Two possibilities:
	   both nodes need to be simulated, or just one */
	if( !generated[edge[e][0]] && !generated[edge[e][1]] )
	  { /* Both nodes need to be simulated */
	    /* Simulating the first node of the edge */
	    ddata[edge[e][0]]=generateBernoulli( ind_prob[edge[e][0]], num_states );
	    generated[edge[e][0]]=1;

	    /* Simulating the second node of the edge */
	    /* Creating the probability distribution first */
	    for( i=0; i<num_states; i++ )
	      prob_vector[i]=edge_prob[e][ddata[edge[e][0]]][i]/ind_prob[edge[e][0]][ddata[edge[e][0]]];

	    ddata[edge[e][1]]=generateBernoulli( prob_vector, num_states );
	    generated[edge[e][1]]=1;
	  } /* Both nodes need to be simulated */
	else
	  { /* Only one node needs to be simulated */
	    if( !generated[edge[e][0]] )
	      { /* Simulating the first node of the edge given the second */
		/* Creating the probability distribution first */
		for( i=0; i<num_states; i++ )
		  prob_vector[i]=edge_prob[e][i][ddata[edge[e][1]]]/ind_prob[edge[e][1]][ddata[edge[e][1]]];
		
		ddata[edge[e][0]]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][0]]=1;
	      } /* Simulating the first node of the edge given the second */
	    
	    if( !generated[edge[e][1]] )
	      { /* Simulating the second node of the edge given the first */
		/* Creating the probability distribution first */
		for( i=0; i<num_states; i++ )
		  prob_vector[i]=edge_prob[e][ddata[edge[e][0]]][i]/ind_prob[edge[e][0]][ddata[edge[e][0]]];

		ddata[edge[e][1]]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][1]]=1;
	      } /* Simulating the second node of the edge given the first */
	  } /* Only one node needs to be simulated */    
      }

    /* Making sure all nodes have been simulated */
    for( i=0; i<dim; i++ )
      if( !generated[i] )
	/* Simulating the node form individual probabilities */
	ddata[i]=generateBernoulli( ind_prob[i], num_states );
      
    delete prob_vector;

    delete generated;

    for( i=0; i<dim; i++ )
      sim->ddata[dim_index[i]]=ddata[i];
    delete ddata;
    
    break;
  case DIST_CONDCHOWLIU:
    /* When generating the model, assuming that postprocessing step is run as well.
       Generating the nodes in the order in which they appear in the edges. */

    ddata=new long[dim];

    generated=new long[2*dim];
    for( i=0; i<dim; i++ )
      {
	generated[i]=1;
	generated[i+dim]=0;
      }

    prob_vector=new double[num_states];

    for( e=0; e<num_edges; e++ )
      {
	/* Two possibilities:
	   both nodes need to be simulated, or just one */
	if( !generated[edge[e][0]] && !generated[edge[e][1]] )
	  { /* Both nodes need to be simulated */
	    /* Simulating the first node of the edge */
	    ddata[edge[e][0]-dim]=generateBernoulli( ind_prob[edge[e][0]], num_states );
	    generated[edge[e][0]]=1;

	    /* Simulating the second node of the edge */
	    /* Creating the probability distribution first */
	    for( i=0; i<num_states; i++ )
	      prob_vector[i]=edge_prob[e][ddata[edge[e][0]-dim]][i]/
		ind_prob[edge[e][0]-dim][ddata[edge[e][0]-dim]];

	    ddata[edge[e][1]-dim]=generateBernoulli( prob_vector, num_states );
	    generated[edge[e][1]]=1;
	  } /* Both nodes need to be simulated */
	else
	  { /* Only one node needs to be simulated */
	    if( !generated[edge[e][0]] )
	      { /* Simulating the first node of the edge given the second */
		/* Creating the probability distribution first */
		if( edge[e][1]<dim )
		  if( prev_datum )
		    temp_value=prev_datum->ddata[dim_index[edge[e][1]]];
		  else
		    temp_value=missing_value( (long)0 );
		else
		  temp_value=ddata[edge[e][1]-dim];

		if( !is_missing( temp_value ) )
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=edge_prob[e][i][temp_value]/ind_prob[edge[e][1]][temp_value];
		else
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=ind_prob[edge[e][0]][i];

		ddata[edge[e][0]-dim]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][0]]=1;
	      } /* Simulating the first node of the edge given the second */
	    
	    if( !generated[edge[e][1]] )
	      { /* Simulating the second node of the edge given the first */
		/* Creating the probability distribution first */
		if( edge[e][0]<dim )
		  if( prev_datum )		    
		    temp_value=prev_datum->ddata[dim_index[edge[e][0]]];
		  else
		    temp_value=missing_value( (long)0 );
		else
		  temp_value=ddata[edge[e][0]-dim];

		if( !is_missing( temp_value ) )
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=edge_prob[e][temp_value][i]/ind_prob[edge[e][0]][temp_value];
		else
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=ind_prob[edge[e][1]][i];

		ddata[edge[e][1]-dim]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][1]]=1;
	      } /* Simulating the second node of the edge given the first */
	  } /* Only one node needs to be simulated */    
      }

    /* Making sure all nodes have been simulated */
    for( i=dim; i<2*dim; i++ )
      if( !generated[i] )
	/* Simulating the node form individual probabilities */
	ddata[i-dim]=generateBernoulli( ind_prob[i], num_states );
      
    delete prob_vector;

    delete generated;

    for( i=0; i<dim; i++ )
      sim->ddata[dim_index[i]]=ddata[i];
    delete ddata;
    
    break;
  case DIST_ME_BIVAR:
    /* Simulation by Gibbs sampling */
    if( is_sim_initialized )
      /* Already performed burn-off */
      num_gibbs_sims=GIBBS_LAG;
    else
      { /* Need to perform the burn-off first */
	is_sim_initialized=1;
	last_sim=new long[dim];

	/* Initializing at random */
	for( i=0; i<dim; i++ )
	  if( drand48()>0.5 )
	    last_sim[i]=1;
	  else
	    last_sim[i]=0;

	num_gibbs_sims=GIBBS_BURNIN;
      } /* Need to perform the burn-off first */

    for( sim_index=0; sim_index<num_gibbs_sims; sim_index++ )
      /* Performing Gibbs sampling */
      for( i=0; i<dim; i++ )
	{ /* Resampling component i */
	  /* Calculating probability of outcome 0: 1/(exp(sum)+1) */
	  sum=0.0;
	  for( j=0; j<i; j++ )
	    if( last_sim[j]==1 )
	      sum+=sigma[i][j];
	  sum+=sigma[i][i];
	  for( j=i+1; j<dim; j++ )
	    if( last_sim[j]==1 )
	      sum+=sigma[j][i];
	  
	  /* Sampling */
	  if( drand48()<1.0/(exp(sum)+1.0) )
	    last_sim[i]=0;
	  else
	    last_sim[i]=1;
	} /* Resampling component i */

    /* Copying the last sample into the output */
    for( i=0; i<dim; i++ )
      sim->ddata[dim_index[i]]=last_sim[i];

    break;
  case DIST_BN_ME_BIVAR:
    for( i=0; i<dim; i++ )
      {
	/* Computing the probability of outcome 0: 1/(exp(sum)+1) */
	sum=sigma[sim_order[i]][0];
	for( j=1; j<num_features[sim_order[i]]; j++ )
	  if( sim->ddata[dim_index[feature_index[sim_order[i]][j]]]==1 )
	    sum+=sigma[sim_order[i]][j];
	
	/* Sampling */
	if( drand48()<1.0/(1.0+exp(sum)) )
	  sim->ddata[dim_index[sim_order[i]]]=0;
	else
	  sim->ddata[dim_index[sim_order[i]]]=1;
      }
    break;
  case DIST_BN_CME_BIVAR:
    for( i=0; i<dim; i++ )
      if( !prev_datum )
	{ /* First entry in a sequence -- using state_prob */
	  if( drand48()<state_prob[i] )
	    sim->ddata[dim_index[i]]=1;
	  else
	    sim->ddata[dim_index[i]]=0;
	} /* First entry in a sequence -- using state_prob */
      else
	{
	  /* Computing the probability of outcome 0: 1/(exp(sum)+1) */
	  sum=sigma[sim_order[i]][0];
	  for( j=1; j<num_features[sim_order[i]]; j++ )
	    if( feature_index[sim_order[i]][j]<dim )
	      {
		if( sim->ddata[dim_index[feature_index[sim_order[i]][j]]]==1 )
		  sum+=sigma[sim_order[i]][j];
	      }
	    else
	      {
		if( prev_datum->ddata[dim_index[feature_index[sim_order[i]][j]-dim]]==1 )
		  sum+=sigma[sim_order[i]][j];
	      }
	      
	  /* Sampling */
	  if( drand48()<1.0/(1.0+exp(sum)) )
	    sim->ddata[dim_index[sim_order[i]]]=0;
	  else
	    sim->ddata[dim_index[sim_order[i]]]=1;
	}
    break;
  case DIST_DELTAEXP:
    i=generateBernoulli( mix_prob, num_states );
    if( i==0 )
      sim->rdata[dim_index[0]]=0.0;
    else
      sim->rdata[dim_index[0]]=generateExponential( exp_param[i-1] );
    break;
  case DIST_DELTAGAMMA:
    i=generateBernoulli( mix_prob, num_states );
    if( i==0 )
      sim->rdata[dim_index[0]]=0.0;
    else
      sim->rdata[dim_index[0]]=generateGamma( gamma_param1[i-1], gamma_param2[i-1] );
    break;
  case DIST_NORMAL:
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }

    rdata=new double[dim];
    SimulateNormal( mu, chol_sigma, rdata, dim );
    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;
    break;
  case DIST_NORMALCHAIN:
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }
    if( !chol_first_sigma )
      {
	chol_first_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_first_sigma[i]=new double[dim];

	cholesky( first_sigma, dim, chol_first_sigma );
      }

    rdata=new double[dim];

    if( prev_datum )
      { /* Not the first observation in the sequence */

	/* Allocating mean vector */
	prob_vector=new double[dim];

	/* Calculating the value for the mean vector */
	for( i=0; i<dim; i++ )
	  {
	    prob_vector[i]=mu[i];
	    for( e=0; e<dim; e++ )
	      prob_vector[i]+=W[i][e]*prev_datum->rdata[dim_index[e]];
	  }

	SimulateNormal( prob_vector, chol_sigma, rdata, dim );

	delete prob_vector;
      } /* Not the first observation in the sequence */
    else
      /* First observation in the sequence */
      SimulateNormal( first_mu, chol_first_sigma, rdata, dim );

    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;

    break;
  case DIST_NORMALCL:
    /* !!! May be replaced by more efficient sampling !!! */
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }

    rdata=new double[dim];
    SimulateNormal( mu, chol_sigma, rdata, dim );
    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;
    break;
  case DIST_LOGISTIC:
  case DIST_TRANSLOGISTIC:
    fprintf( stderr, "Simulation from the logistic data is not yet implemented!\n" );
    exit( -1 );
    break;
  default:;
  }

  return;
}
